// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  J_v(PD4990A) ̃G~[V for Win32
// ---------------------------------------------------------------------------
//	$Id: Calender.cpp,v 1.2 1999/01/11 15:32:07 cisc Exp $
//	ETP, 1Hz @\

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "Calender.h"

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
//	#define CALENDER_LOGGING
#endif

#ifdef CALENDER_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("CALENDER.DMP", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
#endif

// ---------------------------------------------------------------------------
//	Construct/Destruct
// ---------------------------------------------------------------------------

Calender::Calender()
{
	LogBegin();
	
	diff.dwLowDateTime=0;
	diff.dwHighDateTime=0;
	Reset();
}

Calender::~Calender()
{
	LogEnd();
}

// ---------------------------------------------------------------------------
//	Eo
// ---------------------------------------------------------------------------

void Calender::Reset()
{
	datain = 0;
	dataoutmode = 0;
	strobe = 0;
	cmd = 0x80;
}

uint8 Calender::In(uint port)
{
	if (dataoutmode)
		return (reg[0] & 1) << 4;
	else
	{
//		SYSTEMTIME st;
//		GetLocalTime(&st);
//		return (st.wSecond & 1) << 4;
		return 0;
	}
}

void Calender::Out(uint port, uint data)
{
	switch (port)
	{
		uint modified;

	case 0x10:
		pcmd = data & 7;
		datain = (data >> 3) & 1;
		break;

	case 0x40:
		modified = strobe ^ data;
		strobe = data;
		if (modified & data & 2)
			Command();
		if (modified & data & 4)
			ShiftData();
		break;
	}
}

// ---------------------------------------------------------------------------
//	digit->BCD
// ---------------------------------------------------------------------------

inline uint8 Calender::BCD(uint8 n)
{
	return (n/10)*16 + (n%10);
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

void Calender::Command()
{
	if (pcmd == 7)
		cmd = scmd | 0x80;
	else
		cmd = pcmd;

	Log1("Command = %.2x\n", cmd);
	switch (cmd & 15)
	{
	case 0x00:			// register hold
		hold = true;
		dataoutmode = false;
		break;

	case 0x01:			// register shift
		hold = false;
		dataoutmode = true;
		break;
	
	case 0x02:			// time set
		SetTime();
		hold = true;
		dataoutmode = true;
		break;

	case 0x03:			// time read
		GetTime();
		hold = true;
		dataoutmode = false;
		break;
	}
}

// ---------------------------------------------------------------------------
//	f[^Vtg
// ---------------------------------------------------------------------------

void Calender::ShiftData()
{
	if (hold)
	{
		if (cmd & 0x80)
		{
			// shift sreg only
//			Log1("Shift HS %d\n", datain);
			scmd = (scmd >> 1) | (datain << 3);
		}
		else
		{
//			Log1("Shift HP -\n", datain);
		}
	}
	else
	{
		if (cmd & 0x80)
		{
			reg[0] = (reg[0] >> 1) | (reg[1] << 7);
			reg[1] = (reg[1] >> 1) | (reg[2] << 7);
			reg[2] = (reg[2] >> 1) | (reg[3] << 7);
			reg[3] = (reg[3] >> 1) | (reg[4] << 7);
			reg[4] = (reg[4] >> 1) | (reg[5] << 7);
			reg[5] = (reg[5] >> 1) | (scmd   << 7);
			scmd   = (scmd   >> 1) | (datain << 3);
//			Log1("Shift -S %d\n", datain);
		}
		else
		{
			reg[0] = (reg[0] >> 1) | (reg[1] << 7);
			reg[1] = (reg[1] >> 1) | (reg[2] << 7);
			reg[2] = (reg[2] >> 1) | (reg[3] << 7);
			reg[3] = (reg[3] >> 1) | (reg[4] << 7);
			reg[4] = (reg[4] >> 1) | (datain << 7);
//			Log1("Shift -P %d\n", datain);
		}
	}
}

// ---------------------------------------------------------------------------
//	ԌvZ
// ---------------------------------------------------------------------------

static void FTADD(FILETIME& d, const FILETIME& s)
{
	if (uint32(d.dwLowDateTime) > uint32(0xffffffff-s.dwLowDateTime))
	{
		d.dwHighDateTime++;
	}
	d.dwHighDateTime += s.dwHighDateTime;
	d.dwLowDateTime  += s.dwLowDateTime;
}


static void FTSUB(FILETIME& d, const FILETIME& s)
{
	if (uint32(d.dwLowDateTime) < uint32(s.dwLowDateTime))
	{
		d.dwHighDateTime--;
	}
	d.dwHighDateTime -= s.dwHighDateTime;
	d.dwLowDateTime  -= s.dwLowDateTime;
}

// ---------------------------------------------------------------------------
//	Ԏ擾
// ---------------------------------------------------------------------------

void Calender::GetTime()
{
	SYSTEMTIME st;
	FILETIME ft;
	
	GetLocalTime(&st);
	SystemTimeToFileTime(&st, &ft);

	Log3("CT %d/%.2d/%.2d", st.wYear, st.wMonth, st.wDay);
	Log3(" %.2d/%.2d/%.2d", st.wHour, st.wMinute, st.wSecond);
	Log2(" %.8x %.8x\n", ft.dwHighDateTime, ft.dwLowDateTime);
	
	Log2("DIFF %.8x %.8x\n", diff.dwHighDateTime, diff.dwLowDateTime);
	FTADD(ft, diff);
	FileTimeToSystemTime(&ft, &st);

	Log3("PT %d/%.2d/%.2d", st.wYear, st.wMonth, st.wDay);
	Log3(" %.2d/%.2d/%.2d", st.wHour, st.wMinute, st.wSecond);
	Log2(" %.8x %.8x\n", ft.dwHighDateTime, ft.dwLowDateTime);

	reg[5] = BCD(uint8(st.wYear % 100));
	reg[4] = st.wMonth * 16 + st.wDayOfWeek;
	reg[3] = BCD(uint8(st.wDay));
	reg[2] = BCD(uint8(st.wHour));
	reg[1] = BCD(uint8(st.wMinute));
	reg[0] = BCD(uint8(st.wSecond));
}

// ---------------------------------------------------------------------------
//	Ԑݒ
// ---------------------------------------------------------------------------

#define BCDTOI(i)   ((i >> 4) * 10 + (i & 0x0f))

void Calender::SetTime()
{
	SYSTEMTIME st;
	SYSTEMTIME nt;
	FILETIME nft;
	FILETIME sft;

	GetLocalTime(&st);

	nt.wYear = (cmd & 0x80) ? 1900+BCDTOI(reg[5]) : st.wYear;
	if (nt.wYear < 1980)
		nt.wYear += 100;
	nt.wMonth  = reg[4] >> 4;
	nt.wDay    = BCDTOI(reg[3]);
	nt.wHour   = BCDTOI(reg[2]);
	nt.wMinute = BCDTOI(reg[1]);
	nt.wSecond = BCDTOI(reg[0]);

	SystemTimeToFileTime(&st, &sft);
	
	Log3("CT %d/%.2d/%.2d", st.wYear, st.wMonth, st.wDay);
	Log3(" %.2d/%.2d/%.2d", st.wHour, st.wMinute, st.wSecond);
	Log2(" %.8x %.8x\n", sft.dwHighDateTime, sft.dwLowDateTime);

	Log3("NT %d/%.2d/%.2d", nt.wYear, nt.wMonth, nt.wDay);
	Log3(" %.2d/%.2d/%.2d", nt.wHour, nt.wMinute, nt.wSecond);
	Log2(" %.8x %.8x\n", nft.dwHighDateTime, nft.dwLowDateTime);

	SystemTimeToFileTime(&nt, &nft);
	FTSUB(nft, sft);
	diff = nft;

	Log2("DIFF = %.8x %.8x\n", diff.dwHighDateTime, diff.dwLowDateTime);
}
	