// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  CRTC (PD3301) ̃G~[V
// ---------------------------------------------------------------------------
//	$Id: Crtc.cpp,v 1.2 1999/01/11 15:32:07 cisc Exp $

#include <stdio.h>
#include <string.h>
#include "CRTC.h"
#include "PC88.h"
#include "misc.h"

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
	#define CRTC_LOGGING
#endif

#ifdef CRTC_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("CRTC.DMP", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
#endif

// ---------------------------------------------------------------------------
// \z/
// ---------------------------------------------------------------------------

CRTC::CRTC(PC88& _pc) : pc(_pc)
{
	LogBegin();
	Reset(false);
	paravalid = false;
}

CRTC::~CRTC()
{
	LogEnd();
}

// ---------------------------------------------------------------------------
//	I/O
// ---------------------------------------------------------------------------

uint8 CRTC::In(uint p)
{
	switch (p)
	{
	case 0x50:
		return Command(false, 0);
	case 0x51:
		return status;
	}
	return 0xff;
}

void CRTC::Out(uint p, uint d)
{
	switch(p)
	{
	case 0x30:
		setting.widefont = !(d & 1);
		shoulderase = true;
		break;
	case 0x50: case 0x51:
		Command((p&1) != 0, d);
		break;
	}
}

void CRTC::Reset(bool l200)
{
	shoulderase = true;
	setting.cursor_type = cursormode =  4;
	tvramsize = 0;
	status = 0;
	line200 = l200;
}

// ---------------------------------------------------------------------------
//	`p̏擾
// ---------------------------------------------------------------------------

CRTC::DrawStatus CRTC::GetSetting(ScreenSetting* s)
{
	*s = setting;
	if (shoulderase)
	{
		shoulderase = false;
		return erase;
	}
	return (status & 0x10) ? on : off;
}

// ---------------------------------------------------------------------------
//	eLXgʂ擾
// ---------------------------------------------------------------------------

void CRTC::GetTVRAM(uint8* dest)
{
	if (!pc.GetTVRAM(dest, tvramsize))
	{
		shoulderase = true;
		status = (status & ~0x10) | 0x08;		// VE-off U-on
	}
}

// ---------------------------------------------------------------------------
//	R}h
// ---------------------------------------------------------------------------

uint8 CRTC::Command(bool a0, uint8 data)
{
	uint8 result = 0xff;

	Log1(a0 ? "\ncmd:%.2x " : "%.2x ", data);
	
	if (a0)
	{
		cmdc = 0;
		cmdm = data >> 5;
	}
	switch (cmdm)
	{
	case 0:				// RESET
		switch (cmdc)
		{
		case 0:	
			Reset(line200);
			
			shoulderase = true;
			setting.cursor_type = cursormode =  4;
			tvramsize = 0;
			status = 0;
			
			status &= ~0x16;
			break;
			
		case 1:
			setting.width = (data & 0x7f)+2;
			break;
			
		case 2:
			{
				static const uint brtable[4] = { 267, 533, 800, 1067 };
				setting.blinkrate = brtable[data >> 6];
				setting.height = Min(50, (data & 0x3f) + 1);
					// attrcache is 0x1000 bytes large
			}
			break;
			
		case 3:
			setting.skipline = 0 != (data & 0x80);
			cursormode = (data >> 5) & 3;
			setting.linesperchar = (data & 0x1f) + 1;
			if (line200)
				setting.linesperchar *= 2;
			setting.linecharlimit = Min(setting.linesperchar, 16);
			break;
			
		case 4:
//			horizontalretrace = (data & 0x1f) + 2;
			{
				int vrtc = ((data >> 5) & 7) + 1;
				int blankline = line200 ? 120 : 48;

				setting.vblank = Max(0,  vrtc - (blankline+setting.linesperchar-1) / setting.linesperchar);
			}
			break;

		case 5:
			setting.attrperline = (data & 0x1f) + 1;
			
			paravalid = true;
			switch ((data >> 5) & 7)
			{
			case 0:			// transparent b/w
				setting.color = false; 
				setting.controlcode = true;
				break;
			case 1:			// no attribute b/w
				setting.color = false;
				setting.attrperline = 0;
				setting.controlcode = false;
				break;
			case 2:			// transparent color
				setting.color = true;  
				setting.controlcode = true;
				break;
			case 4:			// non transparent b/w
				setting.color = false; 
				setting.controlcode = true;
				break;
			case 5:			// non transparent b/w
				setting.color = false; 
				setting.controlcode = false;
				break;

			default:
				paravalid = false; 
				break;
			}

			break;
		}
		break;
		
	case 1:			// START DISPLAY
		if (cmdc == 0)
		{
			if (!setting.skipline)
				tvramsize = setting.height * (setting.width + setting.attrperline*2);
			else
				tvramsize = ((setting.height+1)/2) * (setting.width + setting.attrperline*2);
			
			setting.inversed = 0 != (data & 1);
			if (paravalid && setting.width <= 80)
				status |= 0x10;
			else
				status &= ~0x10;
			Log3(" Start Display [%.2x;%x;%2d]", status, paravalid, setting.width);
			Log2(" vblank %d tvram size = %.4x", setting.vblank, tvramsize);
		}
		break;
		
	case 4:			// LOAD CURSOR POSITION
		switch (cmdc)
		{
		case 0:
			setting.cursor_type = (data & 1) ? cursormode : 4;
			break;
		case 1:
			setting.cursor_x = data;
			break;
		case 2:
			setting.cursor_y = data;
			break;
		}
		break;
		
	case 2:			// SET INTERRUPT MASK
		if (!(data & 1))
		{
			shoulderase = true;
			status = 0;
		}
		break;
		
	case 3:			// READ LIGHT PEN
		status &= ~1;
		break;
		
	case 5:			// RESET INTERRUPT
		break;
		
	case 6:			// RESET COUNTERS
		shoulderase = true;			// ^C~OɂĂ
		status = 0;					// ȂƂ邩H
		break;
		
	default:
		break;
	}
	cmdc++;
	return result;
}

