// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  DMAC (uPD8257) ̃G~[V
//	ECRTC ̂ƂlĂȂB
// ---------------------------------------------------------------------------
//	$Id: DMAC.cpp,v 1.2 1999/01/11 15:32:07 cisc Exp $

#include <stdio.h>
#include "DMAC.h"
#include "misc.h"

#ifndef RELEASEBUILD
//	#define DMAC_LOGGING
#endif

#ifdef DMAC_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("DMAC.dmp", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
#endif


// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

DMAC::DMAC()
{
	LogBegin();
	Reset();
}

DMAC::~DMAC()
{
	LogEnd();
}

// ---------------------------------------------------------------------------
//	I/O
// ---------------------------------------------------------------------------

void DMAC::Reset()
{
	ff = false;
	for (int i=0; i<4; i++)
		enabled[i] = false;
}

void DMAC::Out(uint d, uint p)
{
	d &= 15;
	int bank = d / 2;
	if (bank < 4)
	{
		if (!(d & 1))
		{
			if (!ff)
				addr[bank] = (addr[bank] & 0xff00) | p;
			else
				addr[bank] = (addr[bank] & 0x00ff) | (p << 8);
			Log2("Bank %d: addr  = %.4x\n", bank, addr[bank]);
		}
		else
		{
			if (!ff)
				count[bank] = (count[bank] & 0xff00) | p;
			else
				count[bank] = (count[bank] & 0x00ff) | (p << 8);
			Log3("Bank %d: count = %.4x  flag = %.4x\n", bank, count[bank] & 0x3fff, count[bank] & 0xc000);
		}
		ff = !ff;
	}
	else
	{
		if (d == 8)
		{
			autoinit = !!(p & 0x80);
			Log0("DMA Control ");
			for (int i=0; i<4; i++)
			{
				enabled[i] = !!(p & 1);
				Log0(enabled[i] ? "*" : "-");
				p >>= 1;
				if (enabled[i])
					ptr[i] = addr[i];
			}
			ff = false;
			Log0("\n");
		}
	}
}

uint8 DMAC::In(uint d)
{
	if ((d & 0x0f) == 8)
	{
		ff = false;
	}
	return 0;
}


// ---------------------------------------------------------------------------
//	DMAC ɓ]vB
//	bank	ΏۃoN̔ԍ
//	a		]JnAhX󂯎|C^
//	c		]oCg󂯎|C^
//	write	ݑEǂݍݑ̑I
//	Ԃl	]łoCg
// ---------------------------------------------------------------------------

bool DMAC::Request(uint bank, uint16 * a, uint16 * c, bool write)
{
	bool result = true;

	if (bank < 4)
	{
		uint16 nbytes = Min(*c, 0xffff & (addr[bank]+(count[bank] & 0x3fff) + 1 - ptr[bank]));
		*a = ptr[bank];
		*c = nbytes;
		Log3("DMA Request ch: %d   addr: %.4x  bytes: %.4x   ", bank, *a, *c);
		result = false;
		if (enabled[bank])
		{
			switch (count[bank] & 0xc000)
			{
			case 0x0000: result = !write; break;
			case 0x4000: result =  write; break;
			case 0x8000: result = !write; break;
			}
		}
		
		if (result)
		{
			ptr[bank] += nbytes;
			if (((ptr[bank] - addr[bank]) & 0xffff) > (count[bank] & 0x3fff))
			{
				ptr[bank] = addr[bank];
				enabled[bank] = autoinit;
			}
		}
		Log0(result ? "true\n" : "false\n");
//		result = true;
	}
	return result;
}

