// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	Emulation of Floppy Disk Controller uPD765A
// ---------------------------------------------------------------------------
//	$Id: FDC.h,v 1.2 1999/01/11 15:32:08 cisc Exp $

#if !defined(AFX_FDC_H__CC0B0147_DEED_11D1_BC2E_00A0C90AC9DA__INCLUDED_)
#define AFX_FDC_H__CC0B0147_DEED_11D1_BC2E_00A0C90AC9DA__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "config.h"

class SubSystem;
class FDU;

#define NUM_FLOPPY_DRIVES		2	//	FDC ɐڑĂhCu̐

// ---------------------------------------------------------------------------
//	FDC (765)
//
class FDC  
{
public:
	enum Result
	{
		ST0_NR = 0x000008, ST0_EC = 0x000010, ST0_SE = 0x000020, 
		ST0_AT = 0x000040, ST0_IC = 0x000080, ST0_AI = 0x0000c0,

		ST1_MA = 0x000100, ST1_NW = 0x000200, ST1_ND = 0x000400,
		ST1_OR = 0x001000, ST1_DE = 0x002000, ST1_EN = 0x008000,

		ST2_MD = 0x010000, ST2_BC = 0x020000, ST2_SN = 0x040000, 
		ST2_SH = 0x080000, ST2_NC = 0x100000, ST2_DD = 0x200000, 
		ST2_CM = 0x400000,

		ST3_HD = 0x04, ST3_TS = 0x08, ST3_T0 = 0x10, ST3_RY = 0x20,
		ST3_WP = 0x40, ST3_FT = 0x80,
	};
	struct ID
	{
		uint8 c, h, r, n;
	};
	struct WIDDESC
	{
		ID* idr;
		uint8 n;
		uint8 sc;
		uint8 gpl;
		uint8 d;
	};

protected:
	enum Stat
	{
		S_D0B = 0x01, S_D1B = 0x02, S_D2B = 0x04, S_D3B = 0x08,
		S_CB  = 0x10, S_NDM = 0x20, S_DIO = 0x40, S_RQM = 0x80,
	};

public:
	FDC();
	virtual ~FDC();

	bool Initialize(SubSystem*);
	void Reset();
	void DriveControl(uint8 data) {}		// 2HD/2DD ؂ւƂ
	void MotorControl(uint8 data) {}		// [^[
	void SetData(uint8 data);				// f[^Zbg
	virtual void TC();						// TC 
	uint8 Status();							// Xe[^X
	uint8 GetData();						// f[^擾

protected:
	enum Phase
	{
		IdlePhase, CommandPhase, ExecutePhase, ExecReadPhase, 
		ExecWritePhase, ResultPhase, TCPhase, ErrorPhase,
	};
	
	virtual void Seek(uint dr, uint cy);
	virtual void ReadData(bool deleted);
	virtual void ReadID();
	virtual void WriteData(bool last, bool deleted);
	virtual void ReadDiag(bool first);
	virtual bool ShiftPhase(Phase ph, uint8 stbits, bool notc);
	virtual void WriteID();

	void Txt(char*);

	FDU* fdu[NUM_FLOPPY_DRIVES];
	uint8 status;			// Xe[^XWX^
	bool tcin;

private:
	typedef void (FDC::*CommandFunc)();
	
	bool IDIncrement(ID*);
	void GetSectorParameters();

	void ShiftToIdlePhase();
	void ShiftToCommandPhase(int);
	void ShiftToExecutePhase(uint8);
	void ShiftToExecReadPhase(int);
	void ShiftToExecWritePhase(int);
	void ShiftToResultPhase(int);
	void ShiftToResultPhase7();
	void ShiftToErrorPhase(uint8);

	void SetDataIdle();
	void SetDataCommand();
	void SetDataExecute();
	void SetDataExecWrite();
	void GetDataResult();
	void GetDataExecute();
	void GetDataExecRead();
	uint8 GetDeviceStatus(uint dr);
	
	uint8* buffer;
	uint8* bufptr;
	SubSystem* system;
	int count;				// Exec*Phase ł̓]coCg
	uint8 data;				// f[^WX^
	uint8 command;			// ݏ̃R}h
	Phase phase;
	Phase tcprevphase;
	bool int_requested;		// SENCEINTSTATUS ̌Ăяov

	uint8 pcn;
	uint8 eot;
	uint8 hdu;			// HD US1 US0
	uint8 hdunext;
	uint8 dtl;
	uint8 seekstate;
	bool lastsect;
	ID idr;
	Result result;
	Result error;
	WIDDESC wid;

	static const CommandFunc CommandTable[32];

	void CmdInvalid();
	void CmdSpecify();
	void CmdRecalibrate();
	void CmdSenceIntStatus();
	void CmdSenceDeviceStatus();
	void CmdSeek();
	void CmdReadData();
	void CmdReadDeletedData();
	void CmdWriteData();
	void CmdWriteDeletedData();
	void CmdReadID();
	void CmdReadDiagnostic();
	void CmdWriteID();
};

#endif // !defined(AFX_FDC_H__CC0B0147_DEED_11D1_BC2E_00A0C90AC9DA__INCLUDED_)
