// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  P88SR tH[}bgɑΉ FDU
// ---------------------------------------------------------------------------
//	$Id: FDUB.cpp,v 1.2 1999/01/11 15:32:08 cisc Exp $

#define STRICT
#include <windows.h>

#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "misc.h"
#include "WinCriticalSection.h"
#include "FDUB.h"

// ---------------------------------------------------------------------------
//	\zEj
// ---------------------------------------------------------------------------

FDUB::FDUB()
{
	file = 0;
	track = 0;
	idleft = 0;
}

FDUB::~FDUB()
{
	if (file)
		Unmount();
}

// ---------------------------------------------------------------------------
//	FDUB::Mount
//	C[WǂݍށihCuɃfBXNj
//
//	fio			t@C
// ---------------------------------------------------------------------------

bool FDUB::Mount(FileIO* fio)
{
	file = fio;
	
	// wb_[ǂݍ

	file->Seek(0, FileIO::begin);
	if (sizeof(header) != file->Read(&header, sizeof(header)))
		return false;

	// gbN 164 gbNȂC[W΍
	if (header.disktype == 0)
		memset(header.trackptr + 84, 0, 320);
	
	if (header.trackptr[0] == 0x2a0)
		memset(header.trackptr + 160, 0, 16);

	// wb_[`FbN
	// ̃C[Wt@Cɂ ID ƂTO݂Ȃ̂
	// ʓ|smB

	int i;
	int failed = 0;
	for (i=0; i<17; i++)
	{
		if (!header.title[i]) break;
	}
	if (i == 17)
		failed = 1;

	for (i=0; i<9; i++)
		failed |= header.reserved[i];

	if (header.disktype != 0)
		failed = 1;

	if (failed)
		return false;
	
	idleft = 0;

	writeprotected = (file->GetFlags() & FileIO::readonly) || header.readonly;
	headermodified = false;
	return true;
}

// ---------------------------------------------------------------------------
//	FDUB::SetHead
//	wbh̎w
// ---------------------------------------------------------------------------

inline bool FDUB::SetHead(uint8 hd)
{
	int nt = (track & ~1) + (hd & 1);
	if (track != nt)
	{
		track = nt;
		idleft = 0;
	}
	return true;
}

// ---------------------------------------------------------------------------
//	FDUB::ReadID
//	ZN^ƂĂ
// ---------------------------------------------------------------------------

FDC::Result FDUB::ReadID(uint8 flags, ID* id)
{
	if (!file)
		return Result(FDC::ST0_AT | FDC::ST1_MA);

	SetHead(flags);

	int i=0;
	do
	{
		if (!idleft)
		{
			if (0 == header.trackptr[track])	// unformated
				return Result(FDC::ST0_AT | FDC::ST1_MA);
			file->Seek(header.trackptr[track], FileIO::begin);
		}
		else if (sectorsize)
		{
			file->Seek(sectorsize, FileIO::current);
		}
		
		if (sizeof(sector) != file->Read(&sector, sizeof(sector)))
			return Result(FDC::ST0_AT | FDC::ST1_ND);
		
		if (!idleft)
			idleft = sector.sectors;
		sectorsize = sector.length;
		idleft--;
		*id = sector.id;
		if ((flags & 0x40) == sector.density)
			continue;
		return Result(0);
	} while (++i < sector.sectors);
	idleft = 0;
	return Result(FDC::ST0_AT | FDC::ST1_MA);
}

// ---------------------------------------------------------------------------
//	FDUB::Seek
//	w肳ꂽV_[ԍփV[N
//
//	cyrinder V[N
// ---------------------------------------------------------------------------

FDC::Result FDUB::Seek(uint8 cyrinder)
{
	cyrinder = Min(cyrinder, 41);
	
	track = cyrinder * 2;
	idleft = 0;
	return Result(0);
}

// ---------------------------------------------------------------------------
//	FDUB::ReadSector
//	ZN^ǂ
//
//	head	wbhԍ
//	id		ǂݍރZN^̃ZN^ ID
//	data	f[^̓]
// ---------------------------------------------------------------------------

FDC::Result FDUB::ReadSector(uint8 flags, ID id, uint8* data)
{
	if (!file || !SetHead(flags))
		return Result(FDC::ST0_AT | FDC::ST1_MA);

	int start = idleft;
	do
	{
		ID readid;
		if (ReadID(flags, &readid) & FDC::ST0_AT)
			return Result(FDC::ST0_AT | FDC::ST1_ND);

		if (!memcmp(&readid, &id, 4))
		{
			int readsize = Min(0x2000, sectorsize);
			sectorsize -= readsize;
			if (readsize == file->Read(data, readsize) && !sector.status)
				return Result(sector.deleted & 0x10 ? FDC::ST2_CM : 0);
			else
				return Result(FDC::ST0_AT | FDC::ST1_DE | FDC::ST2_DD);
		}
	}
	while (start != idleft);
	return Result(FDC::ST0_AT | FDC::ST1_ND);
}

// ---------------------------------------------------------------------------
//	FDUB::WriteSector
//	ZN^ɏ
//
//	Not implemented yet.
// ---------------------------------------------------------------------------

FDC::Result FDUB::WriteSector(uint8 flags, ID id, const uint8* data, bool deleted)
{
	if (!file || !SetHead(flags))
		return Result(FDC::ST0_AT | FDC::ST1_MA);
	if (writeprotected)
		return Result(FDC::ST0_AT | FDC::ST1_NW);

	int start = idleft;
	do
	{
		ID readid;
		if (ReadID(flags, &readid) & FDC::ST0_AT)
			return Result(FDC::ST0_AT | FDC::ST1_ND);

		if (!memcmp(&readid, &id, 4))
		{
			int writesize = 0x80 << Min(8, id.n);

			file->Seek(-int32(sizeof(sector)), FileIO::current);
			sector.deleted = deleted ? 0x10 : 0x00;
			
			if (writesize > sectorsize)
			{
				// ZN^zu蒼
				// gbNTCY͑邱Ƃ
				if (RemapTrack())
					file->Write(data, writesize);
			}
			else
			{
				file->Write(&sector, sizeof(sector));
				file->Write(data, writesize);
				if (writesize < sectorsize)
					file->Seek(sectorsize-writesize, FileIO::current);
			}
		
			sectorsize = 0;
			return Result(0);
		}
	}
	while (start != idleft);
	return Result(FDC::ST0_AT | FDC::ST1_ND);
}

// ---------------------------------------------------------------------------
//	FDUB::RemapTrack
//	WriteID  N < ID  N ȃZN^ւ݂̏̍ۂ̃ZN^ׂ̍Č
// ---------------------------------------------------------------------------

bool FDUB::RemapTrack()
{
	int i;
	int iid = sector.sectors - idleft - 1;		// 0 `
	
	struct Sec
	{
		SectorHeader sh;
		uint8* data;
	};

	Sec sec[0x100];
	uint8 work[0x2000];

	// gbÑZN^ǂݍ
	uint8* pw = work;
	
	file->Seek(header.trackptr[track], FileIO::begin);
	for (i=0; i<sector.sectors; i++)
	{
		file->Read(&sec[i], sizeof(SectorHeader));
		sec[i].data = pw;
		file->Read(pw, sec[i].sh.length);
		pw += sec[i].sh.length;
	}

	// ׂZN^߂
	int sows = 0x80 << Min(sector.id.n, 8);		// ރZN^̃TCY
	sec[iid].sh.length = sows;
	sec[iid].sh.status = 0;
	sows -= 0x80;
	
	for (i=iid+1; sows > 0 && i<sector.sectors-1; i++)
	{
		sec[i].data = 0;
		sows -= sec[i].sh.length + 0xc0;
	}
	if (sows > 0)
	{
		int gap4b = 0x1800 - (sec[i].data - work);
		sec[i].data = 0;
		sows -= sec[i].sh.length + 0x80 + gap4b;
	}
	for (i=0; sows > 0 && i<iid; i++)
	{
		sec[i].data = 0;
		sows -= sec[i].sh.length + 0x80;
	}
	if (sows > 0)
	{
		sec[iid].data = 0;
	}

	// VgbÑTCY߂
	int sont = 0;
	int sot = 0;
	int nos = 0;
	
	for (i=0; i<sector.sectors; i++)
	{
		if (sec[i].data)
		{
			int sos = sec[i].sh.length + sizeof(SectorHeader);
			sot += sos;
			nos++;
			sont += sos;
		}
	}

	// VgbN݂͌̃gbN̈Ɏ܂邩H
	if (sont <= sot)
	{
		file->Seek(header.trackptr[track], FileIO::begin);
	}
	else
	{
		// ܂ȂȂfBXNC[W̍ŌɃgbNp
		file->Seek(0, FileIO::end);
		header.trackptr[track] = file->Tellp();
	}

	// gbN̏
	int pos;
	for (i=0; i<sector.sectors; i++)
	{
		if (sec[i].data)
		{
			sec[i].sh.sectors = nos;
			file->Write(&sec[i].sh, sizeof(SectorHeader));
			if (i != iid)
				file->Write(sec[i].data, sec[i].sh.length);
			else
			{
				pos = file->Tellp();
				file->Seek(0x80 << Min(8, sector.id.n), FileIO::current);
			}
		}
	}

	idleft = 0;

	if (!sec[iid].data)
		return false;

	file->Seek(pos, FileIO::begin);
	return true;
}

// ---------------------------------------------------------------------------
//	FDUB::SenceDeviceStatus
//	foCXEX^[^X𓾂
// ---------------------------------------------------------------------------

uint8 FDUB::SenceDeviceStatus()
{
	return 0x20 | (file ? 8 : 0) | (writeprotected ? 0x40 : 0) 
		| (track & ~1 ? 0 : 0x10) | (track & 1 ? 4 : 0);
}

// ---------------------------------------------------------------------------
//	FDUB::WriteID
// ---------------------------------------------------------------------------

FDC::Result FDUB::WriteID(uint8 flags, WIDDESC wid)
{
	int i;
	
	if (!file || !SetHead(flags))
		return Result(FDC::ST0_AT | FDC::ST1_MA);
	if (writeprotected)
		return Result(FDC::ST0_AT | FDC::ST1_NW);

	// gbNTCYvZ
	int sot = 0;
	wid.n = Min(7, wid.n);
	for (i=wid.sc-1; i>=0; i--)
	{
		int sos = sizeof(SectorHeader) + (0x80 << Min(8, wid.n));
		if (sot + sos > 0x1800)		// 2D
			break;
		sot += sos;
	}
	wid.idr += i+1;
	wid.sc -= i+1;

	if (wid.sc)
	{
		// ݂̃gbNɏނƂ͉\H
		// TCYȂȂt@C̍ŌɃgbNp
		int soct = GetTrackSize();
		if (soct < sot)
		{
			file->Seek(0, FileIO::end);
			header.trackptr[track] = file->Tellp();
		}

		// ZN^
		file->Seek(header.trackptr[track], FileIO::begin);
		
		SectorHeader sh;
		sh.density = (~flags) & 0x40;
		sh.deleted = 0;
		sh.sectors = wid.sc;
		sh.length = 0x80 << Min(8, wid.n);
		memset(sh.reserved, 0, 5);

		uint8 buf[0x80];
		memset(buf, wid.d, 0x80);

		for (i=0; i<wid.sc; i++)
		{
			sh.id = wid.idr[i];
			sh.status = (wid.idr[i].n != wid.n) ? 0xb0 : 0;
			file->Write(&sh, sizeof(SectorHeader));
			for (int j=0; j<(1 << wid.n); j++)
				file->Write(buf, 0x80);
		}
		headermodified = true;
		return Result(0);
	}
	else
	{
		// unformat
		header.trackptr[track] = 0;
		headermodified = true;
		return Result(0);
	}
}

// ---------------------------------------------------------------------------
//	gbÑTCY𓾂
// ---------------------------------------------------------------------------

int FDUB::GetTrackSize()
{
	idleft = 0;
	
	// unformated ?
	if (0 == header.trackptr[track])
		return 0;
	
	file->Seek(header.trackptr[track], FileIO::begin);
	
	SectorHeader sh;
	int sot = 0;
	int i = 0;
	do
	{
		file->Read(&sh, sizeof(SectorHeader));
		file->Seek(sh.length, FileIO::current);
		sot += sizeof(SectorHeader) + sh.length;
	} while (++i < sh.sectors);

	return sot;
}

// ---------------------------------------------------------------------------
//	FDUB::Unmount
// ---------------------------------------------------------------------------

bool FDUB::Unmount()
{
	bool result = true;

	if (file && !writeprotected)
	{
		// egbN͏Ԃɕł邩H
		bool needrelocation = false;
		uint ipos = sizeof(ImageHeader);
		
		for (track=0; track<164; track++)
		{
			int sot = GetTrackSize();

			if ((sot && header.trackptr[track] != ipos) 
				|| (!sot && header.trackptr[track]))
			{
				needrelocation = true;
				break;
			}
			ipos += sot;
		}
		if (needrelocation || ipos != header.disksize)
		{
			uint8* buf[164];
			memset(buf, 0, sizeof(buf));
			result = Relocate(buf);
			for (int i=0; i<164; i++)
				delete[] buf[i];
		}
		else
		{
			if (headermodified)
			{
				header.disksize = ipos;
				file->Seek(0, FileIO::begin);
				file->Write(&header, sizeof(ImageHeader));
			}
		}
	}
	file = 0; 
	return result; 
}

// ---------------------------------------------------------------------------
//	gbN̍Ĕzu
//	t@Cxč蒼
// ---------------------------------------------------------------------------

bool FDUB::Relocate(uint8** buf)
{
	int tr;
	int sot[164];

	for (tr=0; tr<164; tr++)
	{
		track = tr;
		sot[tr] = GetTrackSize();
		if (sot[tr])
		{
			buf[tr] = new uint8[sot[tr]];
			if (!buf[tr])
				return false;
			file->Seek(header.trackptr[tr], FileIO::begin);
			file->Read(buf[tr], sot[tr]);
		}
	}

	if (!file->Reopen(FileIO::create))
		return false;

	int tpos = sizeof(ImageHeader);
	file->Seek(tpos, FileIO::begin);
	
	for (tr=0; tr<164; tr++)
	{
		if (sot[tr])
		{
			header.trackptr[tr] = tpos;
			file->Write(buf[tr], sot[tr]);
			tpos += sot[tr];
		}
		else
		{
			header.trackptr[tr] = 0;
		}
	}
	header.disksize = tpos;

	file->Seek(0, FileIO::begin);
	file->Write(&header, sizeof(ImageHeader));
	return true;
}
