// ---------------------------------------------------------------------------
//	FM Sound Generator - Core Unit
//	Copyright (C) 1998 cisc.
// ---------------------------------------------------------------------------
//	$Id: FMGen.cpp,v 1.4 1999/01/31 01:37:38 cisc Exp $
//
//	Ql:
//	FM sound generator for M.A.M.E., written by Tatsuyuki Satoh.
//
//	:
//		SSG-Type Envelop
//
//	:
//		TvO[g̐ݒ̓O[o
//
//	history:
//	1998.08.20	asm ׂ̈ EG, PG, OP Z
//	1998.09.26	asm Ƃɔ namespace ̗̍p
//	1998.09.27  LFO Ή
//	1998.10.30	VC6 ̃oÖ x86 LFO R[hRpCłȂ̂
//				Ƃ肠܂Ă݂
//
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "misc.h"
#include "FMGen.h"

#ifdef RELEASEBUILD
	#define FM_USE_X86_CODE			// LFO  x86 ł̂ݗL
#endif

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
	#define FM_LOGGING
#endif

#ifdef FM_LOGGING
	static FILE* fp;

	static void LogBegin()   { if (!fp) fp = fopen("OPN.DMP", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }

	class Dummy
	{
	public:
		Dummy() { LogBegin(); }
		~Dummy() { LogEnd(); }
	};

	Dummy dum;

#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif

// ---------------------------------------------------------------------------
//	Global Table/etc
//
namespace FM
{
	const uint8 notetable[128] =
	{
		 0,  0,  0,  0,  0,  0,  0,  1,  2,  3,  3,  3,  3,  3,  3,  3, 
		 4,  4,  4,  4,  4,  4,  4,  5,  6,  7,  7,  7,  7,  7,  7,  7, 
		 8,  8,  8,  8,  8,  8,  8,  9, 10, 11, 11, 11, 11, 11, 11, 11, 
		12, 12, 12, 12, 12, 12, 12, 13, 14, 15, 15, 15, 15, 15, 15, 15, 
		16, 16, 16, 16, 16, 16, 16, 17, 18, 19, 19, 19, 19, 19, 19, 19, 
		20, 20, 20, 20, 20, 20, 20, 21, 22, 23, 23, 23, 23, 23, 23, 23, 
		24, 24, 24, 24, 24, 24, 24, 25, 26, 27, 27, 27, 27, 27, 27, 27, 
		28, 28, 28, 28, 28, 28, 28, 29, 30, 31, 31, 31, 31, 31, 31, 31, 
	};

	uint32 tltable[FM_TLPOS + FM_TLENTS];
	uint32 dltable[3][FM_EGENTS];
	int32 sinetable[FM_OPSINENTS];
	uint32 pmtable[8][FM_LFOENTS];
	uint32 amtable[4][FM_LFOENTS];
	static bool tablemade = false;
	uint32 lfotable[8];
}

#include "FMGenInl.h"

namespace FM
{

// ---------------------------------------------------------------------------
//	e[u쐬
//
void MakeTable()
{
	if (!tablemade)
	{
		int i;
		Operator::MakeTable();
		tablemade = true;
		for (i=-FM_TLPOS; i<0; i++)
			tltable[FM_TLPOS + i] = uint(65535.0 * pow(10.0, i * -96. / FM_TLENTS / 20.));

		for (i=0; i<FM_TLENTS; i++)
			tltable[FM_TLPOS + i] = uint(65535.0 * pow(10.0, i * -96. / FM_TLENTS / 20.));
	}

	static const double pms[8] = { 0, 3.4, 6.7, 10, 14, 20, 40, 80 };
	static const double ams[4] = { 0, 1.4, 5.9, 11.8 };

	int i;
	for (i=0; i<8; i++)
	{
		double pmb = pms[i] / 2400.0;
		for (int j=0; j<FM_LFOENTS; j++)
		{
			pmtable[i][j] =
				uint32(0x10000 * pow(2.0, pmb * sin(2 * FM_PI * j / FM_LFOENTS)));
//			Log3("PM[%d][%3d] = %8x\n", i, j, pmtable[i][j]);
		}
	}
	for (i=0; i<4; i++)
	{
		double amb = -ams[i] / 40.0;
		for (int j=0; j<FM_LFOENTS; j++)
		{
			amtable[i][j] =
				uint32(0x10000 * pow(10.0, amb * (1 + sin(2 * FM_PI * j / FM_LFOENTS))));
//			Log3("AM[%d][%3d] = %8x\n", i, j, amtable[i][j]);
		}
	}
}

//	`bṽTvO[gƐ鉹̃TvO[gƂ̔ݒ
void MakeTimeTable(uint ratio)
{
	Operator::MakeTimeTable(ratio);

	static const int table[8] =
	{
		109,  78,  72,  68,  63,  45,  9,  6
	};

	for (int i=0; i<8; i++)
	{
		lfotable[i] = uint(78.0 / 10000.0 / table[i] * ratio * (1 << (FM_LFOBITS + FM_LFOCBITS - FM_RATIOBITS))); 
		Log3("lfotable[%d] = %8x(%f)\n", i, lfotable[i], lfotable[i]*1.0/(1<<(FM_LFOBITS + FM_LFOCBITS)) );
	}
}

// ---------------------------------------------------------------------------
//	Envelope Generator
//
//  n = .75dB step
#define	EGC(n)				(n << (FM_EGCBITS + FM_EGBITS - 7))
#define EGINDEX(n)			((n) << (FM_EGCBITS + FM_EGBITS))

void Operator::Prepare()
{
	// EG Part
	tllinear = tltable[FM_TLPOS + (tl << (FM_TLBITS-7))];

	ksr = bn >> (3-ks);
	
	switch (phase)
	{
	case attack:
		egdcount = ar ? attacktable[Min(63, ar+ksr)] : 0;
		break;
	case decay:
		egdcount = dr ? decaytable [Min(63, dr+ksr)] : 0;
		eglimit = EGINDEX(dlnormal) + EGC(sl);
		break;
	case sustain:
		egdcount = sr ? decaytable [Min(63, sr+ksr)] : 0;
		break;
	case release:
		egdcount = decaytable [Min(63, rr+ksr)];
		break;
	}

	//	PG Part
	pgdcount = (dp + dttable[detune + bn]) * multable[multiple];
}

//	Envelop  phase ύX
void Operator::ShiftPhase(Phase nextphase)
{
	if (nextphase == next)
		nextphase = Phase(phase + 1);
		
	switch (nextphase)
	{
	case attack:		// Attack Phase
		tl = tll;
		tllinear = tltable[FM_TLPOS + (tl << (FM_TLBITS-7))];
		if ((ar+ksr) < 62)
		{
			if (phase == release)
				egcount = d2atable[(egcount >> (FM_EGCBITS+FM_EGBITS-FM_EGSBITS)) & (FM_EGSENTS-1)];
			else
				egcount = EGINDEX(dlattack);
	
			eglimit = EGINDEX(dlattack) + EGC(128);
			egdcount = ar ? attacktable[Min(63, ar+ksr)] : 0;
			phase = attack;
			Log3("ar=%2d  dcount=%8d (%8d)\n", ar+ksr, egdcount, dltable[(egcount >> FM_EGCBITS) & (FM_EGENTS-1)]);
			break;
		}
	case decay:			// Decay Phase
		if (sl)
		{
			egcount = EGINDEX(dlnormal);
			eglimit = EGINDEX(dlnormal) + EGC(sl);
			egdcount = dr ? decaytable[Min(63, dr+ksr)] : 0;
			phase = decay;
			break;
		}	
	case sustain:		// Sustain Phase
		egcount = EGINDEX(dlnormal) + EGC(sl);
		eglimit = EGINDEX(dlnormal) + EGC(128);
		egdcount = sr ? decaytable[Min(63, sr+ksr)] : 0;
		phase = sustain;
		break;
	
	case release:		// Release Phase
		if (phase == attack || (egcount < EGC(128) && phase != off))
		{
			if (phase == attack)
				egcount = a2dtable[(egcount >> (FM_EGCBITS+FM_EGBITS-FM_EGSBITS)) & (FM_EGSENTS-1)];
			egdcount = rr ? decaytable[Min(63, rr+ksr)] : 0;
			eglimit = EGINDEX(dlnormal) + EGC(128);
			phase = release;
			break;
		}
	case off:			// off
	default:
		egcount = EGINDEX(dlnormal) + EGC(128) - 1;
		eglimit = EGINDEX(dlnormal) + EGC(128);
		egdcount = 0;
		phase = off;
		break;
	}
}

#define DBtoL(db)	(Counter)(65535.0 * pow(10.0, (db) / 20.0))
#define ToTLE(db)	 Counter((db) * (FM_TLENTS - 1))


//	e[u쐬
void Operator::MakeTable()
{
	//	EG Part
	int i;
	const double c = 0.0125;			// K(^^;
	const double logcinv = 1.0 / log(c);
	const double m  = 1.0 / (1.0 - c);
	const double minv = 1.0 - c;
	
	// make dB->linear conversion table
	for (i=0; i<FM_EGENTS; i++)
	{
		double x = double(i) / FM_EGENTS;
		dltable[dlnormal][i] = DBtoL(-96. * x);
		dltable[dlattack][i] = DBtoL(-96. * (1 - m * (1 - pow(c, x))));
//		Log2("%4d  a = %8d\n", i, dltable[dlattack][i]);
	}

	for (i=0; i<FM_EGSENTS; i++)
	{
		double x, y;
		
		x = double(i) / FM_EGSENTS;
		y = 1.0 - m * (1 - pow(c, x));
		a2dtable[i] = EGINDEX(dlnormal  )     + Counter(((1 << (FM_EGCBITS + FM_EGBITS))-1) * y);
		y = 1.0 - log(1.0 - (1.0 - x) * minv) * logcinv;
		d2atable[i] = EGINDEX(dlattack+1) - 1 - Counter(((1 << (FM_EGCBITS + FM_EGBITS))  ) * y);

		Log3("%4d  a2d=%8x  d2a=%8x\n", i, a2dtable[i], d2atable[i]);
	}

	//	PG Part
	static const uint8 table[128] =
	{
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, //FD=0
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
		 0,  0,  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2, //FD=1
		 2,  3,  3,  3,  4,  4,  4,  5,  5,  6,  6,  7,  8,  8,  8,  8,
		 1,  1,  1,  1,  2,  2,  2,  2,  2,  3,  3,  3,  4,  4,  4,  5, //FD=2
		 5,  6,  6,  7,  8,  8,  9, 10, 11, 12, 13, 14, 16, 16, 16, 16,
		 2,  2,  2,  2,  2,  3,  3,  3,  4,  4,  4,  5,  5,  6,  6,  7, //FD=3
		 8,  8,  9, 10, 11, 12, 13, 14, 16, 17, 19, 20, 22, 22, 22, 22,
	};
	
	for (i=0; i<128; i++)
	{
		int val = table[i] << 1;
		dttable[i    ] =  val;
		dttable[i+128] = -val;
	}

	//	OP Part
	for (i=0; i<FM_OPSINENTS; i++)
	{
		double s = sin(2 * FM_PI * i / FM_OPSINENTS);
		sinetable[i] = int32((1 << FM_ISHIFT) * s);
	}
}

//	NbNETvO[gɈˑe[u쐬
void Operator::MakeTimeTable(uint ratio)
{
	int i;

	//	EG Part
	static const int16 table[16] =
	{
		    0,     2,     4,     8,    16,    32,    64,   128,
		  256,   512,  1024,  2048,  4096,  7475, 14238, 29900
	};
	
	double ra = double(ratio) / (1 << FM_RATIOBITS) 
			  * (1 << FM_EGCBITS) * FM_EGENTS / FM_ARRATE;
	double rd = double(ratio) / (1 << FM_RATIOBITS) 
			  * (1 << FM_EGCBITS) * FM_EGENTS / FM_DRRATE;
	
	Counter* pa = attacktable;
	Counter* pd = decaytable;
	
	for (i=0; i<16; i++)
	{
		for (int j=4; j<8; j++)
		{
			*pa++ = Counter(ra * (table[i] * j / 4));
			*pd++ = Counter(rd * Min(32767, i ? (1 << i) * j / 4 : 0));
			Log4("%2d-%d  %10d  %10d\n", i, j-4, pa[-1], pd[-1]);
		}
	}
	attacktable[62] = attacktable[63] = EGC(128);

	// PG Part
	for (i=0; i<16; i++)
	{
		int mul = i ? i * 2 : 1;
#if ((2 + FM_RATIOBITS) > FM_PGBITS)
		multable[i] = (mul * ratio) >> (2 + FM_RATIOBITS - FM_PGBITS);
#else
		multable[i] = (mul * ratio) << (FM_PGBITS - FM_RATIOBITS - 2));
#endif
	}
}

//	Block/F-Num
void Operator::SetFNum(uint f)
{
	dp = (f & 2047) << ((f >> 11) & 7);
	bn = notetable[(f >> 7) & 127];
}

//	static tables
//uint32 Operator::dltable[3][FM_EGENTS] = { 0, };
uint32 Operator::multable[16];
int32  Operator::dttable[256];
Operator::Counter Operator::attacktable[64];
Operator::Counter Operator::decaytable[64];
Operator::Counter Operator::a2dtable[FM_EGSENTS];
Operator::Counter Operator::d2atable[FM_EGSENTS];


// ---------------------------------------------------------------------------
//	4-op Channel
//

const uint8 Channel4::fbtable[8] = { 31, 7, 6, 5, 4, 3, 2, 1 };

// Zbg
void Channel4::Reset()
{
	op[0].Reset();
	op[1].Reset();
	op[2].Reset();
	op[3].Reset();
}

//	Calc ̗p
int Channel4::Prepare()
{
	op[0].Prepare();
	op[1].Prepare();
	op[2].Prepare();
	op[3].Prepare();
	return op[0].IsOn() | op[1].IsOn()
		 | op[2].IsOn() | op[3].IsOn();
}

//	F-Number/BLOCK ݒ
void Channel4::SetFNum(uint f)
{
	uint dp = (f & 2047) << ((f >> 11) & 7);
	uint bn = notetable[(f >> 7) & 127];
	op[0].SetDPBN(dp, bn);
	op[1].SetDPBN(dp, bn);
	op[2].SetDPBN(dp, bn);
	op[3].SetDPBN(dp, bn);
}

//	L[
void Channel4::KeyControl(uint key)
{
	if (key & 0x10) op[0].KeyOn(); else op[0].KeyOff();
	if (key & 0x20) op[1].KeyOn(); else op[1].KeyOff();
	if (key & 0x40) op[2].KeyOn(); else op[2].KeyOff();
	if (key & 0x80) op[3].KeyOn(); else op[3].KeyOff();
}

//	ASYݒ
void Channel4::SetAlgorithm(uint algo)
{
	static const uint8 table1[8][6] = 
	{
		{ 0, 1, 1, 2, 2, 3 },	{ 1, 0, 0, 1, 1, 2 },
		{ 1, 1, 1, 0, 0, 2 },	{ 0, 1, 2, 1, 1, 2 },
		{ 0, 1, 2, 2, 2, 1 },	{ 0, 1, 0, 1, 0, 1 },
		{ 0, 1, 2, 1, 2, 1 },	{ 1, 0, 1, 0, 1, 0 },
	};
	
	in [0] = &buf[table1[algo][0]];
	out[0] = &buf[table1[algo][1]];
	in [1] = &buf[table1[algo][2]];
	out[1] = &buf[table1[algo][3]];
	in [2] = &buf[table1[algo][4]];
	out[2] = &buf[table1[algo][5]];
	op[0].ResetFB();
}

#ifndef FM_USE_X86_CODE

//  
ISample Channel4::Calc()
{
	buf[1] = buf[2] = buf[3] = 0;
	
	buf[0] = op[0].CalcFB(fb);
	*out[0] += op[1].Calc(*in[0]);
	*out[1] += op[2].Calc(*in[1]);
	return *out[2] + op[3].Calc(*in[2]);
}

ISample Channel4::Calc2(uint)
{
	buf[1] = buf[2] = buf[3] = 0;
	
	buf[0] = op[0].CalcFB(fb);
	*out[0] += op[1].Calc(*in[0]);
	*out[1] += op[2].Calc(*in[1]);
	return *out[2] + op[3].Calc(*in[2]);
}

#else

#define THIS	[esi]this
#define OP(n)	[esi]this.op[(n) * TYPE Operator]
#define BUF(n)	[esi]this.buf[(n) * TYPE int]
#define OPD		[edi]op

// -------------------------------------------------------- without LFO

#if 1

#define OPCALC(n)	\
	__asm { lea edi,OP(n-1) } \
	__asm {	mov	ebx,OPD.egcount } \
	__asm {	mov	ecx,OPD.egdcount } \
	__asm { mov edx,OPD.eglimit } \
	__asm {	add	ebx,ecx } \
	__asm {	cmp	ebx,edx } \
	__asm {	mov	OPD.egcount,ebx } \
	__asm { jnb	op##n##_phase } \
	op##n##_1: \
	__asm {	shr ebx,FM_EGCBITS } \
	__asm {	mov	edx,OPD.pgcount } \
	__asm {	mov	eax,OPD.pgdcount } \
	__asm {	mov	ecx,dltable[ebx * 4] } \
	__asm {	mov	ebx,THIS.in[(n-2) * 4] } \
	__asm { add edx,eax } \
	__asm {	mov	eax,[ebx] } \
	__asm {	imul ecx,OPD.tllinear } \
	__asm {	sal	eax,2 + IS2EC_SHIFT } \
	__asm {	mov OPD.pgcount,edx } \
	__asm {	add	edx,eax } \
	__asm { shr ecx,16 } \
	__asm {	shr	edx,20+FM_PGBITS-FM_OPSINBITS-2 } \
	__asm {	mov	ebx,THIS.out[(n-2) * 4] } \
	__asm {	and	edx,(FM_OPSINENTS-1)*4 } \
	__asm { imul ecx,sinetable[edx] }

ISample Channel4::Calc()
{
	__asm
	{
		xor		eax,eax
		mov		esi,ecx						// = this
		mov		BUF(1),eax
		mov		edx,OP(0).egcount			// egcount += egdcount
		mov		ecx,OP(0).egdcount
		mov		BUF(2),eax
		add		edx,ecx
		mov		ebx,OP(0).eglimit			// egcount < eglimit ?
		mov		BUF(3),eax
		cmp		edx,ebx
		mov		OP(0).egcount,edx
		jnb		op1_phase
op1_1:
		shr		edx,FM_EGCBITS
		mov		eax,OP(0).pgdcount
		mov		edi,OP(0).pgcount
		mov		ebx,dltable[edx * 4]
		mov		ecx,THIS.fb
		add		eax,edi
		imul	ebx,OP(0).tllinear
		shr		ebx,16
		cmp		ecx,31
		mov		OP(0).pgcount,eax
		
		jne		op1_fb

		shr		eax,20+FM_PGBITS-FM_OPSINBITS-2
		and		eax,(FM_OPSINENTS-1)*4
		imul	ebx,sinetable[eax]
		mov		BUF(0),ebx
		mov		OP(0).out,ebx
		jmp		op2

op1_fb:
		// ebx = A   eax = 
		mov		edx,OP(0).out
		lea		edi,[edx*2+edx]				// edi = last out
		sal		edx,2 + IS2EC_SHIFT
		sar		edx,cl
		add		eax,edx						// eax = +FB
		sar		eax,20+FM_PGBITS-FM_OPSINBITS-2
		and		eax,(FM_OPSINENTS-1)*4
		imul	ebx,sinetable[eax]
		lea		edx,[ebx*4+ebx]
		mov		OP(0).out,ebx
		add		edx,edi
		sar		edx,3
		mov		BUF(0),edx
	}
op2:
	OPCALC(2)
	__asm {	add	[ebx],ecx }

	OPCALC(3)
	__asm {	add	[ebx],ecx }
	
	OPCALC(4)
	__asm
	{
		pop	edi			//  pop Q̓RpC̏󋵂ɂĂ͏Kv邩
		mov eax,[ebx]
		pop	esi
		add eax,ecx
		pop	ebx
		pop ecx
		pop	ebp
		ret 0
	}

op1_phase:
	op[0].ShiftPhase(Operator::next);
	__asm mov	edx,OP(0).egcount
	__asm jmp	op1_1
op2_phase:
	op[1].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(1).egcount
	__asm jmp	op2_1
op3_phase:
	op[2].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(2).egcount
	__asm jmp	op3_1
op4_phase:
	op[3].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(3).egcount
	__asm jmp	op4_1

	return 0;
}

#endif

#undef OPCALC


// -------------------------------------------------------- with LFO

#define OPCALC(n)	\
	__asm { lea edi,OP(n-1) } \
	__asm {	mov	ebx,OPD.egcount } \
	__asm { mov edx,OPD.eglimit } \
	__asm {	add	ebx,OPD.egdcount } \
	__asm {	cmp	ebx,edx } \
	__asm {	mov	OPD.egcount,ebx } \
	__asm { jnb	op##n##_phase } \
	op##n##_1: \
	__asm {	shr ebx,FM_EGCBITS } \
	__asm {	mov	eax,dltable[ebx * 4] } \
	__asm {	imul eax,OPD.tllinear } \
	__asm { test OPD.amon,-1 } \
	__asm { jz op##n##_noamon } \
	__asm { mul THIS.amc } \
	__asm {	mov	eax,OPD.pgdcount } \
	__asm { mov ecx,edx } \
	__asm { jmp op##n##_2 } \
	op##n##_noamon: \
	__asm { mov ecx,eax } \
	__asm {	mov	eax,OPD.pgdcount } \
	__asm { shr ecx,16 } \
	op##n##_2: \
	__asm { mul THIS.pmc } \
	__asm {	mov	ebx,OPD.pgcount } \
	__asm { shrd eax,edx,16 } \
	__asm {	mov	edx,THIS.in[(n-2) * 4] } \
	__asm { add ebx,eax } \
	__asm {	mov	eax,[edx] } \
	__asm {	mov OPD.pgcount,ebx } \
	__asm {	sal	eax,2 + IS2EC_SHIFT } \
	__asm {	add	ebx,eax } \
	__asm {	shr	ebx,20+FM_PGBITS-FM_OPSINBITS-2 } \
	__asm {	and	ebx,(FM_OPSINENTS-1)*4 } \
	__asm { imul ecx,sinetable[ebx] } \
	__asm {	mov	ebx,THIS.out[(n-2) * 4] }

ISample Channel4::Calc2(uint lfo)
{
	__asm
	{
		xor		edx,edx
		mov		esi,ecx						// = this
		mov		BUF(1),edx
		mov		eax,lfo
		mov		ebx,THIS.ams
		mov		BUF(2),edx
		mov		ecx,[ebx+eax*4]
		mov		BUF(3),edx
		mov		ebx,THIS.pms
		mov		edx,[ebx+eax*4]
		mov		THIS.amc,ecx
		mov		ebx,OP(0).egcount			// egcount += egdcount
		mov		ecx,OP(0).egdcount
		mov		THIS.pmc,edx
		add		ebx,ecx
		mov		edx,OP(0).eglimit			// egcount < eglimit ?
		mov		OP(0).egcount,ebx
		cmp		ebx,edx
		jnb		op1_phase
op1_1:
		// ebx = egcount
		shr		ebx,FM_EGCBITS				// eax = tllinear * dltable[egcount]
		mov		eax,dltable[ebx * 4]
		mov		cl,OP(0).amon
		imul	eax,OP(0).tllinear			// ebx = 32bit
		test	cl,cl
		jz		op1_namon

		mul		THIS.amc
		mov		ebx,edx
		jmp		op1_2
op1_namon:
		shr		eax,16
		mov		ebx,eax
op1_2:
		// ebx = egout
		mov		eax,OP(0).pgdcount
		mul		THIS.pmc
		mov		edi,OP(0).pgcount			// pgcount += pgdcount
		shrd	eax,edx,16
		mov		ecx,THIS.fb
		add		eax,edi
		mov		OP(0).pgcount,eax
		cmp		ecx,31

		jne		op1_fb

		shr		eax,20+FM_PGBITS-FM_OPSINBITS-2
		and		eax,(FM_OPSINENTS-1)*4
		imul	ebx,sinetable[eax]
		mov		BUF(0),ebx
		mov		OP(0).out,ebx
		jmp		op2

op1_fb:
		// ebx = A   eax = 
		mov		edx,OP(0).out
		lea		edi,[edx*2+edx]				// edi = last out
		sal		edx,2 + IS2EC_SHIFT
		sar		edx,cl
		add		eax,edx						// eax = +FB
		sar		eax,20+FM_PGBITS-FM_OPSINBITS-2
		and		eax,(FM_OPSINENTS-1)*4
		imul	ebx,sinetable[eax]
		lea		edx,[ebx*4+ebx]
		mov		OP(0).out,ebx
		add		edx,edi
		sar		edx,3
		mov		BUF(0),edx
	}
op2:
	OPCALC(2)
	__asm {	add	[ebx],ecx }

	OPCALC(3)
	__asm {	add	[ebx],ecx }
	
	OPCALC(4)
	__asm
	{
		pop	edi			//  pop Q̓RpC̏󋵂ɂĂ͏Kv邩
		mov eax,[ebx]
		pop	esi
		add eax,ecx
		pop	ebx
		pop ecx
		pop	ebp
		ret	4
	}

op1_phase:
	op[0].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(0).egcount
	__asm jmp	op1_1
op2_phase:
	op[1].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(1).egcount
	__asm jmp	op2_1
op3_phase:
	op[2].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(2).egcount
	__asm jmp	op3_1
op4_phase:
	op[3].ShiftPhase(Operator::next);
	__asm mov	ebx,OP(3).egcount
	__asm jmp	op4_1

	return 0;
}

#undef OPCALC

#endif // FM_USE_X86_CODE


}	// namespace FM
