// ---------------------------------------------------------------------------
//  M88 - PC88 emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	Config Dialog Box for M88
//	$Id: M88Config2.cpp,v 1.3 1999/01/15 12:58:57 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#define _WIN32_IE		0x200

#include <windows.h>
#include <stdlib.h>
#include <commctrl.h>
#include "resource.h"
#include "M88Config2.h"
#include "misc.h"
#include "MSGs.h"

static const char* AppName = "M88 for Windows";

#define BSTATE(b) (b ? BST_CHECKED : BST_UNCHECKED)

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

M88Config2::M88Config2()
{
	propproc.SetDestination(PropProcGate, this);
	cpupage.SetDestination(CPUPageGate, this);
	screenpage.SetDestination(ScreenPageGate, this);
	soundpage.SetDestination(SoundPageGate, this);
	soundvolpage.SetDestination(SoundVolumePageGate, this);
	functionpage.SetDestination(FunctionPageGate, this);
	switchpage.SetDestination(SwitchPageGate, this);

	page = 0;
}

M88Config2::~M88Config2()
{

}

// ---------------------------------------------------------------------------
//	ݒ_CAO̎s
//	ݒ̔f̓bZ[Wōs\
// ---------------------------------------------------------------------------

bool M88Config2::Show(HINSTANCE hinstance, HWND hwnd, Config* conf)
{
	orgconfig = config = *conf;
	hinst = hinstance;
	hwndparent = hwnd;

	PROPSHEETPAGE psp[term];
	PROPSHEETHEADER psh;
	
	FillInPropPage(&psp[cpu     ], IDD_CONFIG_CPU, cpupage);
	FillInPropPage(&psp[screen  ], IDD_CONFIG_SCREEN, screenpage);
	FillInPropPage(&psp[sound   ], IDD_CONFIG_SOUND, soundpage);
	FillInPropPage(&psp[soundvol], IDD_CONFIG_SOUNDVOL, soundvolpage);
	FillInPropPage(&psp[function], IDD_CONFIG_FUNCTION, functionpage);
	FillInPropPage(&psp[switches], IDD_CONFIG_SWITCHES, switchpage);

	memset(&psh, 0, sizeof(psh));
	psh.dwSize = sizeof(psh);
	psh.dwFlags = PSH_PROPSHEETPAGE | PSH_USECALLBACK; 
	psh.hwndParent = hwndparent;
	psh.hInstance = hinst;
	psh.pszCaption = "ݒ";
	psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
	psh.nStartPage = page;
	psh.ppsp = psp;
	psh.pfnCallback = (PFNPROPSHEETCALLBACK) (void*) propproc;

	PropertySheet(&psh);
	return false;
}

// ---------------------------------------------------------------------------
//	PropSheetProc
//
int M88Config2::PropProc(HWND hwnd, UINT m, LPARAM)
{
	if (m == PSCB_INITIALIZED)
	{
		hwndps = hwnd;
	}
	return 0;
}

int _cdecl M88Config2::PropProcGate
(M88Config2* config, HWND hwnd, UINT m, LPARAM l)
{
	return config->PropProc(hwnd, m, l);
}

// ---------------------------------------------------------------------------
//	Fill in psp
// ---------------------------------------------------------------------------

void M88Config2::FillInPropPage(PROPSHEETPAGE * psp, int iddlg, void* dlgproc)
{
	memset(psp, 0, sizeof(PROPSHEETPAGE));
	psp->dwSize = sizeof(PROPSHEETPAGE);
	psp->dwFlags = 0;
	psp->hInstance = hinst;
	psp->pszTemplate = MAKEINTRESOURCE(iddlg);
	psp->pszIcon = 0;
	psp->pfnDlgProc = (DLGPROC) dlgproc;
	psp->lParam = 0;
}

// ---------------------------------------------------------------------------
//	CPU Page
// ---------------------------------------------------------------------------

BOOL M88Config2::CPUPage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		config.clock = orgconfig.clock;
		config.speed = orgconfig.speed;
		SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_SETLINESIZE, 0, 1);
		SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_SETPAGESIZE, 0, 2);
		SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_SETRANGE, TRUE, MAKELONG(2, 20));
		SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_SETPOS, FALSE, config.speed / 100);
		break;

	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
			SetFocus(GetDlgItem(hdlg, config.flags & Config::fullspeed ? IDC_CPU_NOSUBCPUCONTROL : IDC_CPU_CLOCK));
			SendDlgItemMessage(hdlg, IDC_CPU_CLOCK_SPIN, UDM_SETRANGE, 0, MAKELONG(100, 1));
			SendDlgItemMessage(hdlg, IDC_CPU_CLOCK, EM_SETLIMITTEXT, 3, 0);
			CPUPageUpdate(hdlg);
			SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_SETRANGE, TRUE, MAKELONG(2, 20));
			page = cpu;
			break;

		case PSN_APPLY:
			orgconfig = config;
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;
		}
		break;

	case WM_COMMAND:
		return CPUPageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	case WM_HSCROLL:
		CPUPageUpdateSlider(hdlg);
		PropSheet_Changed( hwndps, hdlg );
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::CPUPageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		switch (id)
		{
		case IDC_CPU_NOWAIT:
			config.flags ^= Config::fullspeed;
			if (config.flags & Config::fullspeed)
				config.flags &= ~Config::cpuburst;
			CPUPageUpdate(hdlg);
			PropSheet_Changed( hwndps, hdlg );
			return TRUE;

		case IDC_CPU_BURST:
			config.flags ^= Config::cpuburst;
			if (config.flags & Config::cpuburst)
				config.flags &= ~Config::fullspeed;
			CPUPageUpdate(hdlg);
			PropSheet_Changed( hwndps, hdlg );
			return TRUE;

		case IDC_CPU_CLOCKMODE:
			config.flags ^= Config::cpuclockmode;
			PropSheet_Changed( hwndps, hdlg );
			return TRUE;

		case IDC_CPU_NOSUBCPUCONTROL:
			config.flags ^= Config::subcpucontrol;
			PropSheet_Changed( hwndps, hdlg );
			return TRUE;
		}
	}
	switch (id)
	{
	case IDC_CPU_CLOCK:
		if (nc == EN_CHANGE)
		{
			int clock = Limit(GetDlgItemInt(hdlg, IDC_CPU_CLOCK, 0, false), 100, 1); 
			if (clock != config.clock)
				PropSheet_Changed( hwndps, hdlg );
			config.clock = clock;
			return TRUE;
		}
		break;
	}
	return FALSE;
}

void M88Config2::CPUPageUpdate(HWND hdlg)
{
	SetDlgItemInt(hdlg, IDC_CPU_CLOCK, config.clock, false);
	CheckDlgButton(hdlg, IDC_CPU_NOWAIT, BSTATE(config.flags & Config::fullspeed));
	
	EnableWindow(GetDlgItem(hdlg, IDC_CPU_CLOCK), !(config.flags & Config::fullspeed));
	
	EnableWindow(GetDlgItem(hdlg, IDC_CPU_SPEED), !(config.flags & Config::cpuburst));
	EnableWindow(GetDlgItem(hdlg, IDC_CPU_SPEED_TEXT), !(config.flags & Config::cpuburst));
	
	CheckDlgButton(hdlg, IDC_CPU_NOSUBCPUCONTROL, BSTATE(!(config.flags & Config::subcpucontrol)));
	CheckDlgButton(hdlg, IDC_CPU_CLOCKMODE, BSTATE(config.flags & Config::cpuclockmode));
	CheckDlgButton(hdlg, IDC_CPU_BURST, BSTATE(config.flags & Config::cpuburst));
	CPUPageUpdateSlider(hdlg);
}

void M88Config2::CPUPageUpdateSlider(HWND hdlg)
{
	char buf[8];
	config.speed = SendDlgItemMessage(hdlg, IDC_CPU_SPEED, TBM_GETPOS, 0, 0) * 100;
	wsprintf(buf, "%d%%", config.speed/10);
	SetDlgItemText(hdlg, IDC_CPU_SPEED_TEXT, buf);
}

BOOL _cdecl M88Config2::CPUPageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->CPUPage(hwnd, m, w, l);
}


// ---------------------------------------------------------------------------
//	Screen Page
// ---------------------------------------------------------------------------

BOOL M88Config2::ScreenPage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
			ScreenPageUpdate(hdlg);
			page = screen;
			break;

		case PSN_APPLY:
			orgconfig = config;
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;
		}
		break;

	case WM_COMMAND:
		return ScreenPageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::ScreenPageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		PropSheet_Changed( hwndps, hdlg );
		switch (id)
		{
		case IDC_SCREEN_REFRESH_1:
			config.refreshtiming = 1; 
			return 0;

		case IDC_SCREEN_REFRESH_2:
			config.refreshtiming = 2; 
			return 0;
		
		case IDC_SCREEN_REFRESH_3:
			config.refreshtiming = 3; 
			return 0;
		
		case IDC_SCREEN_REFRESH_4:
			config.refreshtiming = 4; 
			return 0;

		case IDC_SCREEN_ENABLEPCG:
			config.flags ^= Config::enablepcg;
			return 0;

		case IDC_SCREEN_FV15K:
			config.flags ^= Config::fv15k;
			return 0;

		case IDC_SCREEN_DIGITALPAL:
			config.flags ^= Config::digitalpalette;
			return 0;

		case IDC_SCREEN_FORCE480:
			config.flags ^= Config::force480;
			return 0;

		case IDC_SCREEN_LOWPRIORITY:
			config.flags ^= Config::drawprioritylow;
			return 0;
		}
	}
	return FALSE;
}

void M88Config2::ScreenPageUpdate(HWND hdlg)
{
	static const int item[4] = 
	{ IDC_SCREEN_REFRESH_1, IDC_SCREEN_REFRESH_2, IDC_SCREEN_REFRESH_3, IDC_SCREEN_REFRESH_4 }; 
	CheckDlgButton(hdlg, item[config.refreshtiming-1], BSTATE(true));

	// misc. option
	CheckDlgButton(hdlg, IDC_SCREEN_ENABLEPCG, BSTATE(config.flags & Config::enablepcg));
	CheckDlgButton(hdlg, IDC_SCREEN_FV15K, BSTATE(config.flags & Config::fv15k));
	CheckDlgButton(hdlg, IDC_SCREEN_DIGITALPAL, BSTATE(config.flags & Config::digitalpalette));
	CheckDlgButton(hdlg, IDC_SCREEN_FORCE480, BSTATE(config.flags & Config::force480));
	CheckDlgButton(hdlg, IDC_SCREEN_LOWPRIORITY, BSTATE(config.flags & Config::drawprioritylow));
}

BOOL _cdecl M88Config2::ScreenPageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->ScreenPage(hwnd, m, w, l);
}

// ---------------------------------------------------------------------------
//	Sound Page
// ---------------------------------------------------------------------------

BOOL M88Config2::SoundPage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		config.soundbuffer = orgconfig.soundbuffer;
		CheckDlgButton(hdlg, config.flags & Config::enableopna ? IDC_SOUND44_OPNA : IDC_SOUND44_OPN, BSTATE(true));
		CheckDlgButton(hdlg, 
			config.flags & Config::opnaona8 ? IDC_SOUNDA8_OPNA : 
				(config.flags & Config::opnona8 ? IDC_SOUNDA8_OPN : IDC_SOUNDA8_NONE), 
					   BSTATE(true));
		break;

	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
			{
				UDACCEL ud[2] = { { 0, 25 }, { 1, 100 } };
				SendDlgItemMessage(hdlg, IDC_SOUND_BUFFERSPIN, UDM_SETRANGE, 0, MAKELONG(1000, 100));
				SendDlgItemMessage(hdlg, IDC_SOUND_BUFFERSPIN, UDM_SETACCEL, 2, (LPARAM) ud);
				SendDlgItemMessage(hdlg, IDC_SOUND_BUFFER, EM_SETLIMITTEXT, 4, 0);
				SoundPageUpdate(hdlg);
				page = sound;		// ݑIĂy[W
			}
			break;

		case PSN_APPLY:
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;
		}
		break;

	case WM_COMMAND:
		return SoundPageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::SoundPageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		PropSheet_Changed( hwndps, hdlg );
		switch (id)
		{
		case IDC_SOUND44_OPN:
			config.flags &= ~Config::enableopna;
			return 0;

		case IDC_SOUND44_OPNA:
			config.flags |= Config::enableopna;
			return 0;

		case IDC_SOUNDA8_OPN:
			config.flags = (config.flags & ~Config::opnaona8) | Config::opnona8;
			return 0;

		case IDC_SOUNDA8_OPNA:
			config.flags = (config.flags & ~Config::opnona8) | Config::opnaona8;
			return 0;

		case IDC_SOUNDA8_NONE:
			config.flags = config.flags & ~(Config::opnaona8 | Config::opnona8);
			return 0;

		case IDC_SOUNDA8_INTR:
			config.flags ^= Config::opna8intr;
			return 0;

		case IDC_SOUND_CMDSING:
			config.flags ^= Config::disablesing;
			return 0;

		case IDC_SOUND_NOSOUND:
			config.sound = 0; 
			return 0;

		case IDC_SOUND_11K:
			config.sound = 1; 
			return 0;

		case IDC_SOUND_22K:
			config.sound = 2; 
			return 0;

		case IDC_SOUND_44K:
			config.sound = 3; 
			return 0;
		
		case IDC_SOUND_55K:
			config.sound = 4; 
			return 0;
		}
	}
	switch (id)
	{
	case IDC_SOUND_BUFFER:
		if (nc == EN_CHANGE)
		{
			uint buf;
			buf = Limit(GetDlgItemInt(hdlg, IDC_SOUND_BUFFER, 0, false), 1000, 100) / 25 * 25; 
			if (buf != config.soundbuffer)
				PropSheet_Changed( hwndps, hdlg );
			config.soundbuffer = buf;
			return TRUE;
		}
		break;
	}
	return FALSE;
}

void M88Config2::SoundPageUpdate(HWND hdlg)
{
	static const int itemsr[5] =
	{ IDC_SOUND_NOSOUND, IDC_SOUND_11K, IDC_SOUND_22K, IDC_SOUND_44K, IDC_SOUND_55K, }; 

	CheckDlgButton(hdlg, itemsr[Min(config.sound, 4)], BSTATE(true));

	EnableWindow(GetDlgItem(hdlg, IDC_SOUNDA8_INTR), (config.flags & (Config::opnaona8 | Config::opnona8)));
	CheckDlgButton(hdlg, IDC_SOUNDA8_INTR, BSTATE(config.flags & Config::opna8intr));
	
	CheckDlgButton(hdlg, IDC_SOUND_CMDSING, BSTATE(!(config.flags & Config::disablesing)));
	
	SetDlgItemInt(hdlg, IDC_SOUND_BUFFER, config.soundbuffer, false);
}

BOOL _cdecl M88Config2::SoundPageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->SoundPage(hwnd, m, w, l);
}

// ---------------------------------------------------------------------------
//	ʐݒy[W
// ---------------------------------------------------------------------------

void M88Config2::InitVolumeSlider(HWND hdlg, UINT id, int val)
{
	SendDlgItemMessage(hdlg, id, TBM_SETLINESIZE, 0, 1);
	SendDlgItemMessage(hdlg, id, TBM_SETPAGESIZE, 0, 5);
	SendDlgItemMessage(hdlg, id, TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
	SendDlgItemMessage(hdlg, id, TBM_SETPOS, FALSE, val);
}

BOOL M88Config2::SoundVolumePage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		InitVolumeSlider(hdlg, IDC_SOUND_VOLFM,     orgconfig.volfm);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLSSG,    orgconfig.volssg);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLADPCM,  orgconfig.voladpcm);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLRHYTHM, orgconfig.volrhythm);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLBD,     orgconfig.volbd);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLSD,     orgconfig.volsd);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLTOP,    orgconfig.voltop);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLHH,     orgconfig.volhh);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLTOM,    orgconfig.voltom);
		InitVolumeSlider(hdlg, IDC_SOUND_VOLRIM,    orgconfig.volrim);
		break;

	case WM_HSCROLL:
		SoundVolumePageUpdateSlider(hdlg);
		PropSheet_Changed( hwndps, hdlg );
		return 0;
	

	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
//			SoundVolumePageUpdate(hdlg);
			SoundVolumePageUpdateSlider(hdlg);
			page = soundvol;		// ݑIĂy[W
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLFM,     TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLSSG,    TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLADPCM,  TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLRHYTHM, TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLBD,     TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLSD,     TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOP,    TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLHH,     TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOM,    TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLRIM,    TBM_SETRANGE, TRUE, MAKELONG(-40, 20));
			break;

		case PSN_APPLY:
			config.volfm     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLFM,     TBM_GETPOS, 0, 0);
			config.volssg    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLSSG,    TBM_GETPOS, 0, 0);
			config.voladpcm  = SendDlgItemMessage(hdlg, IDC_SOUND_VOLADPCM,  TBM_GETPOS, 0, 0);
			config.volrhythm = SendDlgItemMessage(hdlg, IDC_SOUND_VOLRHYTHM, TBM_GETPOS, 0, 0);
			config.volbd     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLBD,     TBM_GETPOS, 0, 0);
			config.volsd     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLSD,     TBM_GETPOS, 0, 0);
			config.voltop    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOP,    TBM_GETPOS, 0, 0);
			config.volhh     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLHH,     TBM_GETPOS, 0, 0);
			config.voltom    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOM,    TBM_GETPOS, 0, 0);
			config.volrim    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLRIM,    TBM_GETPOS, 0, 0);

			orgconfig = config;
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;

		case PSN_QUERYCANCEL:
			PostMessage(hwndparent, WM_M88_CHANGEVOLUME, (WPARAM) &orgconfig, 0);
			return FALSE;
		}
		break;

	case WM_COMMAND:
		return SoundVolumePageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::SoundVolumePageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		PropSheet_Changed( hwndps, hdlg );
		switch (id)
		{
		case IDC_SOUND_RESETVOL:
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLFM,     TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLSSG,    TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLADPCM,  TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLRHYTHM, TBM_SETPOS, TRUE, 0);
			SoundVolumePageUpdateSlider(hdlg);
			break;

		case IDC_SOUND_RESETRHYTHM:
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLBD,     TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLSD,     TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOP,    TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLHH,     TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOM,    TBM_SETPOS, TRUE, 0);
			SendDlgItemMessage(hdlg, IDC_SOUND_VOLRIM,    TBM_SETPOS, TRUE, 0);
			SoundVolumePageUpdateSlider(hdlg);
			break;
		}
	}
	return FALSE;
}

void M88Config2::SoundVolumePageUpdateSlider(HWND hdlg)
{
	config.volfm     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLFM,     TBM_GETPOS, 0, 0);
	config.volssg    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLSSG,    TBM_GETPOS, 0, 0);
	config.voladpcm  = SendDlgItemMessage(hdlg, IDC_SOUND_VOLADPCM,  TBM_GETPOS, 0, 0);
	config.volrhythm = SendDlgItemMessage(hdlg, IDC_SOUND_VOLRHYTHM, TBM_GETPOS, 0, 0);
	config.volbd     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLBD,     TBM_GETPOS, 0, 0);
	config.volsd     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLSD,     TBM_GETPOS, 0, 0);
	config.voltop    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOP,    TBM_GETPOS, 0, 0);
	config.volhh     = SendDlgItemMessage(hdlg, IDC_SOUND_VOLHH,     TBM_GETPOS, 0, 0);
	config.voltom    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLTOM,    TBM_GETPOS, 0, 0);
	config.volrim    = SendDlgItemMessage(hdlg, IDC_SOUND_VOLRIM,    TBM_GETPOS, 0, 0);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTFM,     config.volfm);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTSSG,    config.volssg);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTADPCM,  config.voladpcm);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTRHYTHM, config.volrhythm);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTBD,     config.volbd);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTSD,     config.volsd);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTTOP,    config.voltop);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTHH,     config.volhh);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTTOM,    config.voltom);
	SoundSetVolumeText(hdlg, IDC_SOUND_VOLTXTRIM,    config.volrim);
	PostMessage(hwndparent, WM_M88_CHANGEVOLUME, (WPARAM) &config, 0);
}

void M88Config2::SoundSetVolumeText(HWND hdlg, int id, int val)
{
	if (val > -40)
		SetDlgItemInt(hdlg, id, val, TRUE);
	else
		SetDlgItemText(hdlg, id, "Mute");
}

BOOL _cdecl M88Config2::SoundVolumePageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->SoundVolumePage(hwnd, m, w, l);
}

// ---------------------------------------------------------------------------
//	Function Page
// ---------------------------------------------------------------------------

BOOL M88Config2::FunctionPage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		config.padsensibility = orgconfig.padsensibility;
		SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_SETLINESIZE, 0, 256);
		SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_SETPAGESIZE, 0, 4096);
		SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_SETRANGE, TRUE, MAKELONG(10, 32757));
		SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_SETPOS, FALSE, config.padsensibility);
		break;

	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
			FunctionPageUpdate(hdlg);
			SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_SETRANGE, TRUE, MAKELONG(10, 32757));
			page = function;
			break;

		case PSN_APPLY:
			orgconfig = config;
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;
		}
		break;

	case WM_COMMAND:
		return FunctionPageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	case WM_HSCROLL:
		config.padsensibility = SendDlgItemMessage(hdlg, IDC_FUNCTION_PADSENSE, TBM_GETPOS, 0, 0);
		PropSheet_Changed( hwndps, hdlg );
		break;

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::FunctionPageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		PropSheet_Changed( hwndps, hdlg );
		switch (id)
		{
		case IDC_FUNCTION_SAVEDIR:
			config.flags ^= Config::savedirectory;
			return 0;

		case IDC_FUNCTION_ASKBEFORERESET:
			config.flags ^= Config::askbeforereset;
			return 0;

		case IDC_FUNCTION_SUPPRESSMENU:
			config.flags ^= Config::suppressmenu;
			return 0;
		
		case IDC_FUNCTION_USEARROWFOR10:
			config.flags ^= Config::usearrowfor10;
			return 0;
		
		case IDC_FUNCTION_SWAPPADBUTTONS:
			config.flags ^= Config::swappadbuttons;
			return 0;

		case IDC_FUNCTION_WATCHREGISTER:
			config.flags ^= Config::watchregister;
			return 0;
		
		case IDC_FUNCTION_USEQPC:
			config.flags ^= Config::useqpc;
			return 0;

		case IDC_FUNCTION_ENABLEPAD:
			config.flags ^= Config::enablepad;
			FunctionPageUpdate(hdlg);
			return 0;

		case IDC_FUNCTION_RESETF12:
			config.flags ^= Config::disablef12reset;
		}
	}
	return FALSE;
}

void M88Config2::FunctionPageUpdate(HWND hdlg)
{
	CheckDlgButton(hdlg, IDC_FUNCTION_SAVEDIR, BSTATE(config.flags & Config::savedirectory));
	CheckDlgButton(hdlg, IDC_FUNCTION_ASKBEFORERESET, BSTATE(config.flags & Config::askbeforereset));
	CheckDlgButton(hdlg, IDC_FUNCTION_SUPPRESSMENU, BSTATE(config.flags & Config::suppressmenu));
	CheckDlgButton(hdlg, IDC_FUNCTION_USEARROWFOR10, BSTATE(config.flags & Config::usearrowfor10));
	CheckDlgButton(hdlg, IDC_FUNCTION_ENABLEPAD, BSTATE(config.flags & Config::enablepad));
	EnableWindow(GetDlgItem(hdlg, IDC_FUNCTION_SWAPPADBUTTONS), (config.flags & Config::enablepad));
	CheckDlgButton(hdlg, IDC_FUNCTION_SWAPPADBUTTONS, BSTATE(config.flags & Config::swappadbuttons));
	CheckDlgButton(hdlg, IDC_FUNCTION_WATCHREGISTER, BSTATE(config.flags & Config::watchregister));
	CheckDlgButton(hdlg, IDC_FUNCTION_USEQPC, BSTATE(config.flags & Config::useqpc));
	CheckDlgButton(hdlg, IDC_FUNCTION_RESETF12, BSTATE(!(config.flags & Config::disablef12reset)));
}

BOOL _cdecl M88Config2::FunctionPageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->FunctionPage(hwnd, m, w, l);
}

// ---------------------------------------------------------------------------
//	Switch Page
// ---------------------------------------------------------------------------

BOOL M88Config2::SwitchPage(HWND hdlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{

	case WM_NOTIFY:
		switch (((NMHDR*) lp)->code)
		{
		case PSN_SETACTIVE:
			SwitchPageUpdate(hdlg);
			page = switches;
			break;

		case PSN_APPLY:
			orgconfig = config;
			PostMessage(hwndparent, WM_M88_APPLYCONFIG, (WPARAM) &config, 0);
			return PSNRET_NOERROR;
		}
		break;

	case WM_COMMAND:
		return SwitchPageCommand(hdlg, HWND(lp), HIWORD(wp), LOWORD(wp));

	default:
		return FALSE;
	}
	return TRUE;
}

BOOL M88Config2::SwitchPageCommand(HWND hdlg, HWND hwctl, UINT nc, UINT id)
{
	if (nc == BN_CLICKED)
	{
		if (IDC_DIPSW_1H <= id && id <= IDC_DIPSW_CL)
		{
			int n = (id - IDC_DIPSW_1H) / 2;
			int s = (id - IDC_DIPSW_1H) & 1;

			if (!s)	// ON
				config.dipsw &= ~(1 << n);
			else
				config.dipsw |= 1 << n;
			PropSheet_Changed( hwndps, hdlg );
			return 0;
		}

		switch (id)
		{
		case IDC_DIPSW_DEFAULT:
			config.dipsw = 1829;
			PropSheet_Changed( hwndps, hdlg );
			SwitchPageUpdate(hdlg);
			return 0;
		}
	}
	return FALSE;
}

void M88Config2::SwitchPageUpdate(HWND hdlg)
{
	for (int i=0; i<12; i++)
	{
		CheckDlgButton(hdlg, IDC_DIPSW_1L + i*2, BSTATE(0 != (config.dipsw & (1 << i))));
		CheckDlgButton(hdlg, IDC_DIPSW_1H + i*2, BSTATE(0 == (config.dipsw & (1 << i))));
	}
}

BOOL _cdecl M88Config2::SwitchPageGate
(M88Config2* config, HWND hwnd, UINT m, WPARAM w, LPARAM l)
{
	return config->SwitchPage(hwnd, m, w, l);
}



