// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  PC88 ̃x[XfoCX̃G~[V
// ---------------------------------------------------------------------------
//	$Id: PC88.cpp,v 1.2 1999/01/11 15:32:12 cisc Exp $

#include <stdio.h>
#include <string.h>
#include "PC88.h"
#include "Screen.h"
#include "Screen_i.h"
#include "Text.h"
#include "misc.h"
#include "SubSys.h"
#include "Sound.h"

// ---------------------------------------------------------------------------

#ifndef RELEASEBUILD
	#define OUTP_DEBUG
#endif

#ifdef OUTP_DEBUG
	#include <stdio.h>
	#define Log0(a)			if (fp) fprintf(fp, a)
	#define Log1(a, b)		if (fp) fprintf(fp, a, b)
	#define Log2(a, b, c)	if (fp) fprintf(fp, a, b, c)
	#define Log3(a, b, c, d) if (fp) fprintf(fp, a, b, c, d)
	static FILE* fp;
	static void LogBegin()   { if (!fp) fp = fopen("outport.log", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
#else
	#define LogBegin()		 ((void) 0)
	#define LogEnd()		 ((void) 0)
	#define Log0(a)			 ((void) 0)
	#define Log1(a, b)		 ((void) 0)
	#define Log2(a, b, c)	 ((void) 0)
	#define Log3(a, b, c, d) ((void) 0)
#endif


// ---------------------------------------------------------------------------
// \z/
// ---------------------------------------------------------------------------

PC88::PC88(ROM* _rom, Screen* _screen, Device* keyb, Device* cal, Sound88* snd)
:
	rom(_rom),					// 
	screen(_screen),			// OtBbNXEʐ
	keyboard(keyb),				// L[{[h
	calender(cal),				// J_v
	sound(snd),					// 
	kanjirom(_rom->kanji),		//  ROM
	kanjirom2(_rom->kanji2),	//  ROM (2)
	fdsub(0)		
{
	LogBegin();
	crtc = 0;
	tvram = 0;
	textscreen = 0;
	extram = 0;
	sw30 = 0xfa;
	sw31 = 0x29;
	waitbank = -1;
}

PC88::~PC88()
{
	LogEnd();
	delete[] mainram;
	delete[] tvram;
	delete[] extram;
	delete textscreen;
	delete crtc;
}

bool PC88::Init(int nerams)
{
	crtc = new CRTC(*this);		// CRTC initialization
	textscreen = new TextScreen(*crtc, rom->font); 

	mainram = new uint8[0x10000];
	tvram = new uint8[0x1000];

	erambanks = Limit(nerams, 256, 4);
	extram = new uint8[0x8000 * erambanks];

	if (!crtc || !textscreen || !tvram || !extram)
		return false;

	eromtable[0] = rom->e0;	eromtable[1] = rom->e1; 
	eromtable[2] = rom->e2; eromtable[3] = rom->e3; 
	rom60 = &rom->n88[0x6000];
	gvramselected = false;
	sw30 = 0xfb;
	sw31 = 0x69;
	sw6e = 0xff;

	OutPort(0x30, 255); OutPort(0x30, 0);
	OutPort(0x31, 255); OutPort(0x31, 0);
	OutPort(0x32, 255); OutPort(0x32, 0);
	InitMemory();


	return true;
}

void PC88::Reset(BASICMODE mode)
{
	Z80::Reset(1);
	Log0("Reset!");

	sw30 = (sw30 & ~1)    | (mode == N80   ? 0x00 : 0x01);
	sw31 = (sw31 & ~0xc0) | (mode == N88V2 ? 0x40 : (mode == N88V1H ? 0xc0 : 0x80));
	// FA B9
	
	OutPort(0x31, 0);
	OutPort(0x32, 0x80);
	OutPort(0x34, 0);
	OutPort(0x35, 0);
	OutPort(0x40, 0);
	OutPort(0x52, 0);
	OutPort(0x70, 0);
	OutPort(0xe2, 0);
	OutPort(0xe4, 0);
	OutPort(0xe6, 0);

	dmac.Reset();
	screen->Reset();
	calender->Reset();
	keyboard->Reset();
	sound->Reset();
	fdsub->Reset();
	crtc->Reset(line200);
	textscreen->Reset();

	port40 = 0xc4 + (line200 ? 2 : 0) + (fdboot ? 0 : 8);
	
	usetvram = sw31 & 0x40 ? true : false;
	
	SelectROM();
	SelectTVRAM();
	
	Z80::Reset(0);
}

// ---------------------------------------------------------------------------
//	 I/O
// ---------------------------------------------------------------------------

void PC88::Write(uint addr, uint data)
{
	if (screen->IsGVRAMSelected())
	{
		if (addr < 0xc000)
			writeaccess[addr >> 8][addr & 0xff] = data;
		else
			screen->Write(addr & 0x3fff, data);
	}
	else
		writeaccess[addr >> 8][addr & 0xff] = data;
}

uint8 PC88::Read(uint addr)
{
	if (screen->IsGVRAMSelected())
	{
		if (addr < 0xc000)
			return readaccess[addr >> 8][addr & 0xff];
		else
			return screen->Read(addr & 0x3fff);
	}
	else
		return readaccess[addr >> 8][addr & 0xff];
}

// ---------------------------------------------------------------------------
//	eLXg VRAM ̓e tvram ɓ]
//	tvram	1000h oCg̃obt@
//	Ԃl	]̐Es
// ---------------------------------------------------------------------------

bool PC88::GetTVRAM(uint8* dest, uint nbytes)
{
	uint16 addr, count;

	while (nbytes > 0)
	{
		count = uint16(nbytes);
		if (!dmac.Request(2, &addr, &count, false))
			return false;
		
		if (usetvram)
		{
			if ((addr < 0xf000) || (uint32(addr) + count > 0x10000))
				return false;
			memcpy(dest, &tvram[addr & 0xfff], count);
		}
		else
		{
			if (uint32(addr) + count <= 0x10000)	// x̓]ōςޏꍇ
			{
				memcpy(dest, &mainram[addr], count);
			}
			else
			{										// Qx̓]Kvȏꍇ
				uint32 x1 = 0x10000 - addr;
				memcpy(dest, &mainram[addr], x1);
				memcpy(dest+x1, &mainram[0], count-x1);
			}
		}
		dest += count;
		nbytes -= count;
	}
	return true;
}

// ---------------------------------------------------------------------------
//	Device IN
// ---------------------------------------------------------------------------

uint8 PC88::InPort(uint port)
{
	port &= 0xff;
	switch (port)
	{
	case 0x00: case 0x01: case 0x02: case 0x03:
	case 0x04: case 0x05: case 0x06: case 0x07:
	case 0x08: case 0x09: case 0x0a: case 0x0b:
	case 0x0c: case 0x0d: case 0x0e: case 0x0f:
//		Log2("(%.2x)->%.2x\n", port, keyboard->In(port));
		return keyboard->In(port);

	case 0x30:
		return sw30;
	case 0x31:
		return sw31;
	case 0x32:
		return port32;

	case 0x40:
		port40 = (port40 & ~0x20) | (vrtc ? 0x20 : 0x00);
		return port40 | calender->In(port);

	case 0x44: case 0x45: case 0x46: case 0x47:
	case 0xa8: case 0xa9: case 0xac: case 0xad:
		return sound->In(port);

	case 0x50: case 0x51:
		return crtc->In(port);

	case 0x5c:
		return screen->In5c();

	case 0x68:
		return dmac.In(port);

	case 0x6e:
		return sw6e;

	case 0x70:
		return port70;

	case 0x71:
		return port71;
	
	case 0xe2:
//		Log2("(%.2x)->%.2x\n", port, (porte3 & 0xfc) ? 0xff : (~porte2) | 0xee);
		return (porte3 & 0xfc) ? 0xff : (~porte2) | 0xee;
	case 0xe3:
//		Log2("(%.2x)->%.2x\n", port, (porte3 & 0xfc)? 0xff : porte3);
		return (porte3 & 0xfc) ? 0xff : porte3;

	case 0xe8: case 0xe9:
		return kanjirom.In(port);
	
	case 0xec: case 0xed:
		return kanjirom2.In(port);
	
	case 0xfc: case 0xfd: case 0xfe: case 0xff:
		return fdsub->In(port);
	}
	return 0xff;
}

// ---------------------------------------------------------------------------
//	Device OUT
// ---------------------------------------------------------------------------

void PC88::OutPort(uint port, uint data)
{
//	Log2("(%.2x)<-%.2x\n", port, data);
	port &= 0xff; data &= 0xff;
	switch (port)
	{
		uint8 d;

	case 0x00: case 0x01: case 0x02:
		textscreen->Out(port, data);
		return;

	case 0x10:
		calender->Out(port, data);
		return;

	case 0x30:
		crtc->Out(port, data);
		screen->Out30(data);
		return;

	case 0x31:
		d = port31 ^ data;
		port31 = data;
		if (d & 0x06)
			SelectROM();
//		Log0("\n");
		screen->Out31(data);
		return;
	
	case 0x32:
		d = port32 ^ data;
		port32 = data;
		if (d & 0x13)
		{
			SelectEROM();
			SelectROM();
			SelectTVRAM();
		}
		screen->Out32(data);
		gvramselected = screen->IsGVRAMSelected();
		sound->Out(port, data);
//		sintm = data & 0x80 ? ~0x300 : ~0;
		SetInterruptMask();
		SetWaitTable();
		return;

	case 0x34:	
		screen->Out34(data); 
		return;

	case 0x35:
		screen->Out35(data); 
		gvramselected = screen->IsGVRAMSelected();
		SetWaitTable();
		return;
	
	case 0x40:
		calender->Out(port, data);
		sound->Out40(data, GetClockCount());
		port40o = data;
		SetWaitTable();
		return;

	case 0x44: case 0x45: case 0x46: case 0x47:
	case 0xa8: case 0xa9: case 0xaa: case 0xac: case 0xad:
		sound->Out(port, data);
		return;

	case 0x50: case 0x51:
		crtc->Out(port, data);
		return;

	case 0x52:
		screen->Out52(data); 
		return;

	case 0x53:	
		screen->Out53(data); 
		return;

	case 0x54:	
		screen->Out54(data); 
		return;

	case 0x55: case 0x56: case 0x57:
	case 0x58: case 0x59: case 0x5a: case 0x5b:
		screen->Out55to5b(port, data);
		return;
	
	case 0x5c: case 0x5d: case 0x5e: case 0x5f:
		screen->Out5cto5f(port);
		gvramselected = screen->IsGVRAMSelected();
		return;

	case 0x60: case 0x61: case 0x62: case 0x63:
	case 0x64: case 0x65: case 0x66: case 0x67:
		dmac.Out(port, data);
		return;

	case 0x68:
		dmac.Out(port, data);
//		usetvram = !(port32 & 0x10);
		SetWaitTable();
		break;

	case 0x70:
		port70 = data;
		SetTextWindow();
		return;

	case 0x71:
		port71 = data;
		SelectEROM();
		SelectROM();
//		Log0("\n");
		return;

	case 0x78:
		port70++;
		SetTextWindow();
		return;

	case 0xe2:
//		Log2("(%.2x)<-%.2x\n", port, data);
		porte2 = data;
		SelectROM();
		SelectRAM();
		break;

	case 0xe3:
//		Log2("(%.2x)<-%.2x\n", port, data);
		porte3 = data;
		SelectROM();
		SelectRAM();
		break;

	case 0xe4:
		porte4 = data;
		SetInterruptMask();
		return;

	case 0xe6:
		porte6 = data;
		if (!(porte6 & 1)) CancelInterrupt(2*2);
		if (!(porte6 & 2)) CancelInterrupt(1*2);
		if (!(porte6 & 4)) CancelInterrupt(0*2);
		return;

	case 0xe8: case 0xe9:
		kanjirom.Out(port, data);
		return;

	case 0xec: case 0xed:
		kanjirom2.Out(port, data);
		return;

	case 0xfc: case 0xfd: case 0xfe: case 0xff:
		fdsub->Out(port, data);
		return;
	}
}

// ---------------------------------------------------------------------------
//	荞݊֌W
// ---------------------------------------------------------------------------

void PC88::AcceptInterrupt(uint intno)
{
//	Log1("Accept int%d\n", intno);
	CancelInterrupt(intno);
	porte4 = 0;
	SetIMask(0);
}

inline void PC88::SetInterruptMask()
{
	SetIMask((1 << ((porte4 & 15) * 2)) - 1);
}

// ---------------------------------------------------------------------------
//	RAM ύX
// ---------------------------------------------------------------------------

void PC88::SelectRAM()
{
	uint8* ram;
	if ((porte2 & 0x10) && (porte3 < erambanks))
		ram = &extram[porte3 * 0x8000];
	else
		ram = mainram;
	
	if (writeaccess[0] != ram)
	{
		for (int i=0; i<0x80; i+=4, ram+=0x400)
		{
			writeaccess[i  ] = ram;
			writeaccess[i+1] = ram + 0x100;
			writeaccess[i+2] = ram + 0x200;
			writeaccess[i+3] = ram + 0x300;
		}
	}
}

// ---------------------------------------------------------------------------
//	qnlύXꍇ
// ---------------------------------------------------------------------------

void PC88::SelectROM()
{
	uint8* a00;
	uint8* a60;

	if ((porte2 & 1) && (porte3 < erambanks))
	{
		a00 = &extram[porte3 * 0x8000];
		a60 = a00 + 0x6000;
	}
	else
	{
		if ((port31 & 6) == 0)
		{
			// N88 ROM
			a00 = &rom->n88[0x0000];
			a60 = rom60;
			SetTextWindow();
//			Log1("MEMORY: N88(%.2x) ", port70);
		}
		else
		{
			if (port31 & 2)
			{
				// ALL RAM
				a00 = &mainram[0x0000];
				a60 = &mainram[0x6000];
//				Log0("MEMORY: ALL RAM ");
			}
			else
			{
				// N80
				a00 = &rom->n80[0x0000];
				a60 = &rom->n80[0x6000];
//				Log0("MEMORY: N80     ");
			}
			// no text window
			readaccess[0x80] = writeaccess[0x80] = &mainram[0x8000];
			readaccess[0x81] = writeaccess[0x81] = &mainram[0x8100];
			readaccess[0x82] = writeaccess[0x82] = &mainram[0x8200];
			readaccess[0x83] = writeaccess[0x83] = &mainram[0x8300];
		}
	}

	if (readaccess[0] != a00)
	{
		for (int i=0; i<0x60; i+=4, a00 += 0x400)
		{
			readaccess[i+0] = a00;			readaccess[i+1] = a00 + 0x100;
			readaccess[i+2] = a00 + 0x200;	readaccess[i+3] = a00 + 0x300;
		}
	}
	if (readaccess[0x60] != a60)
	{
		for (int i=0x60; i<0x80; i+=4, a60 += 0x400)
		{
			readaccess[i+0] = a60;			readaccess[i+1] = a60 + 0x100;
			readaccess[i+2] = a60 + 0x200;	readaccess[i+3] = a60 + 0x300;
		}
	}
}

// ---------------------------------------------------------------------------
//	EROM ύXꍇ
// ---------------------------------------------------------------------------

void PC88::SelectEROM()
{
	if (!(port71 & 1))
	{
		rom60 = eromtable[port32 & 3];
	}
	else
	{
		rom60 = &rom->n88[0x6000];
	}
}

// ---------------------------------------------------------------------------
//	֌W̏
// ---------------------------------------------------------------------------

void PC88::InitMemory()
{
	int i;
	uint8* a = mainram;
	for (i=0; i<0xf0; i++)
	{
		readaccess[i] = writeaccess[i] = a;
		a += 0x100;
	}
	SelectROM();			// qnl̓ǂݍ݃e[uK
	SelectTVRAM();

	memset(extram, 0, 0x20000);
	memset(tvram, 0, 0x1000);
	
	// Np^[쐬 (SR ^)
	for (i=0; i<0xffff; i+=0x80)
	{
		uint8 b = ((i >> 7) ^ i) & 0x100 ? 0x00 : 0xff;
		memset(&mainram[i     ],  b, 0x40);
		memset(&mainram[i+0x40], ~b, 0x40);
		mainram[i+0x7f] = b;
	}
	mainram[0xffff] = 0;
//	memset(mainram, 0xc3, 0x10000);		// for debug
}

// ---------------------------------------------------------------------------
//	eLXg̎QƈʒuύX
// ---------------------------------------------------------------------------

void PC88::SetTextWindow()
{
	if (0 == (port31 & 6))
	{
		uint window = port70 * 0x100;
		readaccess[0x80] = writeaccess[0x80] = &mainram[(window      )       ];
		readaccess[0x81] = writeaccess[0x81] = &mainram[(window+0x100)&0xffff];
		readaccess[0x82] = writeaccess[0x82] = &mainram[(window+0x200)&0xffff];
		readaccess[0x83] = writeaccess[0x83] = &mainram[(window+0x300)&0xffff];
	}
}

// ---------------------------------------------------------------------------
//	eLXguq`l i0f000h - 0ffffh) I
// ---------------------------------------------------------------------------

void PC88::SelectTVRAM()
{
	uint8* a = !(port32 & 0x10) && usetvram ? tvram : &mainram[0xf000];
//	Log0(!(port32 & 0x10) && usetvram ? "TVRAM\n" : "MAIN RAM\n");
	for (int i=0xf0; i<0x100; i+=4)
	{
		readaccess[i  ] = writeaccess[i  ] = a;
		readaccess[i+1] = writeaccess[i+1] = a + 0x100;
		readaccess[i+2] = writeaccess[i+2] = a + 0x200;
		readaccess[i+3] = writeaccess[i+3] = a + 0x300;
		a += 0x400;
	}
}

// ---------------------------------------------------------------------------
//	ʏۑ
// ---------------------------------------------------------------------------

void PC88::Snap()
{
	screen->Snap();
	textscreen->Snap();
}

// ---------------------------------------------------------------------------
//	ʍXV
// ---------------------------------------------------------------------------

void PC88::UpdateScreen
(uint8* image, int bpl, ulong clock, int* region, bool refresh)
{
	int region1[2], region2[2];
	screen->UpdateScreen(image, bpl, region2);
	textscreen->UpdateScreen(image, bpl, 400, clock, region1, refresh);
	
	region[0] = Min(region1[0], region2[0]);
	region[1] = Max(region1[1], region2[1]);
}

// ---------------------------------------------------------------------------
//	ram     text    gvram   gvramh
//	1	2	1	2	1	42	3	4	4MHz v1
//	0	0	0.5	0.5	1	2	1	2        v1h
//	0	0	0.5	0.5	1	2	1	2        v2
//	1	2	1	2	7	72	4	8	8MHz
//	1	1	2	2	3.5	5	4	5
//	1	1	2	2	3.5	5	4	5

PC88::Waitdesc const PC88::waittable[64] = 
{
//	BLANK/NO DMA	BLANK/DMA		DISPLAY/NO DMA	DISPLAY/DMA
//	4M
	{  2,  2,  2 }, {  2,  2,  2 }, {  2,  2,  2 }, {  4,  4,  4 }, // LNW
	{  2,  2,  2 }, {  2,  2,  2 }, {  2,  2,  2 }, {  4,  4,  4 }, // LNF
	{  2,  2,  2 }, {  2,  2,  2 }, { 84, 84, 84 }, { 84, 84, 84 }, // LGW
	{  2,  6,  6 }, {  2,  6,  6 }, {  2,  6,  6 }, {  4,  8,  8 }, // LGF
	
	{  0,  0,  1 }, {  0,  0,  1 }, {  0,  0,  1 }, {  0,  0,  1 }, // HNW
	{  0,  0,  1 }, {  0,  0,  1 }, {  0,  0,  1 }, {  0,  0,  1 }, // HNF
	{  0,  2,  1 }, {  0,  2,  1 }, {  0,  4,  1 }, {  0,  4,  1 }, // HGW
	{  0,  2,  1 }, {  0,  2,  1 }, {  0,  4,  1 }, {  0,  4,  1 }, // HGW
// 8M
	{  2,  2,  2 }, {  2,  2,  2 }, {  2,  2,  2 }, {  4,  4,  4 }, // LO, M, NORMAL
	{  2,  2,  2 }, {  2,  2,  2 }, {  2,  2,  2 }, {  4,  4,  4 }, // LO, M, FAST
	{  2, 14,  2 }, {  2, 14,  2 }, {144,144,144 }, {144,144,144 }, // LO, G, NORMAL
	{  2,  8,  2 }, {  2,  8,  2 }, { 16, 16, 16 }, { 16, 16, 16 }, // LO, G, FAST

	{  2,  2,  4 }, {  2,  2,  4 }, {  2,  2,  4 }, {  2,  2,  4 },
	{  2,  2,  4 }, {  2,  2,  4 }, {  2,  2,  4 }, {  2,  2,  4 },
	{  2,  7,  4 }, {  2,  7,  4 }, {  2, 10,  4 }, {  2, 10,  4 },
	{  2,  7,  4 }, {  2,  7,  4 }, {  2, 10,  4 }, {  2, 10,  4 },
};

void PC88::SetWaitTable()
{
	int i;
	uint b = (sw6e & 0x80 ? 0 : 0x20)
		  | (sw31 & 0x40 ? 0x10 : 0)
		  | (gvramselected ? 0x08 : 0)
		  | (port40o & 0x10 ? 0x04 : 0)
		  | (vrtc ? 0 : 0x02)
		  | (crtc->IsOn() ? 1 : 0);

	if (waitbank != b)
	{
		Log1("bank = %.2x\n", b);
		waitbank = b;
		uint wait = waittable[b].mainram;
		for (i=0; i<0xf0; i+=0x8)
		{
			memwait [i+0] = memwait [i+1] = memwait [i+2] = memwait [i+3] = wait;
			memwait [i+4] = memwait [i+5] = memwait [i+6] = memwait [i+7] = wait;
			memwait2[i+0] = memwait2[i+1] = memwait2[i+2] = memwait2[i+3] = wait*2;
			memwait2[i+4] = memwait2[i+5] = memwait2[i+6] = memwait2[i+7] = wait*2;
		}
		if (readaccess[0xf0] == tvram)
			wait = waittable[b].tvram;
		for (; i<0x100; i+=0x4)
		{
			memwait [i+0] = memwait [i+1] = memwait [i+2] = memwait [i+3] = wait;
			memwait2[i+0] = memwait2[i+1] = memwait2[i+2] = memwait2[i+3] = wait*2;
		}
		gvramwait = waittable[b].gvram;
		gvramwait2 = gvramwait * 2;
	}
}
