// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  PC88 ̃x[XfoCX̃G~[V
// ---------------------------------------------------------------------------
//	$Id: PC88.h,v 1.2 1999/01/11 15:32:12 cisc Exp $

#if !defined(AFX_PC88_H__8E5D1436_291B_11D1_A283_A086A6A721AF__INCLUDED_)
#define AFX_PC88_H__8E5D1436_291B_11D1_A283_A086A6A721AF__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "Device.h"
#include "DMAC.h"
#include "KanjiROM.h"
#include "Z80.h"

#ifdef Z80_TEST
	#include "Z80Test.h"
	typedef Z80Test Z80;
#else
	#ifdef Z80_X86_NATIVE
		#include "Z80_x86.h"
		typedef Z80_x86 Z80;
	#else
		#if 1
			#include "Z80R.h"
			typedef Z80R Z80;
		#else
			#include "Z80MK.h"
			typedef Z80MK Z80;
		#endif
	#endif
#endif


// ---------------------------------------------------------------------------

class Screen;
class TextScreen;
class CRTC;
class SubSystem;
class Sound88;

class PC88 : public Z80
{
	typedef uint8 (PC88::*Reader)(uint);
	typedef void  (PC88::*Writer)(uint, uint);

public:
	class ROM
	{
	private:
		int LoadROMImage(uint8* dest, char* filename, int32 size);
		bool LoadPC88ROM();
	public:
		uint8* rom;
		
		uint8* n88;
		uint8* n80;
		uint8* e0;
		uint8* e1;
		uint8* e2;
		uint8* e3;
		uint8* kanji;
		uint8* kanji2;
		uint8* disk;
		uint8* font;

		ROM();
		~ROM();
		bool LoadROM();
	};

	enum BASICMODE { N80 = 0, N88V1, N88V2, N88V1H=11 };

public:
	PC88(ROM* _rom, Screen* _screen, Device* keyb, Device* cal, Sound88* snd);
	virtual ~PC88();
	bool Init(int nerams);
	
	void Intr(uint8 intno);
	void IntrClear(uint8 intno);
	bool GetTVRAM(uint8*, uint);
	
	virtual void Reset(BASICMODE mode = N88V2);
	void Snap();
	void UpdateScreen(uint8* image, int bpl, ulong clock, int* region, bool refresh);

protected:
	uint8 Read(uint addr);
	void Write(uint addr, uint data);
	uint8 InPort(uint port);
	void OutPort(uint port, uint data);
	void SetWaitTable();

	void AcceptInterrupt(uint intno);

	SubSystem* fdsub;
	Screen* screen;
	TextScreen* textscreen;
	Device* keyboard;
	Device* calender;
	Sound88* sound;
	ROM* rom;
	CRTC* crtc;
	uint8 sw6e;

protected:
	bool vrtc;
	bool line200;
	bool fdboot;
	bool IsSIntrEnabled() { return !(port32 & 0x80); }
	uint8 sw30, sw31;

private:
	struct Waitdesc
	{
		uint8 mainram, gvram, tvram;
	};
	static const Waitdesc waittable[64];
	
	void InitMemory();
	void SelectEROM();
	void SelectROM();
	void SelectRAM();
	void SelectTVRAM();
	void SetTextWindow();
	void SetInterruptMask();
	
	uint8* eromtable[4];

	uint8* erom;
	uint8* mainram;
	uint8* tvram;
	uint8* rom60;
	uint8* extram;
	uint32 prevclk;
	uint waitbank;
	int erambanks;
	bool usetvram;				// TVRAM eLXg\pƂĎg
	
	uint8 port31, port32;
	uint8 port40, port40o;
	uint8 port70;
	uint8 port71;
	uint8 porte2, porte3;
	uint8 porte4, porte6;
	
	KanjiROM kanjirom;
	KanjiROM kanjirom2;
	DMAC dmac;

	// for debug
	const char* GetDesc();
};

// ---------------------------------------------------------------------------

inline void PC88::Intr(uint8 intno)
{
	if (intno < 3 && !((4 >> intno) & porte6))
		return;
	RequestInterrupt(intno*2);
}

inline void PC88::IntrClear(uint8 intno)
{
	CancelInterrupt(intno*2);
}

#endif // !defined(AFX_PC88_H__8E5D1436_291B_11D1_A283_A086A6A721AF__INCLUDED_)
