// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1997, 1998.
// ---------------------------------------------------------------------------
//	PSG+BEEP Implementation
// ---------------------------------------------------------------------------
//	$Id: PSGB.cpp,v 1.2 1999/01/11 15:32:12 cisc Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "misc.h"
#include "PSGB.h"

// ---------------------------------------------------------------------------
//	Construct
// ---------------------------------------------------------------------------

PSGB::PSGB()
{
	pcount = 0;
	bcount = 0;
	bbcount = 0; 
}

PSGB::~PSGB()
{
}

// ---------------------------------------------------------------------------

bool PSGB::Init(uint rate)
{
	bbperiod = int(2400.0 / rate * (1 << 17));
	return true;
}

// ---------------------------------------------------------------------------
//	EmitWave
// ---------------------------------------------------------------------------

void PSGB::EmitWave(Sample* dest, int nsamples, uint32 bperiod, uint32 pperiod, uint p40)
{
	uint8 chenable[3], nenable[3];
	uint8 r7 = ~reg[7];

	chenable[0] = (r7 & 0x01) && (speriod[0] <= (1 << SSG_TONESHIFT));
	chenable[1] = (r7 & 0x02) && (speriod[1] <= (1 << SSG_TONESHIFT));
	chenable[2] = (r7 & 0x04) && (speriod[2] <= (1 << SSG_TONESHIFT));
	nenable[0]  = (r7 >> 3) & 1;
	nenable[1]  = (r7 >> 4) & 1;
	nenable[2]  = (r7 >> 5) & 1;
	
	int noise, sample;
	uint env;
	uint* p1 = reg[ 8] & 0x10 ? &env : &olevel[0];
	uint* p2 = reg[ 9] & 0x10 ? &env : &olevel[1];
	uint* p3 = reg[10] & 0x10 ? &env : &olevel[2];

	if (!pperiod)
		pcount = p40 & 0x80 ? (0x10000 << SSG_OSR) : 0;
	if (!bperiod)
		bcount = p40 & 0x20 ? (0x10000 << SSG_OSR) : 0;

	#define SCOUNT(ch)	(scount[ch] >> (SSG_TONESHIFT+SSG_OSR))
	
	if (bperiod || pperiod || (p40 & 0xa0))
	{
		for (int i=0; i<nsamples; i++)
		{
			sample = 0;
			for (int j=0; j < (1 << SSG_OSR); j++)
			{
				env = envelop[ecount >> (SSG_ENVSHIFT+SSG_OSR)];
				noise = noisetable[(ncount >> (SSG_NOISESHIFT+SSG_OSR+1)) & 0x7fff];
				sample += 
					  *p1 * (((SCOUNT(0) & chenable[0]) | (nenable[0] & noise)) * 2 - 1)
					+ *p2 * (((SCOUNT(1) & chenable[1]) | (nenable[1] & noise)) * 2 - 1)
					+ *p3 * (((SCOUNT(2) & chenable[2]) | (nenable[2] & noise)) * 2 - 1)
					+ ((pcount | (bcount & bbcount)) & (0x10000 << SSG_OSR) ? 0x2000 : 0x0000);

				scount[0] += speriod[0];
				scount[1] += speriod[1];
				scount[2] += speriod[2];
				ncount += nperiod;
				ecount += eperiod;
				pcount += pperiod;
				bcount += bperiod;
				bbcount += bbperiod;
				if (ecount >= (1 << (SSG_ENVSHIFT+6+SSG_OSR)))
				{
					if ((reg[0x0d] & 0x0b) != 0x0a)
						ecount |= (1 << (SSG_ENVSHIFT+5+ SSG_OSR));
					ecount &= (1 << (SSG_ENVSHIFT+6+SSG_OSR)) - 1;
				}
			}
			*dest++ += sample/(1<<SSG_OSR);
		}
	}
	else
	{
		for (int i=0; i<nsamples; i++)
		{
			sample = 0;
			for (int j=0; j < (1 << SSG_OSR); j++)
			{
				env = envelop[ecount >> (SSG_ENVSHIFT+SSG_OSR)];
				noise = noisetable[(ncount >> (SSG_NOISESHIFT+SSG_OSR+1)) & 0x7fff];

				sample += 
					  *p1 * (((SCOUNT(0) & chenable[0]) | (nenable[0] & noise)) * 2 - 1)
					+ *p2 * (((SCOUNT(1) & chenable[1]) | (nenable[1] & noise)) * 2 - 1)
					+ *p3 * (((SCOUNT(2) & chenable[2]) | (nenable[2] & noise)) * 2 - 1);

				scount[0] += speriod[0];
				scount[1] += speriod[1];
				scount[2] += speriod[2];
				ncount += nperiod;
				ecount += eperiod;

				if (ecount >= (1 << (SSG_ENVSHIFT+6+SSG_OSR)))
				{
					if ((reg[0x0d] & 0x0b) != 0x0a)
						ecount |= (1 << (SSG_ENVSHIFT+5+ SSG_OSR));
					ecount &= (1 << (SSG_ENVSHIFT+6+SSG_OSR)) - 1;
				}
			}
			*dest++ += sample/(1<<SSG_OSR);
		}
	}
}


// ---------------------------------------------------------------------------
//	EmitWave Dual Channel
// ---------------------------------------------------------------------------

void PSGB::EmitWaveDCh(Sample* dest, int nsamples, uint32 bperiod, uint32 pperiod, uint p40)
{
	uint8 chenable[3], nenable[3];
	uint8 r7 = ~reg[7];

	chenable[0] = (r7 & 0x01) && (speriod[0] <= (1 << SSG_TONESHIFT));
	chenable[1] = (r7 & 0x02) && (speriod[1] <= (1 << SSG_TONESHIFT));
	chenable[2] = (r7 & 0x04) && (speriod[2] <= (1 << SSG_TONESHIFT));
	nenable[0]  = (r7 >> 3) & 1;
	nenable[1]  = (r7 >> 4) & 1;
	nenable[2]  = (r7 >> 5) & 1;
	
	int noise, sample;
	uint env;
	uint* p1 = reg[ 8] & 0x10 ? &env : &olevel[0];
	uint* p2 = reg[ 9] & 0x10 ? &env : &olevel[1];
	uint* p3 = reg[10] & 0x10 ? &env : &olevel[2];

	if (!pperiod)
		pcount = p40 & 0x80 ? (0x10000 << SSG_OSR) : 0;
	if (!bperiod)
		bcount = p40 & 0x20 ? (0x10000 << SSG_OSR) : 0;

	#define SCOUNT(ch)	(scount[ch] >> (SSG_TONESHIFT+SSG_OSR))
	
	if (bperiod || pperiod || (p40 & 0xa0))
	{
		for (int i=0; i<nsamples; i++)
		{
			sample = 0;
			for (int j=0; j < (1 << SSG_OSR); j++)
			{
				env = envelop[ecount >> (SSG_ENVSHIFT+SSG_OSR)];
				noise = noisetable[(ncount >> (SSG_NOISESHIFT+SSG_OSR+1)) & 0x7fff];
				sample += 
					  *p1 * (((SCOUNT(0) & chenable[0]) | (nenable[0] & noise)) * 2 - 1)
					+ *p2 * (((SCOUNT(1) & chenable[1]) | (nenable[1] & noise)) * 2 - 1)
					+ *p3 * (((SCOUNT(2) & chenable[2]) | (nenable[2] & noise)) * 2 - 1)
					+ ((pcount | (bcount & bbcount)) & (0x10000 << SSG_OSR) ? 0x2000 : 0x0000);

				scount[0] += speriod[0];
				scount[1] += speriod[1];
				scount[2] += speriod[2];
				ncount += nperiod;
				ecount += eperiod;
				pcount += pperiod;
				bcount += bperiod;
				bbcount += bbperiod;
				if (ecount >= (1 << (SSG_ENVSHIFT+6+SSG_OSR)))
				{
					if ((reg[0x0d] & 0x0b) != 0x0a)
						ecount |= (1 << (SSG_ENVSHIFT+5+ SSG_OSR));
					ecount &= (1 << (SSG_ENVSHIFT+6+SSG_OSR)) - 1;
				}
			}
			*dest++ += sample/(1<<SSG_OSR);
			*dest++ += sample/(1<<SSG_OSR);
		}
	}
	else
	{
		for (int i=0; i<nsamples; i++)
		{
			sample = 0;
			for (int j=0; j < (1 << SSG_OSR); j++)
			{
				env = envelop[ecount >> (SSG_ENVSHIFT+SSG_OSR)];
				noise = noisetable[(ncount >> (SSG_NOISESHIFT+SSG_OSR+1)) & 0x7fff];

				sample += 
					  *p1 * (((SCOUNT(0) & chenable[0]) | (nenable[0] & noise)) * 2 - 1)
					+ *p2 * (((SCOUNT(1) & chenable[1]) | (nenable[1] & noise)) * 2 - 1)
					+ *p3 * (((SCOUNT(2) & chenable[2]) | (nenable[2] & noise)) * 2 - 1);

				scount[0] += speriod[0];
				scount[1] += speriod[1];
				scount[2] += speriod[2];
				ncount += nperiod;
				ecount += eperiod;

				if (ecount >= (1 << (SSG_ENVSHIFT+6+SSG_OSR)))
				{
					if ((reg[0x0d] & 0x0b) != 0x0a)
						ecount |= (1 << (SSG_ENVSHIFT+5+ SSG_OSR));
					ecount &= (1 << (SSG_ENVSHIFT+6+SSG_OSR)) - 1;
				}
			}
			*dest++ += sample/(1<<SSG_OSR);
			*dest++ += sample/(1<<SSG_OSR);
		}
	}
}
