// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1997, 1998
// ---------------------------------------------------------------------------
//  ʐƃOtBbNX̃G~[V
// ---------------------------------------------------------------------------
//	$Id: Screen_i.h,v 1.2 1999/01/11 15:32:13 cisc Exp $

#ifndef SCREEN_INL_H
#define SCREEN_INL_H

// ---------------------------------------------------------------------------
//	Out 30
// ---------------------------------------------------------------------------

inline void Screen::Out30(uint data)
{
	uint i = port30 ^ data;
	port30 = data;
	if (i & 0x02)
	{
		reinitpalette = true;
	}
}

// ---------------------------------------------------------------------------
//	Out 31
// ---------------------------------------------------------------------------
	
inline void Screen::Out31(uint data)
{
	uint i = port31 ^ data;
	port31 = data;
	if (i & 0x19)
	{
		bool l4prev = line400;
		bool cprev = color;
		uint8 dgprev = displaygraphics;
		
		color = !!(data & 0x10);
		line400 = !(data & 0x01) && !color;
		displaygraphics = (data & 8) ? 7 : 0;
		if (l4prev != line400)
			shouldredraw = true;

		if ((cprev != color) || (dgprev != displaygraphics))
			reinitpalette = true;
	}
}

// ---------------------------------------------------------------------------
//	Out 32
// ---------------------------------------------------------------------------

inline void Screen::Out32(uint data)
{
	uint i = port32 ^ data;
	port32 = data;
	if (i & 0x40)
	{
		selectedplane = 3;
		SelectAccessFunc();
	}
	if ((i & 0x20) && !(port31 & 0x10))
	{
		reinitpalette = true;
	}
}

// ---------------------------------------------------------------------------
//	Out 34
// ---------------------------------------------------------------------------

inline void Screen::Out34(uint data)
{
//	LOG3("r:%8d w:%8d     port34: %.2x\n", nreads, nwrites, data);
//	nreads = nwrites = 0;
	for (uint i=0; i<3; i++)
	{
		switch (data & 0x11)
		{
		case 0x00: maskr.byte[i] = 0xff; masks.byte[i] = 0x00; maski.byte[i] = 0x00; break;
		case 0x01: maskr.byte[i] = 0x00; masks.byte[i] = 0xff; maski.byte[i] = 0x00; break;
		case 0x10: maskr.byte[i] = 0x00; masks.byte[i] = 0x00; maski.byte[i] = 0xff; break;
		case 0x11: maskr.byte[i] = 0x00; masks.byte[i] = 0x00; maski.byte[i] = 0x00; break;
		}
		data >>= 1;
	}
}

// ---------------------------------------------------------------------------
//	Out 35
// ---------------------------------------------------------------------------

inline void Screen::Out35(uint data)
{
	port35 = data;
	if (!(port35 & 0x80))
		selectedplane = 3;

	aluread.byte[0] = (data & 1) ? 0xff : 0x00;
	aluread.byte[1] = (data & 2) ? 0xff : 0x00;
	aluread.byte[2] = (data & 4) ? 0xff : 0x00;

	SelectAccessFunc();
//	LOG2("port35: access:%s mux:%d\n", (data & 0x80) ? "GVRAM" : "MAIN ", (data >> 4)&3);
}

// ---------------------------------------------------------------------------
//	Out 5c - 5f
// ---------------------------------------------------------------------------

inline void Screen::Out5cto5f(uint port)
{
	if (!(port32 & 0x40))
	{
		selectedplane = port & 3;
		SelectAccessFunc();
	}
//	LOG1((port == 0x5f) ? "port5f: MAIN RAM\n" : "port5x: GVRAM%d\n", port & 3);
}

inline uint Screen::In5c()
{
	const static uint8 res[4] = { 0xf9, 0xfa, 0xfc, 0xf8 };
	return res[selectedplane];
}
#endif // SCREEN_INL_H
