// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1997, 1998.
// ---------------------------------------------------------------------------
//	Sound Implementation
// ---------------------------------------------------------------------------
//	$Id: Sound.cpp,v 1.3 1999/01/31 01:37:39 cisc Exp $

#include <math.h>
#include <stdio.h>

#include "config.h"
#include "misc.h"
#include "Sound.h"
#include "PC88.h"
#include "file.h"
#include "FMGen.h"

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
	#define SOUND_LOGGING
#endif

#ifdef SOUND_LOGGING
	
#include <windows.h>

	static FILE* fp = 0;
	static int tbase = 0;

	static void LogBegin()   { if (!fp) fp = fopen("SOUND.DMP", "w"); tbase = GetTickCount(); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif

// ---------------------------------------------------------------------------
//	Ej
// ---------------------------------------------------------------------------

Sound88::Sound88()
{
	uscount = 0;
	buffer = 0;
	bwpos = 0;
	enabled = false;
	opn2enabled = false;
	mixingbuf = 0;
	beepmask = 0xa0;

	pindex = 0;
	pswitches = 0;
	bindex = 0;
	bswitches = 0;
	bstep = 0;
	pstep = 0;
	port40 = 0;
}

Sound88::~Sound88()
{
	delete[] mixingbuf;
}

// ---------------------------------------------------------------------------
//	Ƃ
//	rate = SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
// ---------------------------------------------------------------------------

bool Sound88::Init(uint rate, uint32 opnclock, int bufsize, PC88* pc88)
{
	LogBegin();

	enabled = false;
	opn.SetPC(pc88);
	opn2.SetPC(pc88);
	
	if (!psgb.Init(rate))
		return false;
	if (!opn.Init(opnclock, rate, 0))
		return false;
	if (!opn2.Init(opnclock, rate, 0))
		return false;
	
	return SetRate(rate, opnclock, bufsize);
}

// ---------------------------------------------------------------------------
//	[gݒ
//	rate = SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
// ---------------------------------------------------------------------------

bool Sound88::SetRate(uint rate, uint32 opnclock, int bufsize)
{
	bool fm55k = false;
	if (rate == SOUND_55K)
	{
		fm55k = true;
		rate = 44100;
	}
	
	enabled = false;
	samplingrate = rate;
	clock = opnclock;
	
	psgb.Init(rate);
	opn.SetRate(clock, rate, fm55k);
	opn2.SetRate(clock, rate, fm55k);
	psgb.SetPSGClock(clock/4, rate); 
	psg2.SetPSGClock(clock/4, rate); 

	if (!bufsize)
		return false;

	srate50 = rate / 50;
	buffersize = bufsize;
	bvalue = int(2000000.0 * 65536.0 / rate);

	delete[] buffer; delete[] mixingbuf;

	buffer = new uint8[buffersize];
	mixingbuf = new int32[buffersize/2];
	if (!buffer || !mixingbuf)
		return false;

	memset(buffer, 0, buffersize);

	read = 0; write = 0;
	enabled = true;

	return true;
}


// ---------------------------------------------------------------------------
//	Еt
// ---------------------------------------------------------------------------

void Sound88::Cleanup()
{
	LogEnd();
}

// ---------------------------------------------------------------------------
//	Reset
// ---------------------------------------------------------------------------

void Sound88::Reset()
{
	opn.Reset();
	opn2.Reset();
	psgb.Reset();
	psgb.SetPSGClock(clock/4, samplingrate);
	psg2.SetPSGClock(clock/4, samplingrate);
	opn2.EnableIntr(true);
}

// ---------------------------------------------------------------------------
//	Out
// ---------------------------------------------------------------------------

void Sound88::Out(uint port, uint data)
{
	switch (port)
	{
	case 0x32:
//		Log1("p[0x32] = %.2x\n", data);
		opn.EnableIntr(!(data & 0x80));
		break;

	case 0xaa:
		Log1("p[0xaa] = %.2x\n", data);
		if (opn2enabled)
			opn2.EnableIntr(!(data & 0x80));
		break;

	case 0x44:
		regadr1_0 = data;
		if ((regadr1_0 & 0xfc) == 0x2c)
		{
			opn.SetReg(regadr1_0, 0);

			switch (regadr1_0 & 3)
			{
			case 1:
				psgb.SetPSGClock(clock/4, samplingrate); 
				psg2.SetPSGClock(clock/4, samplingrate); 
				break;

			case 2:
				psgb.SetPSGClock(clock/2, samplingrate);
				psg2.SetPSGClock(clock/2, samplingrate);
				break;

			case 3:
				psgb.SetPSGClock(clock/1, samplingrate);
				psg2.SetPSGClock(clock/1, samplingrate);
				break;
			}
		}
		break;

	case 0x45:
		if (regadr1_0 < 0x10)
			psgb.SetReg(regadr1_0, data);
		else
		{
			Log2("reg[0%.2x] = %.2x\n", regadr1_0, data);
			opn.SetReg(regadr1_0, data);
		}
		break;

	case 0x46:
		regadr1_1 = data;
		break;

	case 0x47:
		Log2("reg[1%.2x] = %.2x\n", regadr1_1, data);
		opn.SetReg(0x100 | regadr1_1, data);
		break;

	case 0xa8:
		regadr2_0 = data;
		break;

	case 0xa9:
		if (opn2enabled)
		{
			if (regadr2_0 < 0x10)
				psg2.SetReg(regadr2_0, data);
			else
				opn2.SetReg(regadr2_0, data);
		}
		break;

	case 0xac:
		regadr2_1 = data;
		break;

	case 0xad:
		if (opn2enabled)
			opn2.SetReg(0x100 | regadr2_1, data);
		break;

	}
}

// ---------------------------------------------------------------------------
//	In
// ---------------------------------------------------------------------------

uint8 Sound88::In(uint port)
{
	switch (port)
	{
	case 0x44:
		return opn.ReadStatus();
	
	case 0x45:
		if ((regadr1_0 & 0xfe) == 0x0e)
			return PortIn(regadr1_0 & 1);
		if (regadr1_0 < 0x10)
			return psgb.GetReg(regadr1_0);
		if (regadr1_0 == 0xff && !opn.IsOPNA())
			return 0;
		return opn.GetReg(regadr1_0);

	case 0x46:
		return opn.ReadStatusEx();
	
	case 0x47:
		return opn.GetReg(0x100 | regadr1_1);
	
	// at A8
	case 0xa8:
		return opn2enabled ? opn2.ReadStatus() : 0xff;
	
	case 0xa9:
		if (!opn2enabled)
			return 0xff;
		if (regadr2_0 < 0x10)
			return psg2.GetReg(regadr2_0);
		if (regadr2_0 == 0xff && !opn2.IsOPNA())
			return 0;
		return opn2.GetReg(regadr2_0);

	case 0xac:
		return opn2enabled ? opn2.ReadStatusEx() : 0xff;
	
	case 0xad:
		return opn2enabled ? opn2.GetReg(0x100 | regadr2_1) : 0xff;
	}
	return 0xff;
}

// ---------------------------------------------------------------------------
//	ėp|[g
// ---------------------------------------------------------------------------

uint8 Sound88::PortIn(uint port)
{
	return 0xff;
}

// ---------------------------------------------------------------------------
//	ԏ
// ---------------------------------------------------------------------------
//	todo:
//		pPbgtB

void Sound88::Count(int32 us, bool timer)
{
	uscount += us;
	mixingcount -= us;

	bool c = opn.Count(us) || (opn2enabled && opn2.Count(us));

	if (timer)
		Log0("invoke timer\n");
	if (c || (timer && mixingcount < 0)) 
	{
		int togo = (uscount * srate50 / 10000) & ~1;
		uscount = 0;
		mixingcount = 1000;
		Store(togo * 2);
	}
}

// ---------------------------------------------------------------------------
//	̏܂ł̎Ԃ߂
// ---------------------------------------------------------------------------
	
int32 Sound88::GetNextCount()
{
	if (opn2enabled)
		return Min(opn2.GetNextEvent(), opn.GetNextEvent());
	else
		return opn.GetNextEvent();
}

// ---------------------------------------------------------------------------
//	Oobt@ɉǉ
// ---------------------------------------------------------------------------

void Sound88::Store(int bytes)
{
	if (!enabled)
		return;

	// Oobt@̋󂫂vZ
	int free;
	if (read <= write)
		free = buffersize + read - write;
	else
		free = read - write;

	// ނׂf[^ʂvZ
	bytes = Min(bytes, free-4);
	if (bytes < 0)
		return;
	
//	Log4("read = %6d  write = %6d  free = %6d  bytes = %6d.\n", read, write, free, bytes);

	// 
	if (buffersize - write >= bytes)
	{
		// xŏꍇ
		Mix(&buffer[write], bytes);
	}
	else
	{
		// Qxɕďꍇ
		Mix(&buffer[write], buffersize - write);
		Mix(&buffer[0], bytes - (buffersize - write));
	}
	write += bytes;
	if (write >= buffersize)
		write -= buffersize;
}

// ---------------------------------------------------------------------------
//	Oobt@特Ⴄ
// ---------------------------------------------------------------------------

void Sound88::Fill(void* d, int bytes)
{
	if (!enabled)
		return;

	uint8* dest = (uint8*) d;
	while (bytes > 0)
	{
		int xsize = Min(bytes, buffersize - 4);
		xsize = Min(xsize, buffersize - read);
		
		int avail;
		if (write >= read)
			avail = write - read;
		else
			avail = buffersize + write - read;

		if (xsize > avail)
		{
			Store(xsize - avail);
			uscount = 0;
		}

		memcpy(dest, buffer + read, xsize);
		dest += xsize;
		bytes -= xsize;
		read += xsize;
		if (read >= buffersize)
			read -= buffersize;
	}
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

void Sound88::Mix(void* d, int bytes)
{
	int16* dest = (int16*) d;

	int nsamples = bytes / 4;
	memset(mixingbuf, 0, nsamples * 8);
	opn.Mix(mixingbuf, nsamples);
	psgb.EmitWaveDCh(mixingbuf, nsamples, bstep, pstep, port40);
	if (opn2enabled)
	{
		opn2.Mix(mixingbuf, nsamples);
		psg2.EmitWaveDCh(mixingbuf, nsamples, bstep, pstep, port40);
	}
	int32* src = mixingbuf;
	for (; nsamples>0; nsamples--)
	{
		*dest++ = Limit(*src++, 32767, -32768);
		*dest++ = Limit(*src++, 32767, -32768);
	}
}

// ---------------------------------------------------------------------------
//	BEEP Port ւ Out
// ---------------------------------------------------------------------------

void Sound88::Out40(uint8 data, uint tick)
{
	uint8 modified;

	data &= beepmask;
	
	modified = data ^ port40;
	port40 = data;

	if (modified & 0x80)
	{
		pcount[(pindex++) & 1] = Min(0x10000000, tick - plasttick);
		plasttick = tick;
		pswitches = 10;
	}
	if (modified & 0x20)
	{
		bcount[(bindex++) & 1] = Min(0x10000000, tick - blasttick);
		blasttick = tick;
		bswitches = 10;
	}
}

// ---------------------------------------------------------------------------
//	BEEP pIXV (600Hz)
// ---------------------------------------------------------------------------

void Sound88::UpdateBeepState(uint mhz)
{
	uint level;
	if (pswitches > 0)
	{
		level = (pcount[0] + pcount[1]) / 2 + 1;
		pstep = bvalue * mhz / level;
//		Log3("ps = %8d  pstep = %d  mhz = %d\n", level, pstep, mhz);
		pswitches--;
	}
	else
	{
		pstep = 0;
	}
	if (bswitches > 0)
	{
		level = (bcount[0] + bcount[1]) / 2 + 1;
//		level = (bcount[0] + bcount[1] + bcount[2] + bcount[3]) / 4 + 1;
		bstep = bvalue * mhz / level;
//		Log3("bs = %8d  bstep = %d  mhz = %d\n", level, bstep, mhz);
		bswitches--;
	}
	else
	{
		bstep = 0;
	}
}

// ---------------------------------------------------------------------------
//	CMD SING LE
// ---------------------------------------------------------------------------

void Sound88::EnableSING(bool enable)
{
	beepmask = enable ? 0xa0 : 0x20;
	port40 &= beepmask;
}

void Sound88::EnableOPN2(bool enable)
{
	opn2enabled = enable;
	if (!enable)
		opn2.EnableIntr(false);
}

// ---------------------------------------------------------------------------
//	OPN88::Intr
// ---------------------------------------------------------------------------

void OPN88::Intr(bool flag)
{
	intrpending = flag;
	if (flag && intrenabled)
	{
		Log0("Sound Intr\n");
		if (pc) pc->Intr(4);
	}
}

void OPN88::EnableIntr(bool flag)
{
	if (intrenabled != flag)
	{
		intrenabled = flag;
		if (intrenabled && intrpending)
		{
			Log0("Sound Intr\n");
			if (pc) pc->Intr(4);
		}
	}
}

// ---------------------------------------------------------------------------
//	OPN88::SetReg()
// ---------------------------------------------------------------------------

void OPN88::SetReg(uint regadr, uint data)
{
	if (!opna)
	{
		if (regadr >= 0x100) return;
		if ((regadr & 0xf0) == 0x10) return;
		if (regadr == 0x22) data = 0;
		if (regadr == 0x29) data = 3;
		if (regadr >= 0xb4) data = 0xc0;
	}

	OPNA::SetReg(regadr, data);
}

