// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  TuVXẽG~[V
// ---------------------------------------------------------------------------
//	$Id: SubSys.cpp,v 1.2 1999/01/11 15:32:14 cisc Exp $

#include <stdio.h>
#include <string.h>
#include "SubSys.h"
#include "FDC.h"

// ---------------------------------------------------------------------------
//	for debug

#ifndef RELEASEBUILD
//  #define SUBSYS_LOGGING
#endif

#ifdef SUBSYS_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("subsys.log", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
#endif


// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

SubSystem::SubSystem(uint8* _rom, FDC* _fdc) : rom(_rom), fdc(_fdc)
{
	LogBegin();
	ram = 0;
	piosub.Connect(&piomain);
	piomain.Connect(&piosub);
	Initialize();
	Reset();
}

SubSystem::~SubSystem()
{
	delete[] ram;
	LogEnd();
}

// ---------------------------------------------------------------------------
//	CVXeƂ̃C^[tF[X
// ---------------------------------------------------------------------------

void SubSystem::Out(uint p, uint d)
{
	idlecount = 0;
	reg.haltstate &= ~2;
	if (p == 0xff && d == 0x0f)
		Log0("\ncmd: ");
	if (p == 0xfd)
		Log1("%.2x ", d);
	piomain.Out(p, d);
}

uint8 SubSystem::In(uint p)
{
	uint8 d = piomain.In(p);
	idlecount = 0;
	reg.haltstate &= ~2;
	if (p == 0xfc)
		Log1(">%.2x ", d);
	return d;
}

// ---------------------------------------------------------------------------
//	Reset
// ---------------------------------------------------------------------------

void SubSystem::Reset()
{
	Z80S::Reset(1);
	fdc->Reset();
	Log0("Reset!\n");
	Z80S::Reset(0);
	SetIMask(0xffffffff);
	piosub.Reset();
	piomain.Reset();
	reg.haltstate &= ~2;
	idlecount = 0;
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------
//	TuVXẽ}bv
//	0000-	ROM (2K)
//	0800-	g ROM (2HD @) (6K)
//	2000-	Ȃ
//	4000-	RAM (16K)
//	8000-	Ȃ
//	ߓǂݍݍ̂߂ɃjA\g
//	48KB قǖʂɂȂ邯...

bool SubSystem::Initialize()
{
	// }bv͑f
	gvramselected = false;

	// ̈
	ram = new uint8[0x10100];
	if (!ram) return false;
	
	memcpy(&ram[0],      rom,  0x2000);
	memset(&ram[0x2000], 0xff, 0x2000);
	memset(&ram[0x4000], 0x00, 0x4000);
	memset(&ram[0x8000], 0xff, 0x8000);
	
	// ANZXϐ̏
	for (int i=0; i<0x100; i++)
	{
		readaccess[i] = ram + i * 0x100;
		writeaccess[i] = (i & 0xc0) == 0x40 ? ram + i * 0x100 : ram + 0x10000;
	}

	PatchROM();
	return fdc->Initialize(this);
}

// ---------------------------------------------------------------------------
//	ꉞANZX֐
// ---------------------------------------------------------------------------

uint8 SubSystem::Read(uint addr)
{
	if (addr < 0x2000)
		return rom[addr];
	addr -= 0x4000;
	if (addr < 0x4000)
		return ram[addr];
	return 0xff;
}

void SubSystem::Write(uint addr, uint data)
{
	addr -= 0x4000;
	if (addr < 0x4000)
		ram[addr] = data;
}

// ---------------------------------------------------------------------------
//	o͊֐
//	
//	F4		O	Drive Control		
//	F7		O	Window Port
//	F8		O	Motor Control
//	F8		I	FDC TC in
//	FA		I	FDC Status
//	FB		I/O	FDC	Data
//	FC-FF	i/O	PIO 8255
// ---------------------------------------------------------------------------

void SubSystem::OutPort(uint port, uint data)
{
	idlecount = 0;
	switch (port)
	{
	case 0xf4:
		fdc->DriveControl(data);
		break;
	case 0xf8:
		fdc->MotorControl(data);
		break;
	case 0xfb:
		fdc->SetData(data);
		break;
	case 0xfc: case 0xfd: case 0xfe: case 0xff:
		Log2("<%d %.2x>", port & 3, data);
		piosub.Out(port & 3, data);
		break;
	}
}

uint8 SubSystem::InPort(uint port)
{
	idlecount = (port == 0xfe) ? idlecount + 1 : 0;
	switch (port)
	{
	case 0xf8:
		fdc->TC();
		return 0xff;
	case 0xfa:
		return fdc->Status();
	case 0xfb:
		return fdc->GetData();
	case 0xfc: case 0xfd: case 0xff:
		Log2("(%d %.2x)", port & 3, piosub.In(port & 3));
		return piosub.In(port & 3);
	case 0xfe: 
		return piosub.In(2);
	}
	return 0xff;
}


// ---------------------------------------------------------------------------
//	SubSystem::PatchROM
//	[^[̉]҂ȗpb`𓖂Ă
// ---------------------------------------------------------------------------

void SubSystem::PatchROM()
{
	if (ram[0xfb] == 0xcd && ram[0xfc] == 0xb4 && ram[0xfd] == 0x02)
	{
		ram[0xfb] = ram[0xfc] = ram[0xfd] = 0;
		ram[0x105] = ram[0x106] = ram[0x107] = 0;
	}
}

// ---------------------------------------------------------------------------

void SubSystem::AcceptInterrupt(uint intno)
{
	reg.haltstate &= ~2;
	idlecount = 0;
}
	
