// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  eLXgʂ̕`NX
// ---------------------------------------------------------------------------
//	$Id: Text.cpp,v 1.2 1999/01/11 15:32:14 cisc Exp $

//	T|[g
//		OtBbNX[h o[X̃OtBbNX]

#include <stdio.h>
#include <string.h>
#include "Text.h"
#include "Screen.h"
#include "misc.h"
#include "file.h"
#undef CreateFont

#ifndef RELEASEBUILD
//	#define TEXT_DEBUG
#endif

#ifdef TEXT_DEBUG
	#include <stdio.h>
	#define Log0(a)			if (fp) fprintf(fp, a)
	#define Log1(a, b)		if (fp) fprintf(fp, a, b)
	#define Log2(a, b, c)	if (fp) fprintf(fp, a, b, c)
	#define Log3(a, b, c, d) if (fp) fprintf(fp, a, b, c, d)
	static FILE* fp;
	static void LogBegin()   { if (!fp) fp = fopen("text.log", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
#else
	#define LogBegin()		 ((void) 0)
	#define LogEnd()		 ((void) 0)
	#define Log0(a)			 ((void) 0)
	#define Log1(a, b)		 ((void) 0)
	#define Log2(a, b, c)	 ((void) 0)
	#define Log3(a, b, c, d) ((void) 0)
#endif


// ---------------------------------------------------------------------------
//	Ej
// ---------------------------------------------------------------------------

TextScreen::TextScreen(CRTC& _crtc, const uint8* textfont)
: crtc(_crtc)
{
	LogBegin();

	orgfont = textfont;
	font = new uint8[0x18000];
	pcgram = new uint8[0x400];
	CreateGFont();
	
	bank = 0;
	attr = 0;
	blink = 0;

	vram[0] = new uint8[0x3000];
	vram[1] = vram[0] + 0x1000;
	attrcache = vram[0] + 0x2000;
}

TextScreen::~TextScreen()
{
	delete[] vram[0];
	delete[] font;
	delete[] pcgram;

	LogEnd();
}

// ---------------------------------------------------------------------------
//	UpdateScreen ׂ̈Ɍ݂̉ʂ̏ԂۑB
// ---------------------------------------------------------------------------

void TextScreen::Snap()
{
	crtc.GetTVRAM(vram[bank]);
	state = crtc.GetSetting(&setting);
}

// ---------------------------------------------------------------------------
//	ʂ bpl x nlines ̃C[WɓWJ
//	image	WJ(ɌŒ)
//	nlines	WJC[W̃C (200 only)
//	clock	ms Pʂ݂̌̎
//	region	XV̈n|C^ie Screen::UpdateScreen Ɠ)
// ---------------------------------------------------------------------------

void TextScreen::UpdateScreen
(uint8* image, uint _bpl, uint nlines, ulong clock, int* region, bool refresh)
{
	bpl = _bpl;

	if (state == CRTC::erase)
	{
		ClearText((packed*) image, nlines);
		region[0] = 0, region[1] = 399;
		return;
	}
	if (refresh || fontmodified)
	{
		fontmodified = false;
		ClearText((packed*) image, nlines);
	}

	region[0] = 999, region[1] = -1;
	
	if (state == CRTC::on)		// \XV
	{
		static const uint8 ctype[5] = { 8, 8, 1, 1, 0 };
		cursor_attr = ((setting.cursor_type & 1) && (clock % setting.blinkrate < setting.blinkrate/2))
			? ctype[setting.cursor_type] : 0;
		blinkstate = ((clock % (setting.blinkrate*2)) < setting.blinkrate / 2) ? 2 : 0;
		underlineptr = (setting.linesperchar-1) * bpl;
		
		ExpandImage(image, nlines, region);
	}
}

// ---------------------------------------------------------------------------
//	ʂWJ
// ---------------------------------------------------------------------------

void TextScreen::ExpandImage(uint8 * image, uint nlines, int* region)
{
	uint dy = setting.skipline ? 2 : 1;
	uint ylimit = Min(nlines / setting.linesperchar, setting.height);
	uint8* src = vram[bank];
	const uint8* cache = vram[bank ^= 1];
	uint8* cache_attr = attrcache;
	int ybegin = 100, yend = -1;
	
	int dimage = dy * setting.linesperchar * bpl;
	for (uint y = 0; y < ylimit; y+=dy, image += dimage)
	{
		uint8* dest = image;
		uchar attrflag[256];
		bool linedrawn = false;
		
		attr &= ~0x0c;
		ExpandAttributes(attrflag, src + setting.width, y);
		
		if (!setting.widefont)
		{
			for (uint x = 0; x < setting.width; x++, src++, cache++, dest+=8, cache_attr++)
			{
				uint8 attr = attrflag[x];
				if (*src != *cache || attr != *cache_attr)
				{
					colorp = ATTR_TO_COLORP(attr);
					*cache_attr = attr;
					linedrawn = true;
					PutChar((packed*) dest, *src, attr);
				}
			}
		}
		else
		{
			for (uint x = 0; x < setting.width; x+=2, src+=2, cache+=2, dest+=16, cache_attr+=2)
			{
				uint8 attr = attrflag[x];
				if (*src != *cache || attr != *cache_attr)
				{
					colorp = ATTR_TO_COLORP(attr);
					*cache_attr = attr;
					linedrawn = true;
					PutCharW((packed*) dest, *src, attr);
				}
			}
		}
		if (linedrawn)
		{					
			if (ybegin == 100) ybegin = y;
			yend = y;
		}
		src += setting.attrperline*2, cache += setting.attrperline*2;
	}
	region[0] = setting.linesperchar * ybegin;
	region[1] = setting.linesperchar * (yend + 1) - 1;
}

// ---------------------------------------------------------------------------
//	Agr[gWJ
// ---------------------------------------------------------------------------

void TextScreen::ExpandAttributes(uint8* dest, const uint8* src, uint y)
{
	if (setting.attrperline == 0)
	{
		memset(dest, 0, 80);
		return;
	}
	memset(dest, 0, 80);
	dest[0] = 1;

	uint i;
	uint nattrs = setting.controlcode ? 
			setting.attrperline-1 : setting.attrperline;
	for (i=0; i<nattrs; i++)
		dest[src[i+i]] = 1;

	src++;
	if (setting.cursor_y == y)
	{
		int cx = setting.cursor_x;
		for (i=0; i<setting.width; i++)
		{
			if (dest[i])
				ChangeAttr(*src), src+=2;
			dest[i] = (!cx--) ? attr ^ cursor_attr : attr;
		}
	}
	else
	{
		int cx = setting.cursor_x;
		for (i=0; i<setting.width; i++)
		{
			if (dest[i])
				ChangeAttr(*src), src+=2;
			dest[i] = attr;
		}
	}
}

// ---------------------------------------------------------------------------
//	Agr[gR[h̃tOɕϊ
// ---------------------------------------------------------------------------

inline void TextScreen::ChangeAttr(uint8 code)
{
	if (setting.color)
	{
		if (code & 0x8)
			attr = (attr & 0x0f) | (code & 0xf0);
		else
		{
			attr = (attr & 0xf0) | ((code >> 2) & 0xd) | ((code & 1) << 1);
			if (setting.inversed) attr ^= 1;
			blink = ((code & 2) && !(code & 1)) ? blinkstate : 0;
		}
	}
	else
	{
		attr = 0xe0 | ((code >> 2) & 0x0d) | ((code & 1) << 1) | ((code & 0x80) >> 3);
		if (setting.inversed) attr ^= 1;
		blink = ((code & 2) && !(code & 1)) ? blinkstate : 0;
	}
	attr ^= blink;
}

// ---------------------------------------------------------------------------
//	eLXgtHg\p̃tHg쐬
//	src		pbNꂽtHg(tHg 0x1000)
// ---------------------------------------------------------------------------

void TextScreen::CreateFont()
{
	uint i;
	
	const uint8* src = orgfont;
	uint8* dest = font;
	uint8* destw = font + 0x8000;

	for (i=0; i<256*8; i++)
	{
		uint8 d = *src++;
		for (uint j=0; j<8; j++)
		{
			uint8 b = (d & 0x80) ? TEXT_SET : TEXT_RES;
			*dest++ = b;
			*destw++ = b; *destw++ = b;
			d *= 2;
		}
	}
}

void TextScreen::ModifyFont(uint off, uint d)
{
	uint8* dest = font + 8 * off;
	uint8* destw = font + 0x8000 + 16 * off;
	
	for (uint j=0; j<8; j++)
	{
		uint8 b = (d & 0x80) ? TEXT_SET : TEXT_RES;
		*dest++ = b;
		*destw++ = b; *destw++ = b;
		d *= 2;
	}
	fontmodified = true;
}

const uint8 TextScreen::order[8] = { 0x01, 0x10, 0x02, 0x20, 0x04, 0x40, 0x08, 0x80 };

void TextScreen::CreateGFont()
{
	uint8* dest = font + 0x4000;
	uint8* destw = font + 0x10000;
	
	for (int i=0; i<256; i++)
	{
		for (uint j=0; j<8; j+=2)
		{
			dest [ 0] = dest [ 1] = dest [ 2] = dest [ 3] =
			dest [ 8] = dest [ 9] = dest [10] = dest [11] =
			destw[ 0] = destw[ 1] = destw[ 2] = destw[ 3] =
			destw[ 4] = destw[ 5] = destw[ 6] = destw[ 7] =
			destw[16] = destw[17] = destw[18] = destw[19] =
			destw[20] = destw[21] = destw[22] = destw[23] = 
				i & order[j] ? TEXT_SET : TEXT_RES;
			
			dest [ 4] = dest [ 5] = dest [ 6] = dest [ 7] =
			dest [12] = dest [13] = dest [14] = dest [15] =
			destw[ 8] = destw[ 9] = destw[10] = destw[11] =
			destw[12] = destw[13] = destw[14] = destw[15] =
			destw[24] = destw[25] = destw[26] = destw[27] =
			destw[28] = destw[29] = destw[30] = destw[31] = 
				i & order[j+1] ? TEXT_SET : TEXT_RES;
			
			dest += 16; destw += 32;
		}
	}
}

// ---------------------------------------------------------------------------
//	tHg̃AhX擾
// ---------------------------------------------------------------------------

inline const uint8* TextScreen::GetFont(uint c, bool graphic)
{
	return font + (graphic ? 0x4000 : 0) + c * 64;
}

inline const uint8* TextScreen::GetFontW(uint c, bool graphic)
{
	return font + (graphic ? 0x10000 : 0x8000) + c * 128;
}

// ---------------------------------------------------------------------------
//	eLXg\
// ---------------------------------------------------------------------------

inline void TextScreen::PutChar(packed* dest, uint8 src, uint8 attr)
{
	const packed* font_f = 
		(const packed*) GetFont((attr & 2) ? 0 : src, (attr & 0x10) != 0);
	if (attr & 1)
	{
		PutReversed(dest, font_f);
		PutLineReversed(dest, attr);
	}
	else
	{
		PutNormal(dest, font_f);
		PutLineNormal(dest, attr);
	}
}

inline void TextScreen::PutCharW(packed* dest, uint8 src, uint8 attr)
{
	const packed* font_f = 
		(const packed*) GetFontW((attr & 2) ? 0 : src, (attr & 0x10) != 0);
	if (attr & 1)
	{
		PutReversedW(dest, font_f);
		PutLineReversedW(dest, attr);
	}
	else
	{
		PutNormalW(dest, font_f);
		PutLineNormalW(dest, attr);
	}
}

#define NROW				(bpl/sizeof(packed))
#define DRAW(dest, data)	(dest) = ((dest) & TEXT_MASKP) | (data)
#define DRAWL(dest)			(dest) = ((dest) & ~TEXT_BITP) | TEXT_SETP
#define DRAWLR(dest)		(dest) = ((dest) & ~TEXT_BITP) | TEXT_RESP

void TextScreen::PutNormal(packed * dest, const packed * src)
{
	uint h;

	for (h = 0; h < setting.linecharlimit; h+=2)
	{
		packed x = *src++ | colorp;
		packed y = *src++ | colorp;
		
		DRAW(dest[     0], x);	DRAW(dest[     1], y);
		DRAW(dest[NROW+0], x);	DRAW(dest[NROW+1], y);
		dest += bpl * 2 / sizeof(packed);
	}
	packed p = colorp | TEXT_RESP;
	for (; h < setting.linesperchar; h++)
	{
		DRAW(dest[0], p);	DRAW(dest[1], p);
		dest += bpl / sizeof(packed);
	}
}

void TextScreen::PutNormalW(packed * dest, const packed * src)
{
	uint h;
	
	for (h = 0; h < setting.linecharlimit; h+=2)
	{
		packed v = *src++ | colorp;
		packed w = *src++ | colorp;
		packed x = *src++ | colorp;
		packed y = *src++ | colorp;

		DRAW(dest[     0], v);	DRAW(dest[     1], w);
		DRAW(dest[     2], x);	DRAW(dest[     3], y);
		DRAW(dest[NROW+0], v);	DRAW(dest[NROW+1], w);
		DRAW(dest[NROW+2], x);	DRAW(dest[NROW+3], y);
		dest += bpl * 2 / sizeof(packed);
	}
	packed p = colorp | TEXT_RESP;
	for (; h < setting.linesperchar; h++)
	{
		DRAW(dest[0], p);	DRAW(dest[1], p);
		DRAW(dest[2], p);	DRAW(dest[3], p);
		dest += bpl / sizeof(packed);
	}
}

// ---------------------------------------------------------------------------
//	eLXg]\
// ---------------------------------------------------------------------------

void TextScreen::PutReversed(packed * dest, const packed * src)
{
	uint h;

	for (h = 0; h < setting.linecharlimit; h+=2)
	{
		packed x = (*src++ ^ TEXT_BITP) | colorp;
		packed y = (*src++ ^ TEXT_BITP) | colorp;
		
		DRAW(dest[     0], x);	DRAW(dest[     1], y);
		DRAW(dest[NROW+0], x);	DRAW(dest[NROW+1], y);
		dest += bpl * 2 / sizeof(packed);
	}

	packed p = colorp | TEXT_SETP; 
	for (; h < setting.linesperchar; h++)
	{
		DRAW(dest[     0], p);	DRAW(dest[     1], p);
		dest += bpl / sizeof(packed);
	}
}

void TextScreen::PutReversedW(packed * dest, const packed * src)
{
	uint h;
	
	for (h = 0; h < setting.linecharlimit; h+=2)
	{
		packed v = (*src++ ^ TEXT_BITP) | colorp;
		packed w = (*src++ ^ TEXT_BITP) | colorp;
		packed x = (*src++ ^ TEXT_BITP) | colorp;
		packed y = (*src++ ^ TEXT_BITP) | colorp;
		
		DRAW(dest[     0], v);	DRAW(dest[     1], w);
		DRAW(dest[     2], x);	DRAW(dest[     3], y);
		DRAW(dest[NROW+0], v);	DRAW(dest[NROW+1], w);
		DRAW(dest[NROW+2], x);	DRAW(dest[NROW+3], y);
		dest += bpl * 2 / sizeof(packed);
	}

	packed p = colorp | TEXT_SETP; 
	for (; h < setting.linesperchar; h++)
	{
		DRAW(dest[     0], p);	DRAW(dest[     1], p);
		DRAW(dest[     2], p);	DRAW(dest[     3], p);
		dest += bpl / sizeof(packed);
	}
}

// ---------------------------------------------------------------------------
//	I[o[CAA_[C\
// ---------------------------------------------------------------------------

void TextScreen::PutLineNormal(packed* dest, uint8 attr)
{
	if (attr & 4)		// overline
	{
		DRAWL(dest[0]);		DRAWL(dest[1]);
	}
	if ((attr & 8) && setting.linesperchar > 14)
	{
		dest = (packed*)(((uint8*) dest)+underlineptr);
		DRAWL(dest[0]);		DRAWL(dest[1]);
	}
}

void TextScreen::PutLineReversed(packed* dest, uint8 attr)
{
	if (attr & 4)		// overline
	{
		DRAWLR(dest[0]);	DRAWLR(dest[1]);
	}
	if ((attr & 8) && setting.linesperchar > 14)
	{
		dest = (packed*)(((uint8*) dest)+underlineptr);
		DRAWLR(dest[0]);	DRAWLR(dest[1]);
	}
}

void TextScreen::PutLineNormalW(packed* dest, uint8 attr)
{
	if (attr & 4)		// overline
	{
		DRAWL(dest[0]);		DRAWL(dest[1]);
		DRAWL(dest[2]);		DRAWL(dest[3]);
	}
	if ((attr & 8) && setting.linesperchar > 14)
	{
		dest = (packed*)(((uint8*) dest)+underlineptr);
		DRAWL(dest[0]);		DRAWL(dest[1]);
		DRAWL(dest[2]);		DRAWL(dest[3]);
	}
}

void TextScreen::PutLineReversedW(packed* dest, uint8 attr)
{
	if (attr & 4)		// overline
	{
		DRAWLR(dest[0]);	DRAWLR(dest[1]);
		DRAWLR(dest[2]);	DRAWLR(dest[3]);
	}
	if ((attr & 8) && setting.linesperchar > 14)
	{
		dest = (packed*)(((uint8*) dest)+underlineptr);
		DRAWLR(dest[0]);	DRAWLR(dest[1]);
		DRAWLR(dest[2]);	DRAWLR(dest[3]);
	}
}

// ---------------------------------------------------------------------------
//	eLXgʏ
// ---------------------------------------------------------------------------

void TextScreen::ClearText(packed* dest, uint nlines)
{
	for (uint i = nlines * bpl / 8 / sizeof(packed); i > 0; i--)
	{
		dest[0] = (dest[0] & TEXT_MASKP) | TEXT_RESP;
		dest[1] = (dest[1] & TEXT_MASKP) | TEXT_RESP;
		dest[2] = (dest[2] & TEXT_MASKP) | TEXT_RESP;
		dest[3] = (dest[3] & TEXT_MASKP) | TEXT_RESP;
		dest[4] = (dest[4] & TEXT_MASKP) | TEXT_RESP;
		dest[5] = (dest[5] & TEXT_MASKP) | TEXT_RESP;
		dest[6] = (dest[6] & TEXT_MASKP) | TEXT_RESP;
		dest[7] = (dest[7] & TEXT_MASKP) | TEXT_RESP;
		dest += 8;
	}
	// ׂẴeLXgV[NbgƂƂ
	memset(attrcache, 1, 0x1000);
}

// ---------------------------------------------------------------------------
//	fobO\
// ---------------------------------------------------------------------------

void TextScreen::PutInfo(uint8* image, int bpl, char* infotxt)
{
	int len = Min(80, strlen(infotxt));

	uint8* dest = image + bpl * 384;
	packed color = CRTC::on == crtc.GetSetting(&setting)
		? PACK(0x20) : PACK(0x10);
	for (int i=0; i<len; i++)
	{
		PutInfoChar((packed*) dest, infotxt[i], color);
		dest += 8;
	}
}

void TextScreen::PutInfoChar(packed * dest, uint ch, packed color)
{
	uint h;

	const packed* src = (const packed*) GetFont(ch, false);
	for (h = 0; h < 16; h+=2)
	{
		packed x = (*src++ ^ TEXT_BITP) | color;
		packed y = (*src++ ^ TEXT_BITP) | color;
		
		dest[     0] = x;	dest[     1] = y;
		dest[NROW+0] = x;	dest[NROW+1] = y;
		dest += bpl * 2 / sizeof(packed);
	}
}

// ---------------------------------------------------------------------------
//	PCG I/O
// ---------------------------------------------------------------------------

void TextScreen::Reset()
{
	memcpy(pcgram, orgfont+0x400, 0x400);
	CreateFont();
}

void TextScreen::Out(uint p, uint d)
{
	switch (p)
	{
	case 0:
		pcgdat = d;
		break;
	case 1:
		pcgadr = (pcgadr & 0xff00) | d;
		break;
	case 2:
		pcgadr = (pcgadr & 0x00ff) | (d << 8);
		break;
	}
	if (pcgadr & 0x1000)
	{
		Log2("PCG: %.4x <- %.2x\n", pcgadr, pcgdat);
		pcgram[pcgadr & 0x3ff] = pcgdat;
		if (pcgenable)
			ModifyFont(0x400 + (pcgadr & 0x3ff), pcgdat);
	}
}

void TextScreen::EnablePCG(bool enable)
{
	pcgenable = enable;
	if (!pcgenable)
	{
		CreateFont();
		fontmodified = true;
	}
	else
	{
		for (int i=0; i<0x400; i++)
		{
			ModifyFont(0x400 + i, pcgram[i]);
		}

#ifdef TEXT_DEBUG
		FileIO file("textdump.bin", FileIO::create);

		uint8 buf[0x1000];
		crtc.GetTVRAM(buf);
		file.Write(buf, 0x1000);
#endif
	}
}
