// ---------------------------------------------------------------------------
//  M88 - PC88 emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	User Interface for Win32
//	$Id: Ui.cpp,v 1.3 1999/01/15 12:58:58 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#define _WIN32_IE		0x200

#include <windows.h>
#include <stdio.h>
#include <crtdbg.h>
#include <commdlg.h>
#include <mmsystem.h>
#include <dsound.h>
#include <commctrl.h>
#include "resource.h"
#include "Win88.h"
#include "WinScreen.h"
#include "WinKeyIF.h"
#include "ui.h"
#include "misc.h"
#include "WinSound.h"
#include "M88About.h"
#include "M88Config.h"
#include "M88Config2.h"
#include "File.h"
#include "MSGs.h"

#ifndef RELEASEBUILD
	#define UI_LOGGING
#endif

#ifdef UI_LOGGING
	#include <stdio.h>
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("ui.dmp", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif


// ---------------------------------------------------------------------------

char m88dir[_MAX_PATH];
char m88ini[_MAX_PATH];

// ---------------------------------------------------------------------------
//	InitPathInfo
// ---------------------------------------------------------------------------

static void InitPathInfo()
{
	char buf[_MAX_PATH];
	GetModuleFileName(0, buf, _MAX_PATH);
	
	LPTSTR filepart;
	GetFullPathName(buf, _MAX_PATH, m88dir, &filepart);
	*filepart = 0;

	strcpy(m88ini, m88dir);
	strcat(m88ini, "M88.ini");
}

// ---------------------------------------------------------------------------
//	WinMain
// ---------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hinst, HINSTANCE, LPSTR, int nwinmode)
{
	InitPathInfo();
	InitCommonControls();

//	INITCOMMONCONTROLSEX icc;
//	icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
//	icc.dwICC = ICC_WIN95_CLASSES;
//	InitCommonControlsEx(&icc);

	M88UI ui(hinst);
	if (!ui.InitWindow(nwinmode))
		return -1;
	
	return ui.Main();
}

// ---------------------------------------------------------------------------
//	M88UI 
//	Ej
// ---------------------------------------------------------------------------

M88UI::M88UI(HINSTANCE hinstance)
: hinst(hinstance)
{
	winproc.SetDestination((void*)(WinProcGate), (void*) this);
	timerid = 0;
	screen = 0;
	pc = 0;
	sound = 0;
	point.x = point.y = 0;
	resizewindow = 0;
	displaychangedat = GetTickCount();
	report = true;

	GetDLLAddress();

	for (int i=0; i<2; i++)
	{
		diskinfo[i].hmenu = 0;
		diskinfo[i].ndisks = 0;
		diskinfo[i].currentdisk = 0;
		diskinfo[i].idchgdisk = i ? IDM_CHANGEDISK_2 : IDM_CHANGEDISK_1;
	}
	LogBegin();
}

M88UI::~M88UI()
{
	LogEnd();
}

// ---------------------------------------------------------------------------
//	M88UI::InitM88
//	M88 ̏
// ---------------------------------------------------------------------------

bool M88UI::InitM88()
{
	if (!rom.LoadROM())	return false;

	screen = new WinScreen();
	if (!screen || !screen->Init0(hwnd))
		return false;

	sound = new WinSound();
	if (!sound)
		return false;

	pc = new Win88(&rom, screen, &keyif, sound);
	if (!pc)
		return false;

	pc->Wait(true);

	config.LoadConfig(m88ini, true);

	static const uint srate[] = { 0, 11025, 22050, 44100, SOUND_55K, 48000 };
	sound->Init(hwnd, srate[config.sound], config.opnclock, pc, config.soundbuffer);
	if (!pc->Initalize(config.nerams)) return false;
	if (!screen->Init(pc)) return false;
	if (!keyif.Init(hwnd)) return false;
	
	config.LoadDefaultDirectory(m88ini);

	pc->Reset();
	ApplyConfig();
	pc->Reset();
	
	pc->Wait(false);
	active = true;
	return true;
}

// ---------------------------------------------------------------------------
//	M88UI::CleanupM88
//	M88 ̌Еt
// ---------------------------------------------------------------------------

void M88UI::CleanupM88()
{
	if (pc)
	{
		SaveConfig();
		pc->Wait(true);
	}
	if (screen)
	{
		screen->Cleanup();
		delete screen;
		screen = 0;
	}
	if (sound)
	{
		delete sound;
		sound = 0;
	}
	if (pc)
	{
		delete pc;
		pc = 0;
	}
}

// ---------------------------------------------------------------------------
//	M88UI::InitWindow
//	M88 ̑쐬
// ---------------------------------------------------------------------------

bool M88UI::InitWindow(int nwinmode)
{
	WNDCLASS wcl;
	char* szwinname = "M88forWin_Class";

	wcl.hInstance = hinst;
	wcl.lpszClassName = szwinname;
	wcl.lpfnWndProc = WNDPROC((void*)(winproc));
	wcl.style = 0;
	wcl.hIcon = LoadIcon(hinst, MAKEINTRESOURCE(IDI_ICON1));
	wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcl.cbClsExtra = 0;
	wcl.cbWndExtra = 0;
	wcl.hbrBackground = NULL;
	wcl.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);

	accel = LoadAccelerators(hinst, MAKEINTRESOURCE(IDR_ACC_M88UI));

	if (!RegisterClass(&wcl)) 
		return false;

//	wstyle = WS_POPUP;
	wstyle = WS_CAPTION | WS_OVERLAPPED | WS_SYSMENU;

	hwnd = CreateWindowEx(
		0, //WS_EX_TOPMOST,
		szwinname,
		"M88",
		wstyle,
		CW_USEDEFAULT,		// x
		CW_USEDEFAULT,		// y
		640,				// w
		400,				// h
		NULL,
		NULL, //LoadMenu(hinst, MAKEINTRESOURCE(IDR_MENU1)),	// menu loadmenu
		hinst,
		NULL
		);
	ShowWindow(hwnd, nwinmode);
	UpdateWindow(hwnd);

	HDC hdc = GetDC(hwnd);
	haspalette = !!(GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE);
	ReleaseDC(hwnd, hdc);
	
	return true;
}

// ---------------------------------------------------------------------------
//	M88UI::Main
//	bZ[W[v
// ---------------------------------------------------------------------------

int M88UI::Main()
{
	if (!InitM88())
		PostQuitMessage(1);
	
	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0))
	{
		if (!TranslateAccelerator(msg.hwnd, accel, &msg))
		{
			if ((msg.message == WM_SYSKEYDOWN || msg.message == WM_SYSKEYUP)
				&& !(config.flags & Config::suppressmenu))
				TranslateMessage(&msg);
		}
		DispatchMessage(&msg);
	}

	CleanupM88();
	return msg.wParam;
}


// ---------------------------------------------------------------------------
//	M88UI::WinProc
//	bZ[Wnh
// ---------------------------------------------------------------------------

#define PROC_MSG(msg, func)		case (msg): ret = (func)(hwnd, wp, lp); break

LRESULT M88UI::WinProc(HWND hwnd, UINT umsg, WPARAM wp, LPARAM lp)
{
	uint ret;
	keyif.Disable(true);
	switch (umsg)
	{
	PROC_MSG(WM_COMMAND,			WmCommand);
	PROC_MSG(WM_PAINT,				WmPaint);
	PROC_MSG(WM_CREATE,				WmCreate);
	PROC_MSG(WM_DESTROY,			WmDestroy);
	PROC_MSG(WM_CLOSE,				WmClose);
	PROC_MSG(WM_TIMER,				WmTimer);
	PROC_MSG(WM_ACTIVATE,			WmActivate);
	PROC_MSG(WM_PALETTECHANGED,		WmPaletteChanged);
	PROC_MSG(WM_QUERYNEWPALETTE,	WmQueryNewPalette);
	PROC_MSG(WM_INITMENU,			WmInitMenu);
	PROC_MSG(WM_KEYUP,				WmKeyUp);
	PROC_MSG(WM_KEYDOWN,			WmKeyDown);
	PROC_MSG(WM_SYSKEYUP,			WmSysKeyUp);
	PROC_MSG(WM_SYSKEYDOWN,			WmSysKeyDown);
	PROC_MSG(WM_M88_SENDKEYSTATE,	M88SendKeyState);
	PROC_MSG(WM_ENTERMENULOOP,		WmEnterMenuLoop);
	PROC_MSG(WM_EXITMENULOOP,		WmExitMenuLoop);
//	PROC_MSG(WM_M88_DISPLAYCHANGED,	M88DisplayChanged);
	PROC_MSG(WM_M88_CHANGEDISPLAY,	M88ChangeDisplay);
	PROC_MSG(WM_DISPLAYCHANGE,		WmDisplayChange);
	PROC_MSG(WM_M88_APPLYCONFIG,	M88ApplyConfig);
	PROC_MSG(WM_M88_CHANGEVOLUME,	M88ChangeVolume);
	
	default:
		ret = DefWindowProc(hwnd, umsg, wp, lp);
	}
	keyif.Disable(false);
	return ret;
}

LRESULT _stdcall M88UI::WinProcGate(M88UI* ui, HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	return ui->WinProc(hwnd, umsg, wparam, lparam);
}

// ---------------------------------------------------------------------------
//	M88UI::M88SendKeyState
// ---------------------------------------------------------------------------

inline uint M88UI::M88SendKeyState(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	uint8* dest = reinterpret_cast<uint8*> (wparam);
	GetKeyboardState(dest);
	SetEvent((HANDLE) lparam);
	return 0;
}

inline uint M88UI::WmKeyDown(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if ((uint) wparam == VK_F12 && !(config.flags & Config::disablef12reset))
		;
	else
		keyif.KeyDown((uint) wparam, (uint32) lparam);
	
	return 0;
}

inline uint M88UI::WmKeyUp(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if ((uint) wparam == VK_F12 && !(config.flags & Config::disablef12reset))
		Reset88(config.basicmode);
	else
		keyif.KeyUp((uint) wparam, (uint32) lparam);
	
	return 0;
}

inline uint M88UI::WmSysKeyDown(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if (config.flags & Config::suppressmenu)
	{
		keyif.KeyDown((uint) wparam, (uint32) lparam);
		return 0;
	}
	return DefWindowProc(hwnd, WM_SYSKEYDOWN, wparam, lparam);
}


inline uint M88UI::WmSysKeyUp(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if (config.flags & Config::suppressmenu)
	{
		keyif.KeyUp((uint) wparam, (uint32) lparam);
		return 0;
	}
	return DefWindowProc(hwnd, WM_SYSKEYUP, wparam, lparam);
}

// ---------------------------------------------------------------------------
//	M88UI::WmCommand
//	WM_COMMAND nh
// ---------------------------------------------------------------------------

uint M88UI::WmCommand(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	uint wid = LOWORD(wparam);
	switch (wid)
	{
	case IDM_EXIT:
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		break;

	case IDM_RESET:
		Reset88(config.basicmode);
		break;

	case IDM_LOGSTART:
		if (pc)
			pc->StartMonitor();
		break;

	case IDM_LOGEND:
		if (pc)
			pc->StopMonitor();
		break;

	case IDM_N88V1:
		Reset88(PC88::N88V1);
		break;

	case IDM_N88V1H:
		Reset88(PC88::N88V1H);
		break;

	case IDM_N88V2:
		Reset88(PC88::N88V2);
		break;

	case IDM_NMODE:
		Reset88(PC88::N80);
		break;

	case IDM_DRIVE_1:
	case IDM_CHANGEIMAGE_1:
		ChangeDiskImage(hwnd, 0, IDM_DRIVE_1);
		break;

	case IDM_DRIVE_2:
	case IDM_CHANGEIMAGE_2:
		ChangeDiskImage(hwnd, 1, IDM_DRIVE_2);
		break;

	case IDM_ABOUTM88:
		if (screen) screen->SetGUIFlag(true);
		if (!background) DrawMenuBar(hwnd);
		M88About().Show(hinst, hwnd);
		if (screen) screen->SetGUIFlag(false);
		break;

	case IDM_CONFIG:
		if (screen) screen->SetGUIFlag(true);
		if (!background) DrawMenuBar(hwnd);
		ResizeWindow();
		m88config.Show(hinst, hwnd, &config);
		if (screen) screen->SetGUIFlag(false);
		break;

	case IDM_TOGGLEDISPLAY:
		ToggleDisplayMode();
		break;

	case IDM_CAPTURE:
		CaptureScreen();
		break;

	case IDM_WATCHREGISTER:
		config.flags ^= Config::watchregister;
		ApplyConfig();
		break;

	default:
		if (IDM_CHANGEDISK_1 <= wid && wid < IDM_CHANGEDISK_1 + 64)
		{
			SelectDisk(0, wid - IDM_CHANGEDISK_1);
			break;
		}
		if (IDM_CHANGEDISK_2 <= wid && wid < IDM_CHANGEDISK_2 + 64)
		{
			SelectDisk(1, wid - IDM_CHANGEDISK_2);
			break;
		}
		break;
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmPaint
//	WM_PAINT nh
// ---------------------------------------------------------------------------

uint M88UI::WmPaint(HWND hwnd, WPARAM wp, LPARAM lp)
{
	if (screen)
		screen->RequestPaint(true);
//	ValidateRect(hwnd, NULL);
	return DefWindowProc(hwnd, WM_PAINT, wp, lp);
}

// ---------------------------------------------------------------------------
//	M88UI::WmCreate
//	WM_CREATE nh
// ---------------------------------------------------------------------------

uint M88UI::WmCreate(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	RECT rect;
	rect.left = 0;	rect.right =  640;
	rect.top  = 0;  rect.bottom = 400;
	
	AdjustWindowRectEx(&rect, wstyle, TRUE, 0);
	SetWindowPos(hwnd, 0, 0, 0, rect.right-rect.left, rect.bottom-rect.top,
				 SWP_NOMOVE | SWP_NOZORDER);
	
	timerid = SetTimer(hwnd, 1, 1000, 0);
	Log0("WmCreate\n");

	EnableIME(hwnd, false);
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);

	GetWindowRect(hwnd, &rect);
	point.x = rect.left; point.y = rect.top;

	fullscreen = false;
	ChangeDisplayType();
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmDestroy
//	WM_DESTROY nh
// ---------------------------------------------------------------------------

uint M88UI::WmDestroy(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	PostQuitMessage(0);
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmClose
//	WM_CLOSE nh
// ---------------------------------------------------------------------------

uint M88UI::WmClose(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	// mF
	if (config.flags & Config::askbeforereset)
	{
		if (screen) screen->SetGUIFlag(true);
		int res = MessageBox(hwnd, "M88 I܂", "M88", MB_ICONEXCLAMATION | MB_OKCANCEL | MB_DEFBUTTON2);
		if (screen) screen->SetGUIFlag(false);
		if (res != IDOK)
			return 0;
	}
	
	// fBXN𔲂
	for (int drive=0; drive<2; drive++)
	{
		if (pc && !pc->Unmount(drive))
		{
			if (screen) screen->SetGUIFlag(true);
			MessageBox(hwnd, 
				"fBXNC[W̎舵ɏQ܂",
				"M88",
				MB_ICONERROR | MB_OK);
			if (screen) screen->SetGUIFlag(false);
		}
	}
	
	// 
	KillTimer(hwnd, timerid);
	timerid = 0;
	active = false;

	DestroyWindow(hwnd);
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmTimer
//	WM_TIMER nh
// ---------------------------------------------------------------------------

uint M88UI::WmTimer(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if (wparam == timerid)
	{
		// g,\t[擾
		int	fcount = screen ? screen->GetFCount() : 0;
		int	icount = pc ? pc->GetICount() : 0;
		
		// |[gꍇ̓^Cgo[XV
		if (report)
		{
			if (active)
			{
				char buf[64];
				uint freq = icount / 10000;
				wsprintf(buf, "M88 - %d fps.  %d.%.2d MHz", 
					fcount, freq / 100, freq % 100);
				SetWindowText(hwnd, buf);
			}
			else
				SetWindowText(hwnd, "M88");
		}

		if (resizewindow > 0)
			ResizeWindow();
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmActivate
//	WM_ACTIVATE nh
// ---------------------------------------------------------------------------

uint M88UI::WmActivate(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	background = LOWORD(wparam) == WA_INACTIVE;
	keyif.Activate(!background);
	if (screen)
		screen->QueryNewPalette(background);
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmPaletteChanged
//	WM_PALETTECHANGED nh
// ---------------------------------------------------------------------------

uint M88UI::WmPaletteChanged(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if ((HWND) wparam != hwnd)
	{
		if (screen)
		{
			screen->QueryNewPalette(background);
			return 1;
		}
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmQueryNewPalette
//	WM_QUERYNEWPALETTE nh
// ---------------------------------------------------------------------------

uint M88UI::WmQueryNewPalette(HWND hwnd, WPARAM wparam, LPARAM lparam)
{
	if (screen)
	{
		screen->QueryNewPalette(background);
		return 1;
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmInitMenu
//	WM_INITMENU nh
// ---------------------------------------------------------------------------

uint M88UI::WmInitMenu(HWND hwnd, WPARAM wp, LPARAM lp)
{
	HMENU hmenu = (HMENU) wp;
#ifndef DEBUG_MONITOR
	EnableMenuItem(hmenu, IDM_LOGSTART, MF_GRAYED);
	EnableMenuItem(hmenu, IDM_LOGEND, MF_GRAYED);
#endif
	CheckMenuItem(hmenu, IDM_N88V1, (config.basicmode == PC88::N88V1) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hmenu, IDM_N88V1H,(config.basicmode == PC88::N88V1H)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hmenu, IDM_N88V2, (config.basicmode == PC88::N88V2) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hmenu, IDM_NMODE, (config.basicmode == PC88::N80)   ? MF_CHECKED : MF_UNCHECKED);

	CheckMenuItem(hmenu, IDM_WATCHREGISTER, (config.flags & Config::watchregister) ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hmenu, IDM_SUBCPUCONTROL, (config.flags & Config::subcpucontrol) ? MF_CHECKED : MF_UNCHECKED);

	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::ChangeDiskImage
//	fBXNւ
// ---------------------------------------------------------------------------

void M88UI::ChangeDiskImage(HWND hwnd, int drive, UINT item)
{
	HANDLE hthread = GetCurrentThread();
	int prev = GetThreadPriority(hthread);
	SetThreadPriority(hthread, THREAD_PRIORITY_ABOVE_NORMAL);
	
	if (screen) screen->SetGUIFlag(true);
	if (!background) DrawMenuBar(hwnd);
	if (!pc->Unmount(drive))
	{
		MessageBox(hwnd, 
			"fBXNC[W̎舵ɏQ܂",
			"M88",
			MB_ICONERROR | MB_OK);
	}

	OPENFILENAME ofn;
	memset(&ofn, 0, sizeof(ofn));
	char filename[MAX_PATH];
	filename[0] = 0;
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = "88 Disk Image (*.d88)\0*.d88\0"
					  "All Files (*.*)\0*.*\0";
	ofn.lpstrFile = filename;
	ofn.nMaxFile = MAX_PATH;
	ofn.Flags = OFN_CREATEPROMPT;
	ofn.lpstrDefExt = "d88";
	ofn.lpstrTitle = "Open disk image";
	
	EnableIME(hwnd, true);
	bool isopen = !!GetOpenFileName(&ofn);
	EnableIME(hwnd, false);
	if (screen) screen->SetGUIFlag(false);

	if (isopen)
	{
		char title[MAX_PATH];
		strcpy(title, filename+ofn.nFileOffset);
		
		if (ofn.nFileExtension)
			title[ofn.nFileExtension - ofn.nFileOffset - 1] = 0;
		
		OpenDiskImage(drive, filename, title, ofn.Flags & OFN_READONLY);
	}
	else
		OpenDiskImage(drive, 0, 0, 0);

	SetThreadPriority(hthread, prev);
}

// ---------------------------------------------------------------------------
//	M88UI::SelectDisk
//	fBXNZbg
// ---------------------------------------------------------------------------

bool M88UI::SelectDisk(int drive, int id)
{
	DiskInfo& dinfo = diskinfo[drive];
	if (!pc || drive >= 2 || id >= dinfo.ndisks)
		return false;

	if (dinfo.currentdisk >= 0)
	{
		CheckMenuItem(dinfo.hmenu,
					  dinfo.idchgdisk+dinfo.currentdisk,
					  MF_BYCOMMAND | MF_UNCHECKED);
	}
	if (pc->SetDisk(drive, id))
	{
		CheckMenuItem(dinfo.hmenu,
					  dinfo.idchgdisk+id,
					  MF_BYCOMMAND | MF_CHECKED);
		dinfo.currentdisk = id;
	}
	return true;
}


// ---------------------------------------------------------------------------
//	M88UI::OpenDiskImage
//	fBXNC[WJ
// ---------------------------------------------------------------------------

bool M88UI::OpenDiskImage(int drive, char* filename, char* title, bool readonly)
{
	char buf[MAX_PATH + 16];
	MENUITEMINFO mii;
	memset(&mii, 0, sizeof(mii));
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_SUBMENU;
	mii.fType = MFT_STRING;
	mii.dwTypeData = buf;
	mii.hSubMenu = 0;
	SetMenuItemInfo(GetMenu(hwnd), drive ? IDM_DRIVE_2 : IDM_DRIVE_1, false, &mii);
	
	mii.fMask = MIIM_TYPE  | MIIM_SUBMENU;
	
	DiskInfo& dinfo = diskinfo[drive];

	if (filename)
		dinfo.ndisks = pc->Mount(drive, filename, readonly);
	else
		dinfo.ndisks = 0;

	CreateDiskMenu(drive);

	if (!dinfo.ndisks || !pc->IsMount(drive))
	{
		wsprintf(buf, "Drive &%d...", drive+1);
		mii.hSubMenu = 0;
	}
	else if (dinfo.ndisks == 1 || !dinfo.hmenu)
	{
		wsprintf(buf, "Drive &%d - %s", drive+1, title);
		mii.hSubMenu = 0;
	}
	else // ndisks >= 2
	{
		wsprintf(buf, "Drive &%d - %s", drive+1, title);
		mii.hSubMenu = dinfo.hmenu;
		SelectDisk(drive, 0);
	}
	SetMenuItemInfo(GetMenu(hwnd), drive ? IDM_DRIVE_2 : IDM_DRIVE_1, false, &mii);
	return true;
}

// ---------------------------------------------------------------------------
//	M88UI::CreateDiskMenu
//	}`fBXNC[Wpj[̍쐬
// ---------------------------------------------------------------------------

bool M88UI::CreateDiskMenu(int drive)
{
	DiskInfo& dinfo = diskinfo[drive];
	if (dinfo.hmenu)
	{
		DestroyMenu(dinfo.hmenu);
		dinfo.hmenu = 0;
	}
	dinfo.currentdisk = -1;

	if (dinfo.ndisks < 2)
		return false;

	// j[쐬
	dinfo.hmenu = CreatePopupMenu();
	if (!dinfo.hmenu)
		return false;

	for (int i=0; i<dinfo.ndisks; i++)
	{
		char buf[32];
		char* title = pc->GetDiskTitle(drive, i);
		
		if (!title)
			title = "(???)";
		if (!title[0])
			title = "(no name)";
		
		if (i < 9)
			wsprintf(buf, "&%d %.16s", i+1, title);
		else
			wsprintf(buf, "%d %.16s", i+1, title);
		
		AppendMenu(	dinfo.hmenu,
					MF_STRING,
					dinfo.idchgdisk + i,
					buf );
	}

	AppendMenu(dinfo.hmenu, MF_SEPARATOR, 0, 0);
	AppendMenu(dinfo.hmenu, MF_STRING, dinfo.idchgdisk + 64, "&0 Change Disk");
	SetMenuDefaultItem(dinfo.hmenu, dinfo.idchgdisk + 64, FALSE );

	return true;
}


// ---------------------------------------------------------------------------
//	M88UI::LoadConfig
//	ݒ̓ǂݍ
// ---------------------------------------------------------------------------
	
void M88UI::LoadConfig(char* ininame)
{
	if (ininame)
		config.LoadConfig(ininame, false);
	else
		config.LoadConfig(m88ini, true);

	ApplyConfig();
}

// ---------------------------------------------------------------------------
//	M88UI::SaveConfig
//	ݒ̕ۑ
// ---------------------------------------------------------------------------

void M88UI::SaveConfig(char* ininame)
{
	if (ininame)
		config.SaveConfig(ininame, false);
	else
		config.SaveConfig(m88ini, true);
}


// ---------------------------------------------------------------------------
//	M88UI::ApplyConfig
//	ݒ̔f
// ---------------------------------------------------------------------------

void M88UI::ApplyConfig()
{
	if (pc && screen && sound)
	{
		config.mainsubratio = 
			(config.clock >= 8 || (config.flags & Config::fullspeed)) ? 2 : 1;
		
		pc->ApplyConfig(&config);
		sound->ApplyConfig(&config);
		screen->ApplyConfig(&config);
		keyif.ApplyConfig(&config);
		ChangeDisplayType();

		MENUITEMINFO mii;
		memset(&mii, 0, sizeof(mii));
		mii.cbSize = sizeof(mii);
		mii.fMask = MIIM_TYPE;
		mii.fType = MFT_STRING;
		mii.dwTypeData = (config.flags & Config::disablef12reset) ? "&Reset" : "&Reset\tF12";
		SetMenuItemInfo(GetMenu(hwnd), IDM_RESET, false, &mii);
	}
}

// ---------------------------------------------------------------------------
//	M88UI::Reset88
//	88 Zbg
// ---------------------------------------------------------------------------

void M88UI::Reset88(PC88::BASICMODE newmode)
{
	if (pc)
	{
		if (config.flags & Config::askbeforereset)
		{
			if (screen) screen->SetGUIFlag(true);
			int res = MessageBox(hwnd, "Zbg܂H", "M88", MB_ICONQUESTION | MB_OKCANCEL | MB_DEFBUTTON2);
			if (screen) screen->SetGUIFlag(false);
			if (res != IDOK)
				return;
		}
		config.basicmode = newmode;
		pc->ApplyConfig(&config);
		pc->Reset();
	}
}

// ---------------------------------------------------------------------------
//	M88UI::GetDLLAddress
//	IME ̂߂ API AhX擾
// ---------------------------------------------------------------------------

void M88UI::GetDLLAddress()
{
	HMODULE hmod = ::GetModuleHandle("user32.dll");
	
	penableime = 0;
	if (hmod)
	{
		penableime = PENABLEIME(::GetProcAddress(hmod, "WINNLSEnableIME"));
	}
}

// ---------------------------------------------------------------------------
//	M88UI::EnableIME
//	IME ̐
// ---------------------------------------------------------------------------

bool M88UI::EnableIME(HWND hwnd, BOOL flag)
{
	if (penableime)
	{
		return !!(*penableime)(hwnd, flag);
	}
	return false;
}

// ---------------------------------------------------------------------------
//	M88UI::WmEnterMenuLoop
//	WM_ENTERMENULOOP nh
// ---------------------------------------------------------------------------

uint M88UI::WmEnterMenuLoop(HWND, WPARAM wp, LPARAM)
{
	if (!wp)
	{
		if (screen)
			screen->SetGUIFlag(true);
		if (!background)
			DrawMenuBar(hwnd);
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::WmExitMenuLoop
//	WM_EXITMENULOOP nh
// ---------------------------------------------------------------------------

uint M88UI::WmExitMenuLoop(HWND, WPARAM wp, LPARAM)
{
	if (!wp)
	{
		if (screen)
			screen->SetGUIFlag(false);
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::M88ChangeDisplay
//	\\bh̕ύX
// ---------------------------------------------------------------------------

uint M88UI::M88ChangeDisplay(HWND hwnd, WPARAM mode, LPARAM)
{
	if (!screen || !screen->ChangeDisplayDriver((WinScreen::DisplayType) mode))
		return 0;
	
	switch ((WinScreen::DisplayType) mode)
	{
		RECT rect;
	
	case WinScreen::GDI:
	case WinScreen::DDWin:
		wstyle = (GetWindowLong(hwnd, GWL_STYLE) & ~WS_POPUP) 
			   | (WS_CAPTION | WS_OVERLAPPED | WS_SYSMENU);
		SetWindowLong(hwnd, GWL_STYLE, wstyle);
		SetWindowLong(hwnd, GWL_EXSTYLE, GetWindowLong(hwnd, GWL_EXSTYLE) & ~WS_EX_TOPMOST);
		
		rect.left = 0;	rect.right =  640;
		rect.top  = 0;  rect.bottom = 400;
		AdjustWindowRectEx(&rect, wstyle, TRUE, 0);
	
		SetWindowPos(hwnd, 0, 
					 point.x, point.y, 
					 rect.right - rect.left, rect.bottom - rect.top,
					 SWP_NOZORDER);
		
		report = true;
		break;

	case WinScreen::DDFull:
		wstyle = (GetWindowLong(hwnd, GWL_STYLE) & ~(WS_CAPTION | WS_OVERLAPPED | WS_SYSMENU)) 
					| WS_POPUP;
		SetWindowLong(hwnd, GWL_STYLE, wstyle);
		SetWindowLong(hwnd, GWL_EXSTYLE, GetWindowLong(hwnd, GWL_EXSTYLE) | WS_EX_TOPMOST);
		report = false;
		SetWindowText(hwnd, "M88");
		break;
	}

	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::ToggleDisplayMode
//	SʁEEBhE\ؑ  (ALT+ENTER)
// ---------------------------------------------------------------------------

void M88UI::ToggleDisplayMode()
{
	uint tick = GetTickCount();
	if ((tick - displaychangedat) < 1000)
		return;

	displaychangedat = GetTickCount();
	fullscreen = !fullscreen;
	
	ChangeDisplayType(fullscreen);
}

// ---------------------------------------------------------------------------
//	M88UI::WmDisplayChange
//	WM_DISPLAYCHANGE nh
// ---------------------------------------------------------------------------

uint M88UI::WmDisplayChange(HWND hwnd, WPARAM wp, LPARAM lp)
{
	if (!fullscreen)
	{
		resizewindow = 5;
	}
	return DefWindowProc(hwnd, WM_DISPLAYCHANGE, wp, lp);
}

void M88UI::ResizeWindow()
{
	if (!fullscreen)
	{
		wstyle = (GetWindowLong(hwnd, GWL_STYLE) & ~(WS_POPUP | WS_MAXIMIZE)) 
			   | (WS_CAPTION | WS_OVERLAPPED | WS_SYSMENU);
		SetWindowLong(hwnd, GWL_STYLE, wstyle);
		SetWindowLong(hwnd, GWL_EXSTYLE, GetWindowLong(hwnd, GWL_EXSTYLE) & ~WS_EX_TOPMOST);
		
		RECT rect, rect2;
		rect.left = 0;	rect.right =  640;
		rect.top  = 0;  rect.bottom = 400;
		AdjustWindowRectEx(&rect, wstyle, TRUE, 0);

		SetWindowPos(hwnd, 0, 0, 0,
					 rect.right - rect.left, rect.bottom - rect.top,
					 SWP_NOMOVE | SWP_NOZORDER);

		GetWindowRect(hwnd, &rect2);

		if (--resizewindow < 0)
			resizewindow = 0;
	}
}

// ---------------------------------------------------------------------------
//	M88UI::ChangeDisplayType
//	\\bhύX
// ---------------------------------------------------------------------------

void M88UI::ChangeDisplayType(bool savepos)
{
	if (savepos)
	{
		RECT rect;
		GetWindowRect(hwnd, &rect);
		point.x = rect.left; point.y = rect.top;
	}
	if (!fullscreen)
	{
		if (haspalette /* || (config.flags & Config::ddonwin) */)
			PostMessage(hwnd, WM_M88_CHANGEDISPLAY, WinScreen::DDWin, 0);
		else
			PostMessage(hwnd, WM_M88_CHANGEDISPLAY, WinScreen::GDI, 0);
	}
	else
		PostMessage(hwnd, WM_M88_CHANGEDISPLAY, WinScreen::DDFull, 0);
}

// ---------------------------------------------------------------------------
//	M88UI::ApplyConfig
//	Config K
// ---------------------------------------------------------------------------

uint M88UI::M88ApplyConfig(HWND, WPARAM newconfig, LPARAM)
{
	if (newconfig)
	{
		// \łȁB
		if (memcmp(&config, (Config*)newconfig, sizeof(Config)))
		{
			config = *((Config*)newconfig);
			ApplyConfig();
		}
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::M88ChangeVolume
// ---------------------------------------------------------------------------

uint M88UI::M88ChangeVolume(HWND, WPARAM c, LPARAM)
{
	if (c && sound)
	{
		sound->SetVolume((Config*) c);
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	M88UI::CaptureScreen
// ---------------------------------------------------------------------------

void M88UI::CaptureScreen()
{
	if (!screen) return;

	const int bmpsize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO) 
					  + 15 * sizeof(RGBQUAD) + 320 * 400;
	uint8* bmp = new uint8[bmpsize];
	if (!bmp) return;

	if (!screen->Capture(bmp))
	{
		delete[] bmp;
		return;
	}
	
	screen->SetGUIFlag(true);
	
	OPENFILENAME ofn;
	memset(&ofn, 0, sizeof(ofn));
	char filename[MAX_PATH];
	filename[0] = 0;
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = "bitmap image [4bpp] (*.bmp)\0*.bmp\0";
	ofn.lpstrFile = filename;
	ofn.nMaxFile = MAX_PATH;
	ofn.Flags = OFN_CREATEPROMPT;
	ofn.lpstrDefExt = "bmp";
	ofn.lpstrTitle = "Save captured image";
	
	EnableIME(hwnd, true);
	bool save = !!GetSaveFileName(&ofn);
	EnableIME(hwnd, false);
	if (screen) screen->SetGUIFlag(false);

	if (save)
	{
		FileIO file;
		if (file.Open(filename, FileIO::create))
			file.Write(bmp, bmpsize);
	}
	delete[] bmp;
}
