// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	PC88 Hardware Emulation for Win32
// ---------------------------------------------------------------------------
//	$Id: Win88.cpp,v 1.3 1999/01/31 10:49:20 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <process.h>
#include <crtdbg.h>
#include <mmsystem.h>
#include <dsound.h>
#include "Win88.h"
#include "WinScreen.h"
#include "WinSound.h"
#include "WinFDC.h"
#include "SubSys.h"
#include "misc.h"
#include "Text.h"
#include "WinKeyIF.h"

extern char m88dir[];


// ---------------------------------------------------------------------------
//	萔
// ---------------------------------------------------------------------------

#define FV_24	57					// g
#define FV_15	62	

// ---------------------------------------------------------------------------
//	for debug
// ---------------------------------------------------------------------------

#ifndef RELEASEBUILD
	#define WIN88_LOGGING
#endif

#ifdef WIN88_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("win88.log", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

Win88::Win88(ROM* _rom, WinScreen* _screen, WinKeyIF* keyb, WinSound* ws)
:	PC88(_rom, _screen, keyb, &calender, ws), sound(ws), 
	winscreen(_screen), winkeyb(keyb)
{
	winfdc = 0;
	fdsub = 0;
	LogBegin();

	SetTimeMode(true);
}

bool Win88::Initalize(int nerams)
{
	if (!Init(nerams))
		return false;

	winfdc = new WinFDC();
	fdsub = new SubSystem(rom->disk, winfdc);
	
	if (!winfdc->Init())
		return false;

	wait = true;
	shouldterminate = false;
	hthread = HANDLE(_beginthreadex(NULL, 0, ThreadEntry,
									reinterpret_cast<LPVOID>(this), 0, &idthread));
	if (!hthread)
		return false;
	
	vsynccount = 0;
	timercount = 0;
	refreshcount = 0;
	
	config.flags = 0;
	config.refreshtiming = 2;
	config.clock = 8;
	config.mainsubratio = 2;
	config.basicmode = N88V2;
	
	char buf[_MAX_PATH+10];
	strcpy(buf, m88dir);
	strcat(buf, "m88.bin");
	winfdc->Mount(0, buf, true);
	return true;
}


Win88::~Win88()
{
	LogEnd();
	if (hthread)
	{
		shouldterminate = true;
		if (WAIT_TIMEOUT == WaitForSingleObject(hthread, 3000))
			TerminateThread(hthread, 0);
		CloseHandle(hthread);
	}
	_RPT0(_CRT_WARN, "Win88 I\n");
}

// ---------------------------------------------------------------------------
//	Zbg
// ---------------------------------------------------------------------------

void Win88::Reset()
{
	CriticalSection::Lock lock(cs_cpu);
	Log0("Reset\n");

	vsynctime1 = 100000 / (line200 ? FV_15 : FV_24);
	vsynctime2 = 900000 / (line200 ? FV_15 : FV_24);
	sw6e = (sw6e & 0x7f) | ((!config.clock || Abs(config.clock) >= 6) ? 0 : 0x80);
	sw30 = (sw30 & 0xc1) | ((config.dipsw << 1) & 0x3e);
	sw31 = (sw31 & 0xc0) | ((config.dipsw >> 5) & 0x3f);
	fdboot = !(config.dipsw & (1 << 11));
	
	PC88::Reset(config.basicmode);
}

// ---------------------------------------------------------------------------
//	CPU Xbh
// ---------------------------------------------------------------------------

uint Win88::ThreadMain()
{
	DWORD time = GetTime();
	vsynccount = 1;
	timercount = 1;
	int cpuclock = 10;			// NbN in MHz
	icount = 0;
	refreshcount = 1;

//	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);

	while (!shouldterminate)
	{
		if (!wait)
		{
			if (!config.clock)
			{
				time = GetTime();
				DWORD ms = 0;
				int eclk = 0;
				do
				{
					Execute(cpuclock, 5 * config.speed);
					eclk+=5;
					ms = GetTime() - time;
				} while (ms < 10);

				// ̎NbNvZ  A 1GHz? (^^;
				cpuclock = Min((Min(100, eclk) * cpuclock / ms) + 1, 1000);
			}
			else
			{
				if (config.clock > 0)
				{
//					Log2("time = %8d   GetTime = %8d  ", time, GetTime());
					int dt = GetTime() - time;
					int ms = Limit(dt, 100, 4);				// max s
					time += dt;								// time  I\莞
					Execute(config.clock, config.speed * ms);
					
					int idletime = time - GetTime() + 10;	// idletime  ]莞
					if (idletime > 0)
						Sleep(idletime);
//					Log3("dt = %6d  ms = %6d  idletime = %6d\n", dt, ms, Max(0, idletime));
				}
				else
				{
					Execute(-config.clock, 5000);
				}
			}
		}
		else
		{
			Sleep(5);
			time = GetTime();
		}
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	botC[v
//	mcps	bot̃NbN(MHz)
//	ustodo	s鎞
// ---------------------------------------------------------------------------

//	fminterval-fmcount us

int32 Win88::Execute(int32 mcps, int32 ustodo)
{
	CriticalSection::Lock lock(cs_cpu);

	if (!wait)
	{
		icount += ustodo * mcps;
		for (int32 timeremain = ustodo; timeremain > 0; )
		{
			int32 soundcount = sound->GetNextCount();
			int32 us = Min(Min(vsynccount, timercount), Min(timeremain, soundcount));
			us = Max(10, us);
			
			vsynccount -= us;
			timercount -= us;
			timeremain -= us;

//			Log1("\n%5d ", us * mcps);

			if (vsynccount <= 0)
			{
				if (!vrtc)
				{
#if 1 && defined(WIN88_LOGGING)
					static int prevv;
					int cv = GetTime();
					Log1("vsync = %d\n", cv - prevv);
					prevv = cv;
#endif
					vrtc = true;
					vsynccount += vsynctime1;
					Intr(1);
					winkeyb->VSync();
					sound->VSync();
					SetWaitTable();
					if (--refreshcount <= 0)
					{
						refreshcount = config.refreshtiming;
						winscreen->RequestPaint();
					}
				}
				else
				{
//					Log0("vsyncc  ");
					vrtc = false;
					SetWaitTable();
//					IntrClear(1);
					vsynccount += vsynctime2;
				}
			}

			bool rtcon = false;
			if (timercount <= 0)
			{
				if (!rtc)
				{
					rtcon = true;
					rtc = true;
					Intr(2);
//					Log1("timer(%d) ", timercount);
					timercount += 1000000 / 1200;
					sound->UpdateBeepState(mcps);
				}
				else
				{
					rtc = false;
					timercount += 1000000 / 1200;
				}
			}
			sound->Count(us, rtcon);
			
			bool subidle = fdsub->IsIdle(0 != (config.flags & Config::subcpucontrol));
			if (config.flags & Config::cpuclockmode)
			{
				if (config.mainsubratio == 1)
					StepDualP(mcps*us, fdsub, 24, 24);
				else
					StepDualP(mcps*us, fdsub, 48, 24);
			}
			else
			{
				if (subidle)
					Step(mcps*us);
				else
				{
					if (config.mainsubratio == 1)
						StepDual(mcps*us, fdsub);
					else
						StepDual2(mcps*us, fdsub);
				}
			}
		}
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	sNbNJEg̒lԂAJE^Zbg
// ---------------------------------------------------------------------------

int Win88::GetICount()
{
	int i = icount;
	icount = 0;
	return i;
}


// ---------------------------------------------------------------------------
//	TuXbhJn_
// ---------------------------------------------------------------------------

uint __stdcall Win88::ThreadEntry(void * arg)
{
	Win88* w88 = reinterpret_cast<Win88*>(arg);
	if (w88)
		return w88->ThreadMain();
	return 0;
}

// ---------------------------------------------------------------------------
//	s~邩H
// ---------------------------------------------------------------------------

void Win88::Wait(bool dowait)
{
	CriticalSection::Lock lock(cs_cpu);
	wait = dowait;
}

// ---------------------------------------------------------------------------
//	ݒ
// ---------------------------------------------------------------------------

void Win88::ApplyConfig(const Config* newconfig)
{
	CriticalSection::Lock lock(cs_cpu);
	config = *newconfig;
	if (config.flags & Config::fullspeed)
		config.clock = 0;
	if (config.flags & Config::cpuburst)
		config.clock = -config.clock;
	textscreen->EnablePCG(0 != (config.flags & Config::enablepcg));
	line200 = 0 != (config.flags & Config::fv15k);
	SetTimeMode(0 != (config.flags & Config::useqpc));
}

// ---------------------------------------------------------------------------

void Win88::AcceptInterrupt(uint intno)
{
	PC88::AcceptInterrupt(intno);
//	Log1(" [int%d]", intno/2);
}

// ---------------------------------------------------------------------------

void Win88::StartMonitor()
{
	Z80::StartMonitor();
	fdsub->StartMonitor();
}

void Win88::StopMonitor()
{
	Z80::StopMonitor();
	fdsub->StopMonitor();
}

// ---------------------------------------------------------------------------

void Win88::UpdateScreen
(uint8* image, int bpl, ulong clock, int* region, bool info, bool refresh)
{
	PC88::UpdateScreen(image, bpl, clock, region, refresh);
	
	if (info)
	{
		char buf[100];

#if 1 || defined(RELEASEBUILD)
		wsprintf(buf, "PC:%.4x AF:%.4x H:%.4x D:%.4x B:%.4x S:%.4x X:%.4x Y:%.4x FDD:%2d PC:%.4x",
				reg.pc & 0xffff, reg.r.w.af & 0xffff,
				reg.r.w.hl & 0xffff, reg.r.w.de & 0xffff,
				reg.r.w.bc & 0xffff, reg.r.w.sp & 0xffff,
				reg.r.w.ix & 0xffff, reg.r.w.iy & 0xffff,
				fdsub->IdleCount(), fdsub->GetPC()
				);
#else
		wsprintf(buf, "%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x",
				 InPort(0), InPort(1), InPort(2), InPort(3), 
				 InPort(4), InPort(5), InPort(6), InPort(7), 
				 InPort(8), InPort(9), InPort(10), InPort(11),
				 InPort(12), InPort(13), InPort(14), InPort(15));
#endif
		textscreen->PutInfo(image, bpl, buf);
		region[0] = Min(384, region[0]);
		region[1] = 400;
	}
}

// ---------------------------------------------------------------------------
//	Disk I/F
// ---------------------------------------------------------------------------

int Win88::Mount(uint drive, const char* filename, bool readonly)
{
	return winfdc->Mount(drive, filename, readonly);
}

bool Win88::Unmount(uint drive)
{
	return winfdc->Unmount(drive);
}

bool Win88::IsMount(uint drive)
{
	return winfdc->IsMount(drive);
}

bool Win88::SetDisk(uint drive, uint id)
{
	return winfdc->SetDisk(drive, id);
}

char* Win88::GetDiskTitle(uint drive, uint id)
{
	return winfdc->GetDiskTitle(drive, id);
}

// ---------------------------------------------------------------------------
//	Ԃ擾
// ---------------------------------------------------------------------------

uint32 Win88::GetTime()
{
	uint32 p = precision;
	if (p)
	{
		LARGE_INTEGER count;
		QueryPerformanceCounter(&count);
		uint32 r;

		__asm { mov edx,count.u.HighPart }
		__asm { mov eax,count.u.LowPart }
		__asm { div p }
		__asm { mov r,eax }
		return r;
	}
	else
	{
		return GetTickCount();
	}
}

// ---------------------------------------------------------------------------
//	Ԏ擾̕@ݒ
// ---------------------------------------------------------------------------

void Win88::SetTimeMode(bool useqpc)
{
	LARGE_INTEGER freq;

	if (useqpc && QueryPerformanceFrequency(&freq) && freq.u.HighPart < 1000)
	{
		uint r;
		__asm { mov edx,freq.u.HighPart }
		__asm { mov eax,freq.u.LowPart }
		__asm { mov ecx,1000 }
		__asm { div ecx }
		__asm { mov r,eax }
		precision = r;
	}
	else
		precision = 0;
	Log1("precision = %8d\n", precision);
}

