// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	FDC obNOEhp
// ---------------------------------------------------------------------------
//	$Id: WinFDC.cpp,v 1.2 1999/01/11 15:32:16 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <crtdbg.h>
#include <stdlib.h>
#include <process.h>

#include "WinFDC.h"
#include "WinFDU.h"

// ---------------------------------------------------------------------------
//	bZ[W
// ---------------------------------------------------------------------------

#define UM_FDCOP		(WM_USER + 0x100)
#define FDCOP_READDIAG	0x02
#define FDCOP_WRITEDATA	0x05
#define FDCOP_READDATA	0x06
#define FDCOP_READID	0x0a
#define FDCOP_WRITEID	0x0d
#define FDCOP_SEEK		0x0f
#define FDCOP_TC		0xff

// ---------------------------------------------------------------------------
//	Ej
// ---------------------------------------------------------------------------

WinFDC::WinFDC()
{
	hthread = 0;
	threaded = true;
}

WinFDC::~WinFDC()
{

}

// ---------------------------------------------------------------------------
//	En
// ---------------------------------------------------------------------------

bool WinFDC::Init()
{
	hthread = HANDLE(_beginthreadex(NULL, 0, ThreadEntry, 
					   reinterpret_cast<LPVOID>(this), 0, &idthread));
	if (!hthread)
		return false;
	while (!PostThreadMessage(idthread, 0, 0, 0))
		Sleep(10);

	for (int i=0; i<NUM_FLOPPY_DRIVES; i++)
	{
		if (!(fdu[i] = winfdu[i] = new WinFDU))
			return false;
	}

	return true;
}

void WinFDC::Cleanup()
{
	if (hthread)
	{
		int i=300;
		do
		{
			PostThreadMessage(idthread, WM_QUIT, 0, 0);
		} while (WAIT_TIMEOUT == WaitForSingleObject(hthread, 10) && --i);
		
		if (!i)
		{
			TerminateThread(hthread, 0);
		}
		CloseHandle(hthread);
		hthread = 0;
	}
}

// ---------------------------------------------------------------------------
//	fBXNւ֌W
// ---------------------------------------------------------------------------

int WinFDC::Mount(uint drive, const char* filename, bool readonly)
{
	if (drive < NUM_FLOPPY_DRIVES)
	{	
		CriticalSection::Lock lock(cs);
		return winfdu[drive]->Mount(filename, readonly);
	}
	return 0;
}

bool WinFDC::SetDisk(uint drive, uint num)
{
	if (drive < NUM_FLOPPY_DRIVES)
	{	
		CriticalSection::Lock lock(cs);
		return winfdu[drive]->SetDisk(num);
	}
	return false;
}

bool WinFDC::Unmount(uint drive)
{
	if (drive < NUM_FLOPPY_DRIVES)
	{
		CriticalSection::Lock lock(cs);
		return winfdu[drive]->Unmount();
	}
	return false;
}

bool WinFDC::IsMount(uint drive)
{
	if (drive < NUM_FLOPPY_DRIVES)
	{
		return winfdu[drive]->IsMount();
	}
	return false;
}

char* WinFDC::GetDiskTitle(uint drive, uint id)
{
	if (drive < NUM_FLOPPY_DRIVES)
	{
		return winfdu[drive]->GetDiskTitle(id);
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	XbhGg
// ---------------------------------------------------------------------------

uint __stdcall WinFDC::ThreadEntry(LPVOID arg)
{
	WinFDC* instance = reinterpret_cast<WinFDC*> (arg);
	if (instance)
		return instance->ThreadMain();
	return 0;
}

// ---------------------------------------------------------------------------
//	XbhC[v
// ---------------------------------------------------------------------------

uint WinFDC::ThreadMain()
{
	MSG msg;
//	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);
	
	while (TRUE == GetMessage(&msg, NULL, 0, 0))
	{
		if (msg.message == UM_FDCOP)
		{
//			Sleep(50);
			CriticalSection::Lock lock(cs);

			switch (msg.wParam)
			{
			case FDCOP_READDATA:
				FDC::ReadData(msg.lParam != 0);
				break;
			
			case FDCOP_WRITEDATA:
				FDC::WriteData((1 & msg.lParam)!=0, (2 & msg.lParam)!=0);
				break;

			case FDCOP_READID:
				FDC::ReadID();
				break;

			case FDCOP_SEEK:
				FDC::Seek((msg.lParam >> 8) & 0xff, msg.lParam & 0xff);
				break;

			case FDCOP_TC:
				FDC::TC();
				break;

			case FDCOP_READDIAG:
				FDC::ReadDiag(0 != msg.lParam);
				break;

			case FDCOP_WRITEID:
				FDC::WriteID();
				break;
			}
		}
	}
	return 0;
}

void WinFDC::Reset()
{
	CriticalSection::Lock lock(cs);
	FDC::Reset();
}

// ---------------------------------------------------------------------------
//	WinFDC::ReadData
// ---------------------------------------------------------------------------

void WinFDC::ReadData(bool deleted)
{
	if (!threaded)
		FDC::ReadData(deleted);
	else
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_READDATA, deleted);
}

// ---------------------------------------------------------------------------
//	WinFDC::WriteData
// ---------------------------------------------------------------------------

void WinFDC::WriteData(bool last, bool deleted)
{
	if (!threaded)
		FDC::WriteData(last, deleted);
	else
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_WRITEDATA, last + (deleted ? 2 : 0));
}

// ---------------------------------------------------------------------------
//	WinFDC::ReadID
// ---------------------------------------------------------------------------

void WinFDC::ReadID()
{
	if (!threaded)
		FDC::ReadID();
	else
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_READID, 0);
}

// ---------------------------------------------------------------------------
//	WinFDC::Seek
// ---------------------------------------------------------------------------

void WinFDC::Seek(uint dr, uint cy)
{
	if (!threaded)
		FDC::Seek(dr, cy);
	else
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_SEEK, (dr << 8) + cy);
}

// ---------------------------------------------------------------------------
//	WinFDC::ReadDiag
// ---------------------------------------------------------------------------

void WinFDC::ReadDiag(bool first)
{
	if (!threaded)
		FDC::ReadDiag(first);
	else
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_READDIAG, first);
}

// ---------------------------------------------------------------------------
//	WinFDC::WriteID
// ---------------------------------------------------------------------------

void WinFDC::WriteID()
{
//	if (!threaded)
		FDC::WriteID();
//	else
//		PostThreadMessage(idthread, UM_FDCOP, FDCOP_WRITEID, 0);
}

// ---------------------------------------------------------------------------
//	WinFDC::TC
// ---------------------------------------------------------------------------

void WinFDC::TC()
{
	if (!threaded)
		FDC::TC();
	else
	{
		CriticalSection::Lock lock(csphase);
		tcin = true;
		Txt("TCin");
		status &= ~(S_RQM | S_DIO);
		PostThreadMessage(idthread, UM_FDCOP, FDCOP_TC, 0);
	}
}

// ---------------------------------------------------------------------------
//	WinFDC::ShiftPhase
// ---------------------------------------------------------------------------

bool WinFDC::ShiftPhase(Phase p, uint8 s, bool t)
{
	CriticalSection::Lock lock(csphase);
	return FDC::ShiftPhase(p, s, t);
}
