// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator.
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//  FDU I@\t
// ---------------------------------------------------------------------------
//	$Id: WinFDU.cpp,v 1.2 1999/01/11 15:32:17 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "WinFDU.h"
#include "FDU0.h"
#include "FDUB.h"

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

WinFDU::WinFDU()
{
	fdu = 0;
	cy = 0;
}

WinFDU::~WinFDU()
{
	delete fdu;
}

// ---------------------------------------------------------------------------
//	fBXNւ֌W
// ---------------------------------------------------------------------------

int WinFDU::Mount(const char* filename, bool readonly)
{
	Unmount();					// Unmount() locks cs

	if (!readonly && file.CreateNew(filename))
	{
		// P88SR ō쐬
		char buf[0x2b0];
		memset(buf, 0, 0x2b0);
		*(int*)(buf+0x1c) = 0x2b0;
		file.Write(buf, 0x2b0);
		file.Close();
	}

	// }`C[Wۂ𒲂ׂ
	
	if (CheckMultiDiskImage(filename))
		readonly = true;

	// t@CJ

	if (!(file.Open(filename, readonly ? FileIO::readonly : 0)))
	{
		if (!(file.Open(filename, FileIO::readonly)))
			return 0;
	}
	
	if (SetDisk(0))
		return ndisks;

	file.Close();
	return 0;
}

// ---------------------------------------------------------------------------
//	}`fBXNC[WǂmF
//	}`fBXNC[WȂ炻̂߂̏쐬
// ---------------------------------------------------------------------------

bool WinFDU::CheckMultiDiskImage(const char* filename)
{
	int i;
	FileIO fio;
	if (!(fio.Open(filename, FileIO::readonly)))
	{
		ndisks = 0;
		return false;
	}
	ndisks = 1;
	diskdesc[0].title[0] = 0;
	diskdesc[0].pos = 0;

	// fBXN P88SR `H
	struct ImageHeader
	{
		char title[17];
		uint8 reserved[9];
		uint8 readonly;
		uint8 disktype;
		uint32 disksize;
	} header;
	
	fio.Read(&header, 32);
	if (!memcmp("M88 RawDiskImage", header.title, 16))
		return false;
	if (!header.disksize)			// TransDisk 0.02 oO΍
		return false;

	for (i=0; i<9; i++) if (header.reserved[i]) return false;
	header.title[17] = 0;		// = reserved[0];
	if (strlen(header.title) == 17) return false;
	
	// fBXNt@CH
	fio.Seek(0, FileIO::end);
	uint32 size = fio.Tellp();
	if (size <= header.disksize) return false;

	// efBXÑ^Cgƈʒu擾
	uint32 fpos = 0;
	for (i=0; i<WINFDU_MAXDISKINFILE && fpos < size; i++)
	{
		fio.Seek(fpos, FileIO::begin);
		fio.Read(&header, sizeof(ImageHeader));
		strncpy(diskdesc[i].title, header.title, 17);
		diskdesc[i].pos = fpos;
		fpos += header.disksize;
	}
	ndisks = i;
	return true;
}

// ---------------------------------------------------------------------------
//	fBXN^Cg𓾂
// ---------------------------------------------------------------------------

char* WinFDU::GetDiskTitle(uint diskid)
{
	if (diskid >= ndisks)
		return 0;
	return diskdesc[diskid].title;
}

// ---------------------------------------------------------------------------
//	fBXNւ
// ---------------------------------------------------------------------------

bool WinFDU::SetDisk(uint diskid)
{
	// fBXN͂
	bool result = true;
	if (fdu)
	{
		CriticalSection::Lock lock(cs);
		result = fdu->Unmount();
		delete fdu;
		fdu = 0;
	}
	if (!result)
		return false;
	
	Sleep(100);

	if (diskid < ndisks)
	{
		file.SetLogicalOrigin(diskdesc[diskid].pos);

		if (DoMount(new FDU0))
			return true;
		if (DoMount(new FDUB))
			return true;
	}
	return false;
}

// ---------------------------------------------------------------------------
//	fBXNC[W͂
// ---------------------------------------------------------------------------

bool WinFDU::Unmount()
{
	bool result = true;
	ndisks = 0;
	if (fdu)
	{
		CriticalSection::Lock lock(cs);
		result = fdu->Unmount();
		delete fdu;
		fdu = 0;
		file.Close();
	}
	return result;
}

// ---------------------------------------------------------------------------
//	fBXN̓}EgĂ邩H
// ---------------------------------------------------------------------------

bool WinFDU::IsMount()
{
	CriticalSection::Lock lock(cs);
	return fdu ? true : false;
}

// ---------------------------------------------------------------------------
//	Mount ł邩݂
// ---------------------------------------------------------------------------

bool WinFDU::DoMount(FDU* t_fdu)
{
	if (t_fdu)
	{
		if (t_fdu->Mount(&file))
		{
			CriticalSection::Lock lock(cs);
			fdu = t_fdu;
			fdu->Seek(cy);
			return true;
		}
		delete t_fdu;
	}
	return false;
}

// ---------------------------------------------------------------------------
//	WinFDU::SenceDeviceStatus
// ---------------------------------------------------------------------------

uint8 WinFDU::SenceDeviceStatus()
{
	CriticalSection::Lock lock(cs);

	return fdu ? fdu->SenceDeviceStatus() : FDC::ST3_RY;
}

// ---------------------------------------------------------------------------
//	WinFDU::ReadSector
// ---------------------------------------------------------------------------

FDC::Result WinFDU::ReadSector(uint8 flags, ID id, uint8* data)
{
	CriticalSection::Lock lock(cs);

	if (fdu)
		return fdu->ReadSector(flags, id, data);
	else
		return Result(FDC::ST0_AT | FDC::ST0_NR);
}

// ---------------------------------------------------------------------------
//	WinFDU::WriteSector
// ---------------------------------------------------------------------------

FDC::Result WinFDU::WriteSector(uint8 flags, ID id, const uint8* data, bool deleted)
{
	CriticalSection::Lock lock(cs);

	if (fdu)
		return fdu->WriteSector(flags, id, data, deleted);
	else
		return Result(FDC::ST0_AT | FDC::ST0_NR);
}

// ---------------------------------------------------------------------------
//	WinFDU::Seek
// ---------------------------------------------------------------------------

FDC::Result WinFDU::Seek(uint8 cyrinder)
{
	CriticalSection::Lock lock(cs);

	cy = cyrinder;
	if (fdu)
		return fdu->Seek(cyrinder);
	else
		return Result(0);
}

// ---------------------------------------------------------------------------
//	WinFDU::ReadID
// ---------------------------------------------------------------------------

FDC::Result WinFDU::ReadID(uint8 flags, ID* id)
{
	CriticalSection::Lock lock(cs);

	if (fdu)
		return fdu->ReadID(flags, id);
	else
		return Result(FDC::ST0_AT | FDC::ST1_MA);
}

// ---------------------------------------------------------------------------
//	WinFDU::WriteID
// ---------------------------------------------------------------------------

FDC::Result WinFDU::WriteID(uint8 flags, WIDDESC id)
{
	CriticalSection::Lock lock(cs);

	if (fdu)
		return fdu->WriteID(flags, id);
	else
		return Result(FDC::ST0_AT | FDC::ST1_MA);
}
