// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	Joystick Implementation for Win32
// ---------------------------------------------------------------------------
//	$Id: WinJoy.cpp,v 1.2 1999/01/11 15:32:17 cisc Exp $

// #define	DIRECTINPUT_VERSION		0x0300

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include "WinJoy.h"
#include "M88Config.h"

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

WinJoyPad::WinJoyPad()
{
	enabled = false;
	button1 = JOY_BUTTON1 | JOY_BUTTON3;
	button2 = JOY_BUTTON2 | JOY_BUTTON4;
	threshold = 16384;
}

WinJoyPad::~WinJoyPad()
{

}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

bool WinJoyPad::Init()
{
	paravalid = false;
	enabled = false;
	if (!joyGetNumDevs())
		return false;
	
	JOYINFO joyinfo;
	if (joyGetPos(JOYSTICKID1, &joyinfo) == JOYERR_UNPLUGGED)
		return false;
	enabled = true;
	return true;
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

uint8 WinJoyPad::In(uint port)
{
	if (!paravalid)
	{
		JOYINFO joyinfo;
		if (enabled && joyGetPos(JOYSTICKID1, &joyinfo) == JOYERR_NOERROR)
		{
			data[0] =  (joyinfo.wYpos < (32768-threshold) ? 0 : 1)	// U
					 | (joyinfo.wYpos > (32768+threshold) ? 0 : 2)	// D
					 | (joyinfo.wXpos < (32768-threshold) ? 0 : 4)	// L
					 | (joyinfo.wXpos > (32768+threshold) ? 0 : 8)	// R
					 | 0xf0;
			data[1] =  (joyinfo.wButtons & button1 ? 0 : 1)
					 | (joyinfo.wButtons & button2 ? 0 : 2)
					 | 0xfc;
		}
		else
			data[0] = data[1] = 0xff;
		paravalid = true;
	}
	return data[port];
}

// ---------------------------------------------------------------------------
//	RtBM[VXV
// ---------------------------------------------------------------------------

void WinJoyPad::ApplyConfig(const Config* config)
{
	if (config->flags & Config::swappadbuttons)
	{
		button1 = JOY_BUTTON2 | JOY_BUTTON4;
		button2 = JOY_BUTTON1 | JOY_BUTTON3;
	}
	else
	{
		button1 = JOY_BUTTON1 | JOY_BUTTON3;
		button2 = JOY_BUTTON2 | JOY_BUTTON4;
	}
	threshold = config->padsensibility;
}