// ---------------------------------------------------------------------------
//  M88 - PC88 emulator
//  Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	ʊ֌W for Windows
// ---------------------------------------------------------------------------
//	$Id: WinScreen.cpp,v 1.2 1999/01/11 15:32:18 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <process.h>
#include <crtdbg.h>
#include <malloc.h>
#include <ddraw.h>

#include "misc.h"
#include "WinCriticalSection.h"
#include "WinScreen.h"
#include "Win88.h"
#include "WinDraw.h"
#include "WinDrawGDI.h"
#include "WinDrawDDW.h"
#include "WinDrawDDS.h"
#include "MSGs.h"

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
	#define WINSCRN_LOGGING
#endif

#ifdef WINSCRN_LOGGING
	#include <stdio.h>
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("winscrn.dmp", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif


// ---------------------------------------------------------------------------
// \z/
// ---------------------------------------------------------------------------

WinScreen::WinScreen()
{
	draw = 0;
	drawtype = None;
	drawtypenew = None;
	idthread = 0;
	hthread = 0;
	shouldterminate = false;
	putinfo = false;
	hevredraw = 0;
	pc = 0;
	updateall = false;
	fcount = 0;
	guicount = 0;
	palettemode = 0;
	force480 = false;
	LogBegin();
}

WinScreen::~WinScreen()
{
	Cleanup();
	LogEnd();
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

bool WinScreen::Init0(HWND hwindow)
{
	hwnd = hwindow;
	draw = 0;
	drawtype = None;
	
	Screen::Init();

	fcount = 0;
	hthread = 0;
	hevredraw = CreateEvent(NULL, FALSE, FALSE, NULL);

	return true;
}

void WinScreen::Cleanup()
{
	if (hthread)
	{
		int i;
		SetThreadPriority(hthread, THREAD_PRIORITY_NORMAL);
		SetEvent(hevredraw);
		for (i=300; i>0 && (WAIT_TIMEOUT == WaitForSingleObject(hthread, 10)); i--)
		{
			shouldterminate = true;
			SetEvent(hevredraw);
		}   
		if (!i)
		{
			_RPT0(_CRT_WARN, "WinScreen XbhI\n");
			TerminateThread(hthread, 0);
		}
		CloseHandle(hthread), hthread = 0;
	}
	if (hevredraw)
		CloseHandle(hevredraw), hevredraw = 0;
	
	delete draw;	draw = 0;
}

bool WinScreen::Init(Win88 * _pc)
{
	pc = _pc;
	
	shouldterminate = false;
	hthread = HANDLE(_beginthreadex(NULL, 0, ThreadEntry, 
						   reinterpret_cast<LPVOID>(this), 0, &idthread));
	return !!hthread;
}

// ---------------------------------------------------------------------------
//	pbgZbg
// ---------------------------------------------------------------------------

void WinScreen::SetPalette(int num, uint8 red, uint8 green, uint8 blue)
{
	static const uint8 ctable[2][8] = 
	{
		{   0,  36,  73, 109, 146, 182, 219, 255 }, // analogue
		{   0, 255, 255, 255, 255, 255, 255, 255 }, // digital
	};
	
	palentry[num].peRed = ctable[palettemode][red];
	palentry[num].peGreen = ctable[palettemode][green];
	palentry[num].peBlue = ctable[palettemode][blue];
	palchanged = true;
}

// ---------------------------------------------------------------------------
//	ʕ`p̃Xbh[v
// ---------------------------------------------------------------------------

uint WinScreen::ThreadMain()
{
//	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);

	updated = true;
	while (!shouldterminate)
	{
		PaintWindow();
		WaitForSingleObject(hevredraw, 100);
	}
	return 0;
}

// ---------------------------------------------------------------------------
//	Xbhɓ
// ---------------------------------------------------------------------------

uint __stdcall WinScreen::ThreadEntry(LPVOID arg)
{
	WinScreen* instance = reinterpret_cast<WinScreen*> (arg);
	if (instance)
		return instance->ThreadMain();
	return 0;
}

// ---------------------------------------------------------------------------
//	pbgf
// ---------------------------------------------------------------------------

void WinScreen::QueryNewPalette(bool background)
{
	CriticalSection::Lock lock(csdraw);
	if (draw)
	{
		draw->QueryNewPalette();
	}
}

// ---------------------------------------------------------------------------
//	yCg
// ---------------------------------------------------------------------------

void WinScreen::RequestPaint(bool redraw)
{
	if (redraw)
	{
		updateall = true;
		SetEvent(hevredraw);
	}
	else
	{
		if (updated)
		{
			updated = false;
			//pc->Snap();
			SetEvent(hevredraw);
		}
//		PulseEvent(hevredraw);
	}
}

// ---------------------------------------------------------------------------
//	`悷
// ---------------------------------------------------------------------------

void WinScreen::PaintWindow()
{
	uint8* image;
	int bpl;
	int region[2];
		
	// ʍ
	{
		CriticalSection::Lock lock(csdraw);
		if (draw && draw->Lock(&image, &bpl))
		{
			pc->Snap();
			pc->UpdateScreen(image, bpl, GetTickCount(), region, putinfo, redraw);
			draw->Unlock();
			if (palchanged)
				draw->SetPalette(palentry);
			palchanged = false, redraw = false;
		}
	}
	updated = true;

	// `
	bool update = updateall;
	updateall = false;

	region[0] = Max(  0, region[0]);
	region[1] = Min(399, region[1]);
	
	{
		CriticalSection::Lock lock(csdraw);
		if (draw && (update || (region[0] <= region[1])))
		{
			draw->UpdateScreen(region[0], region[1], update);
			fcount ++;
		}
	}
}

// ---------------------------------------------------------------------------
//	t[JEg擾
// ---------------------------------------------------------------------------

int WinScreen::GetFCount()
{
	int f = int(fcount);
	fcount = 0;
	return f;
}

// ---------------------------------------------------------------------------
//	GUI gptOݒ
// ---------------------------------------------------------------------------

void WinScreen::SetGUIFlag(bool usegui)
{
	CriticalSection::Lock lock(csdraw);
	if (usegui)
	{
		if (!guicount++ && draw)
		{
			draw->SetGUIMode(true);
			int count = ShowCursor(true);
			Log1("SetGUIFlag (false) : %d\n", count);
		}
	}
	else
	{
		if (!--guicount && draw)
		{
			draw->SetGUIMode(false);
			int count = ShowCursor(false);
			Log1("SetGUIFlag (true)  : %d\n", count);
		} 
	}
}

// ---------------------------------------------------------------------------
//	`惂[hύX
// ---------------------------------------------------------------------------

bool WinScreen::ChangeDisplayDriver(DisplayType type)
{
	if (type != drawtype)
	{
		// ʂ̕ۑ
		uint8* imcache = new uint8[640*400];
		if (!imcache)
			return false;
		SaveScreen(imcache);

		// ܂ł̃hCop
		if (draw)
			draw->SetGUIMode(true);
		{
			CriticalSection::Lock lock(csdraw);		
			delete draw; draw = 0;
		}

		// VhCo̗p
		WinDraw* newdraw;
		switch (type)
		{
		case GDI: default:
			newdraw = new WinDrawGDI;
			break;
		case DDWin:
			newdraw = new WinDrawDDW;
			break;
		case DDFull:
			newdraw = new WinDrawDDS(force480);
			break;
		}
		
		bool result = true;
		
		if (!newdraw || !newdraw->Init(hwnd))
		{
			// Ɏsꍇ GDI hCoōĒ
			delete newdraw;
			newdraw = new WinDrawGDI, type = GDI;
			result = false;

			if (!newdraw || !newdraw->Init(hwnd))
				newdraw = 0, drawtype = None;
		}
		
		if (newdraw)
			newdraw->SetGUIMode(false);
		guicount = 0;
		
		// VhCogp\ɂ
		{
			CriticalSection::Lock lock(csdraw);
			draw = newdraw;
		}

		// ʂ𕜌
		updateall = true, palchanged = true;
		LoadScreen(imcache);
		delete[] imcache;

		if (type == DDFull)		ShowCursor(false);
		if (drawtype == DDFull)	ShowCursor(true);
		drawtype = type;
		
		return result;
	}
	return false;
}

// ---------------------------------------------------------------------------
//	ʂ̒g擾
//	dest	ۑA640x400 bytes
// ---------------------------------------------------------------------------

void WinScreen::SaveScreen(uint8 * dest)
{
	if (draw)
	{
		uint8* src;
		int bpl;
		if (draw->Lock(&src, &bpl))
		{
			for (int y=0; y<400; y++)
			{
				memcpy(dest, src, 640);
				dest+=640, src+=bpl;
			}
			draw->Unlock();
		}
	}
}

// ---------------------------------------------------------------------------
//	ʂ̒g
//	src		u摜  640x400 bytes
// ---------------------------------------------------------------------------

void WinScreen::LoadScreen(const uint8 * src)
{
	if (draw)
	{
		uint8* dest;
		int bpl;
		if (draw->Lock(&dest, &bpl))
		{
			for (int y=0; y<400; y++)
			{
				memcpy(dest, src, 640);
				src+=640, dest+=bpl;
			}
			draw->Unlock();
		}
	}
}

// ---------------------------------------------------------------------------
//	ݒXV
//	config	Vݒ
// ---------------------------------------------------------------------------

void WinScreen::ApplyConfig(const Config* config)
{
	putinfo = 0 != (config->flags & Config::watchregister);
	force480 = 0 != (config->flags & Config::force480);
	int newpalmode = config->flags & Config::digitalpalette ? 1 : 0;
	if (newpalmode != palettemode) 
		reinitpalette = true;
	palettemode = newpalmode;
	if (hthread)
	{
		SetThreadPriority(hthread, 
			config->flags & Config::drawprioritylow ? 
				THREAD_PRIORITY_BELOW_NORMAL : THREAD_PRIORITY_NORMAL);
	}

}

// ---------------------------------------------------------------------------
//	ʂ 640x400x4  BMP ɕϊ
//	dest	ϊ BMP ̒uꏊAu邾̗̈悪KvB
//	ret		Iłǂ
// ---------------------------------------------------------------------------

bool WinScreen::Capture(uint8* dest)
{
	uint8* src = new uint8[640*400];
	if (!src) return false;

	SaveScreen(src);

	// \̂̏

	BITMAPFILEHEADER* filehdr = (BITMAPFILEHEADER*) dest;
	BITMAPINFO* binfo = (BITMAPINFO*) (filehdr+1);
	uint8* image = ((uint8*)(binfo+1)) + 15 * sizeof(RGBQUAD);
	
	// headers
	filehdr->bfType = 'MB';
	filehdr->bfSize	= image + 640*400/2 - dest ;
	filehdr->bfReserved1 = 0;
	filehdr->bfReserved2 = 0;
	filehdr->bfOffBits = image - dest;
	binfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	binfo->bmiHeader.biWidth = 640;
	binfo->bmiHeader.biHeight = 400;
	binfo->bmiHeader.biPlanes = 1;
	binfo->bmiHeader.biBitCount = 4;
	binfo->bmiHeader.biCompression = BI_RGB;
	binfo->bmiHeader.biSizeImage = 0;
	binfo->bmiHeader.biXPelsPerMeter = 0;
	binfo->bmiHeader.biYPelsPerMeter = 0;
	binfo->bmiHeader.biClrImportant = 0;

	// PUFpbg̍쐬
	RGBQUAD* palette = binfo->bmiColors;
	memset(palette, 0, sizeof(RGBQUAD)*16);
	
	uint8 ctable[256];
	memset(ctable, 0, sizeof(ctable));
	
	int colors=0;
	for (int index=0; index<128; index++)
	{
		RGBQUAD rgb;
		rgb.rgbBlue  = palentry[index].peBlue;
		rgb.rgbRed   = palentry[index].peRed;
		rgb.rgbGreen = palentry[index].peGreen;
		uint32 entry = *((uint32*)&rgb);
		
		int k;
		for (k=0; k<colors; k++)
		{
			if (!((*((uint32*)&palette[k]) ^ entry) & 0xffffff))
				goto match;
		}
		if (colors<15)
			palette[colors++] = rgb;
		else 
			k=15;
match:
		ctable[64+index] = k;
	}
	binfo->bmiHeader.biClrUsed = colors;
	
	// Fϊ
	uint8* d = image;
	for (int y=0; y<400; y++)
	{
		uint8* s = src + 640 * (399-y);
		
		for (int x=0; x<320; x++, s+=2)
			*d++ = ctable[s[0]] * 16 + ctable[s[1]];
	}

	delete[] src;
	return true;
}

