// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1997, 1998.
// ---------------------------------------------------------------------------
//	Sound Implementation for Win32
// ---------------------------------------------------------------------------
//	$Id: WinSound.cpp,v 1.2 1999/01/11 15:32:18 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#include <dsound.h>
#include <process.h>
#include <crtdbg.h>
#include <math.h>
#include "WinSound.h"
#include "misc.h"
#include "M88Config.h"
#include "WinJoy.h"

// ---------------------------------------------------------------------------
//	fobOp

#ifndef RELEASEBUILD
	#define WINSOUND_LOGGING
#endif

#ifdef WINSOUND_LOGGING
	#include <stdio.h>
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("winsound.dmp", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

WinSound::WinSound()
{
	LogBegin();
}

WinSound::~WinSound()
{
	Cleanup();
	LogEnd();
}

void WinSound::Reset()
{
	EnableOPN2(!!(cfgflg & (Config::opnaona8 | Config::opnona8)));
	opn.SetMode(!!(cfgflg & Config::enableopna));
	opn2.SetMode(!!(cfgflg & Config::opnaona8));

	Sound88::Reset();
}

// ---------------------------------------------------------------------------
//	E㏈
// ---------------------------------------------------------------------------

bool WinSound::Init(HWND hwindow, uint rate, uint32 clock, PC88* pc, uint buflen)
{
	uint samprate = rate;
	currentrate = rate;
	currentbuflen = buflen;		
	hwnd = hwindow;
	
	switch (rate)
	{
	case SOUND_55K:
		samprate = 44100;
	case 11025: case 22050: case 44100: 
		break;

	default:
		rate = 100; samprate = 0;
		break;
	}
	
	joy.Init();
	
	int bufsize = 0;
	if (rate > 1000 && driver.Init(this, hwnd, samprate, 2, buflen))
	{
		bufsize = (rate * 4 * buflen / 2000) & ~15;		// 1/2 buflen
	}
		
	if (!Sound88::Init(rate, clock, bufsize, pc))
		return false;

	return true;
}

void WinSound::Cleanup()
{
	driver.Cleanup();
	
	Sound88::Cleanup();
}

bool WinSound::ChangeRate(uint rate, uint32 clock, uint buflen)
{
	if (currentrate != rate || currentbuflen != buflen)
	{
		CriticalSection::Lock lock(csmix);

		bool resetdriver = true;
		if (currentrate + rate == 44100 + SOUND_55K)
			resetdriver = false;

		uint samprate = rate;
		currentrate = rate;
		currentbuflen = buflen;		
		
		if (rate == SOUND_55K)
		{
			samprate = 44100;
		}
		else if (rate < 8000)
		{
			rate = 100; samprate = 0;
		}
		
		int bufsize = (rate * buflen / 500) & ~15;		// 1/2 buflen
		
		if (resetdriver)
		{
			driver.Cleanup();

			if (!(rate > 1000 && driver.Init(this, hwnd, samprate, 2, buflen)))
				bufsize = 0;
		}
			
		if (!SetRate(rate, clock, bufsize))
			return false;
	}
	return true;
}

// ---------------------------------------------------------------------------
//	ԏ
// ---------------------------------------------------------------------------

void WinSound::Count(int32 us, bool timer)
{
	CriticalSection::Lock lock(csmix);
	Sound88::Count(us, timer);
}

// ---------------------------------------------------------------------------
//	I
// ---------------------------------------------------------------------------

void WinSound::VSync()
{
	joy.VSync();
//	driver.Send();
}

// ---------------------------------------------------------------------------
//	ėp|[g
// ---------------------------------------------------------------------------

uint8 WinSound::PortIn(uint port)
{
	return enablepad ? joy.In(port) : 0xff;
}

// ---------------------------------------------------------------------------
//	ʐݒ
// ---------------------------------------------------------------------------

inline int WinSound::ConvertVolume(int volume)
{
	return volume > -40 ? volume : -200;
}

void WinSound::SetVolume(const Config* config)
{
	CriticalSection::Lock lock(csmix);
	psgb.SetVolume(ConvertVolume(config->volssg));
	psg2.SetVolume(ConvertVolume(config->volssg));

	SetOPNVolume(config, opn);
	SetOPNVolume(config, opn2);
}

void WinSound::SetOPNVolume(const Config* config, OPN88& o)
{
	o.SetVolumeFM(ConvertVolume(config->volfm));
	o.SetVolumeADPCM(ConvertVolume(config->voladpcm));
	o.SetVolumeRhythmTotal(ConvertVolume(config->volrhythm));
	o.SetVolumeRhythm(0, ConvertVolume(config->volbd));
	o.SetVolumeRhythm(1, ConvertVolume(config->volsd));
	o.SetVolumeRhythm(2, ConvertVolume(config->voltop));
	o.SetVolumeRhythm(3, ConvertVolume(config->volhh));
	o.SetVolumeRhythm(4, ConvertVolume(config->voltom));
	o.SetVolumeRhythm(5, ConvertVolume(config->volrim));
}

// ---------------------------------------------------------------------------
//	ݒXV
// ---------------------------------------------------------------------------

void WinSound::ApplyConfig(const Config* config)
{
	SetVolume(config);
	EnableSING(!(config->flags & Config::disablesing));
	
	cfgflg = config->flags;
	
	joy.ApplyConfig(config);
	enablepad = !!(config->flags & Config::enablepad);
	
	static const uint srate[] = { 0, 11025, 22050, 44100, SOUND_55K, 48000 };
	ChangeRate(srate[config->sound], config->opnclock, config->soundbuffer);
}


// ---------------------------------------------------------------------------
//	ĐhCo
// ---------------------------------------------------------------------------

const uint WinSound::Driver::num_blocks = 5;
const uint WinSound::Driver::timer_resolution = 20;


WinSound::Driver::Driver()
{
	playing = false;
	lpds = 0;
	lpdsb = 0;
	lpdsb_primary = 0;
	timerid = 0;
	sending = false;
}

WinSound::Driver::~Driver()
{
	Cleanup();
}

//   -------------------------------------------------------------------

bool WinSound::Driver::Init(WinSound* ws, HWND hwnd, uint rate, uint ch, uint buflen)
{
	if (playing)
		return false;

	winsound = ws;
	buffer_length = buflen;

	// vZ
	buffersize = (rate * ch * 2 * buffer_length / 1000) & ~7;

	// DirectSound object 쐬
	if (DS_OK != DirectSoundCreate(NULL, &lpds, 0))
		return false;

	// xݒ
	if (DS_OK != lpds->SetCooperativeLevel(hwnd, DSSCL_PRIORITY))
	{
		if (DS_OK != lpds->SetCooperativeLevel(hwnd, DSSCL_NORMAL))
			return false;
	}
	
	DSBUFFERDESC dsbd;
    memset(&dsbd, 0, sizeof(DSBUFFERDESC));
    dsbd.dwSize = sizeof(DSBUFFERDESC);
    dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;
    dsbd.dwBufferBytes = 0; 
    dsbd.lpwfxFormat = 0;
	if (DS_OK != lpds->CreateSoundBuffer(&dsbd, &lpdsb_primary, 0))
		return false;

	// ĐtH[}bgݒ
	WAVEFORMATEX wf;
    memset(&wf, 0, sizeof(WAVEFORMATEX));
    wf.wFormatTag = WAVE_FORMAT_PCM;
	wf.nChannels = ch;
    wf.nSamplesPerSec = rate;
	wf.wBitsPerSample = 16;
	wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
	wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;

	lpdsb_primary->SetFormat(&wf);

	// ZJ_obt@쐬
    memset(&dsbd, 0, sizeof(DSBUFFERDESC));
    dsbd.dwSize = sizeof(DSBUFFERDESC);
    dsbd.dwFlags = DSBCAPS_CTRLDEFAULT | DSBCAPS_STICKYFOCUS | DSBCAPS_GETCURRENTPOSITION2;

    dsbd.dwBufferBytes = buffersize; 
    dsbd.lpwfxFormat = &wf;
    
	HRESULT res = lpds->CreateSoundBuffer(&dsbd, &lpdsb, NULL); 
	if (DS_OK != res)
		return false;

	// Đ
	lpdsb->Play(0, 0, DSBPLAY_LOOPING);
//	lpdsb_primary->Play(0, 0, DSBPLAY_LOOPING);

	// ^C}[쐬
	timeBeginPeriod(buffer_length / num_blocks);
	timerid = timeSetEvent(buffer_length / num_blocks, timer_resolution,
		TimeProc, reinterpret_cast<DWORD> (this), TIME_PERIODIC);
	nextwrite = 4;
	
	if (!timerid)
	{
		timeEndPeriod(buffer_length / num_blocks);
		return false;
	}

	playing = true;
	return true;
}

//  Еt -----------------------------------------------------------------

bool WinSound::Driver::Cleanup()
{
	playing = false;
	if (timerid)
	{
		timeKillEvent(timerid);
		timeEndPeriod(buffer_length / num_blocks);
		timerid = 0;
	}
	if (lpdsb)
	{
		lpdsb->Stop();
		lpdsb->Release(), lpdsb = 0;
	}
	if (lpdsb_primary)
		lpdsb_primary->Release(), lpdsb_primary = 0;
	if (lpds)
		lpds->Release(), lpds = 0;
	return true;
}

//  TimeProc  ----------------------------------------------------------------

void CALLBACK WinSound::Driver::TimeProc(UINT uid, UINT, DWORD user, DWORD, DWORD)
{
	Driver* inst = reinterpret_cast<Driver*> (user);
	if (inst)
		inst->Send();
}

//  ubN -------------------------------------------------------------

void WinSound::Driver::Send()
{
	if (playing && !InterlockedExchange(&sending, true))
	{
		bool restored = false;

		// Buffer Lost ?
		DWORD status;
		lpdsb->GetStatus(&status);
		if (DSBSTATUS_BUFFERLOST & status)
		{
			// restore the buffer
//			lpdsb_primary->Restore();
			if (DS_OK != lpdsb->Restore())
				goto ret;
			nextwrite = 0;
			restored = true;
		}

		// ʒu擾
		DWORD cplay, cwrite;
		lpdsb->GetCurrentPosition(&cplay, &cwrite);

		if (cplay == nextwrite && !restored)
			goto ret;

		// ݃TCYvZ
		int writelength;
		if (cplay < nextwrite)
			writelength = cplay + buffersize - nextwrite;
		else
			writelength = cplay - nextwrite;

		writelength &= ~3;

		if (writelength <= 0)
			goto ret;

		Log3("play = %5d  write = %5d  length = %5d\n", cplay, nextwrite, writelength);
		{
			CriticalSection::Lock lock(winsound->csmix);
			
			void* a1, * a2;
			DWORD al1, al2;
			// Lock
			if (DS_OK != lpdsb->Lock(nextwrite, writelength,
									 (void**) &a1, &al1, (void**) &a2, &al2, 0))
				goto ret;

		// 

			if (a1)
				winsound->Fill(a1, al1);
			if (a2)
				winsound->Fill(a2, al2);

			// Unlock
			lpdsb->Unlock(a1, al1, a2, al2);
		}

		nextwrite += writelength;
		if (nextwrite >= buffersize)
			nextwrite -= buffersize;

		if (restored)
			lpdsb->Play(0, 0, DSBPLAY_LOOPING);

		// I
ret:
		InterlockedExchange(&sending, false);
	}
	return;
}

