// ---------------------------------------------------------------------------
//	Z80 emulator in C++
//	Copyright (C) cisc 1997, 1998.
// ----------------------------------------------------------------------------
//	$Id: Z80R.cpp,v 1.2 1999/01/11 15:32:18 cisc Exp $

#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>
#include "Z80R.h"

// ---------------------------------------------------------------------------

#define SHORT_IOADDRESS		// I/O |[g̃AhX 8 rbgłꍇ
							// 88 ̃G~[Vɂ͕͂Kv

#if !defined(_MSC_VER) || _MSC_VER < 1200)
	#define __assume(x)		(0)
#endif

// ---------------------------------------------------------------------------

#define CF		(1 << 0)
#define NF		(1 << 1)
#define PF		(1 << 2)
#define HF		(1 << 4)
#define ZF		(1 << 6)
#define SF		(1 << 7)

#define WF		(1 << 3)

// ---------------------------------------------------------------------------

//#define Read8(a)		(Read(a))
//#define Write8(a, d)	(Write(a, d))
//#define Read16(a)		(Read(a) + Read(a+1) * 256)
// inline void Write16(uint a, uint d) { Write8(a, d); Write8(a+1, d>>8); }

#define Inp(p)			(InPort((p) & 0xff))
#define Outp(p, d)		(OutPort((p) & 0xff, d))

// ---------------------------------------------------------------------------
//	88 ˑANZX[`
// ---------------------------------------------------------------------------

inline void Z80R::Write8(uint addr, uint8 data)
{
	addr &= 0xffff;
#ifndef DEBUG_F
	if (gvramselected)
	{
		if (addr < 0xc000)
			writeaccess[addr >> 8][addr & 0xff] = data;
		else
			Write(addr, data);
	}
	else
	{
		writeaccess[addr >> 8][addr & 0xff] = data;
	}
#else
	Write(addr, data);
#endif
}

inline uint8 Z80R::Read8(uint addr)
{
	addr &= 0xffff;
#ifndef DEBUG_F
	if (gvramselected)
	{
		if (addr < 0xc000)
			return readaccess[addr >> 8][addr & 0xff];
		else
			return Read(addr);
	}
	else
		return readaccess[addr >> 8][addr & 0xff];
#else
	return Read(addr);
#endif
}

uint Z80R::Read16(uint a)
{
	a &= 0xffff;
#ifndef DEBUG_F
	if (!gvramselected || a < 0xbfff)
	{
#ifndef READ16_AT_ONCE
		return readaccess[a     >> 8][a     & 0xff] 
			 + readaccess[(a+1) >> 8][(a+1) & 0xff] * 256;
#else
		uint bank = a >> 8;
		uint offs = a & 0xff;
		if (offs < 0xff)
			return *(uint16*)&readaccess[bank][offs];
		return readaccess[bank][0xff] + readaccess[(bank+1)&255][0] * 256;
#endif
	}
#endif
	return Read(a) + 256 * Read(a+1);
}

void Z80R::Write16(uint a, uint d)
{
	a &= 0xffff;
#ifndef DEBUG_F
	if (!gvramselected || a < 0xbfff)
	{
		writeaccess[ a    >> 8][ a    & 0xff] = uint8(d);
		writeaccess[(a+1) >> 8][(a+1) & 0xff] = uint8(d >> 8);
		return;
	}
#endif
	Write(a,   uint8(d & 0xff));
	Write(a+1, uint8((d >> 8) & 0xff));
	return;
}

// ---------------------------------------------------------------------------
//	}NQ -----------------------------------------------------------------

#ifdef STAT_COUNT
	#define INP_REF(p)			// (itable[(p) & 0xff]++)
	#define OUTP_REF(p)			// (otable[(p) & 0xff]++)
	#define EXEC_REF(p)			(etable[(p) & 0xff]++)
	#define EXEC_REF_I(p)		(eitable[(p) & 0xff]++)
	#define EXEC_REF_CB(p)		(ectable[(p) & 0xff]++)
	#define EXEC_REF_ED(p)		(eetable[(p) & 0xff]++)
#else
	#define INP_REF(p)			(0)
	#define OUTP_REF(p)			(0)
	#define EXEC_REF(p)			(0)
	#define EXEC_REF_I(p)		(0)
	#define EXEC_REF_CB(p)		(0)
	#define EXEC_REF_ED(p)		(0)
#endif

	#define INP_REF(p)			(0)
	#define OUTP_REF(p)			(0)

#ifndef SHORT_IOADDRESS
	#define InpC()			Inp(RegBC)
	#define OutpC(d)		(void)Outp(RegBC, (d))
#else
	#define InpC()			Inp(RegC)
	#define OutpC(d)		(void)Outp(RegC, (d))
#endif


#define RegA			(reg.r.b.a)
#define RegB			(reg.r.b.b)
#define RegC			(reg.r.b.c)
#define RegD			(reg.r.b.d)
#define RegE			(reg.r.b.e)
#define RegH			(reg.r.b.h)
#define RegL			(reg.r.b.l)
#define RegXH			(*ref_h[index_mode])
#define RegXL			(*ref_l[index_mode])
#define RegF			(reg.r.b.flags)

#define RegXHL			(*ref_hl[index_mode])
#define RegHL			(reg.r.w.hl)
#define RegDE			(reg.r.w.de)
#define RegBC			(reg.r.w.bc)
#define RegAF			(reg.r.w.af)
#define RegSP			(reg.r.w.sp)


#define Fetch8()		Read8(reg.pc)
#define Fetch16()		Read16(reg.pc)

#define GetNF()			(RegF & NF)
#define SetXF(n)		(xf = n)

#define SetFlags(m,f)	(void)((RegF = (RegF & ~(m)) | (f)), (uf &= ~(m)))

#define	Jump(d)			(void)(reg.pc = (d))

#define Ret()			(void)Jump(Pop())

#define JumpR()			(void)(Jump(reg.pc + 1 + (int8)(Fetch8()) ),CLK(5))

#define RES(n, bit)		((n) & ~(1 << (bit)))
#define SET(n, bit)		((n) | (1 << (bit))) 
#define BIT(n, bit)		(void)SetFlags(ZF|HF|NF|SF, HF|(((n) & (1 << (bit))) ? (n & SF) : ZF)), \
						SetXF(n)

#define	CLK(count)		(clockcount -= (count))

// ---------------------------------------------------------------------------
//	KɊ֐Q
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
//  ֐ -----------------------------------------------------------------

void Z80R::Call()
{
	uint d = Fetch16();
	Push(reg.pc+2);
	Jump(d);
	CLK(7);
}

// ---------------------------------------------------------------------------
//	ANZX⏕֐ ---------------------------------------------------------

void Z80R::SetM(uint8 n)
{
	if (index_mode == USEHL)
		Write8(RegHL, n);
	else
	{
		Write8(RegXHL + int8(Fetch8()), n);
		reg.pc++;
		CLK(12);
	}
}

uint8 Z80R::GetM()
{
	if (index_mode == USEHL)
		return Read8(RegHL);
	else
	{
		uint8 r = Read8(RegXHL + int8(Fetch8()));
		reg.pc++;
		CLK(12);
		return r;
	}
}

uint Z80R::GetAF()
{
	RegF &= 0xd7;
#ifndef DEBUG_F
	RegF |= xf & 0x28; 
#endif
	if (uf & (CF|ZF|SF|HF|PF))
		GetCF(), GetZF(), GetSF(), GetHF(), GetPF();
	return RegAF;
}

void Z80R::SetAF(uint n)
{
	RegAF = n;
	uf = 0;
}

// ---------------------------------------------------------------------------
//	X^bN֐ -------------------------------------------------------------

void Z80R::Push(uint n)
{
	RegSP -= 2;
	Write16(RegSP, n);
}

uint Z80R::Pop()
{
	uint a = Read16(RegSP);
	RegSP += 2;
	return a;
}

// ---------------------------------------------------------------------------
//	ZpZ֐ -------------------------------------------------------------

void Z80R::ADDA(uint8 n)
{
	fx = uint(RegA) * 2;
	fy = uint(n   ) * 2;
	uf = SF|ZF|HF|PF|CF;
	nfa = 0;
	RegF &= ~NF;
	RegA += n;
	SetXF(RegA);
}

void Z80R::ADCA(uint8 n)
{
	uint8 a = RegA;
	uint8 cy = GetCF();
	
	RegA = a + n + cy;
	SetXF(RegA);

	fx = uint(a) * 2 + 1;
	fy = uint(n) * 2 + cy;
	uf = SF|ZF|HF|PF|CF;
	nfa = 0;
	RegF &= ~NF;
}

void Z80R::SUBA(uint8 n)
{
	fx = RegA * 2;
	fy = n    * 2;
	uf = SF|ZF|HF|PF|CF;
	nfa = 1;
	RegF |= NF;
	RegA -= n;
	SetXF(RegA);
}

void Z80R::CPA(uint8 n)
{
	fx = RegA * 2;
	fy = n    * 2;
	uf = SF|ZF|HF|PF|CF;
	nfa = 1;
	RegF |= NF;
	SetXF(RegA);
}

void Z80R::SBCA(uint8 n)
{
	uint8 a = RegA;
	uint8 cy = GetCF();
	RegA = (a - n - cy);
	
	fx = a * 2;
	fy = n * 2 + cy;
	uf = SF|ZF|HF|PF|CF;
	nfa = 1;
	RegF |= NF;
	SetXF(RegA);
}

void Z80R::ANDA(uint8 n)
{
	uint8 b = RegA & n;
	SetZSP(b);
	SetFlags(HF|NF|CF, HF);
	RegA = (b);
}

void Z80R::ORA(uint8 n)
{
	uint8 b = RegA | n;
	SetZSP(b);
	SetFlags(HF|NF|CF, 0);
	RegA = (b);
}

void Z80R::XORA(uint8 n)
{
	uint8 b = RegA ^ n;
	SetZSP(b);
	SetFlags(HF|NF|CF, 0);
	RegA = (b);
}

uint8 Z80R::Inc8(uint8 y)
{
	y++;
	SetFlags(SF|ZF|HF|PF|NF,
		  ((y == 0) ? ZF : 0)
		| ((y & 0x80) ? SF : 0)
		| ((y == 0x80) ? PF : 0)
		| ((y & 0x0f) ? 0 : HF)
		);
	SetXF(y);
	return y;
}

uint8 Z80R::Dec8(uint8 y)
{
	y--;
	SetFlags(SF|ZF|HF|PF|NF,
		  ((y == 0) ? ZF : 0)
		| ((y & 0x80) ? SF : 0)
		| ((y == 0x7f) ? PF : 0)
		| (((y & 0x0f) == 0xf) ? HF : 0)
		| NF
		);
	SetXF(y);
	return y;
}

uint Z80R::ADD16(uint x, uint y)
{
	x &= 0xffff;
	y &= 0xffff;
	fx32 = x * 2;
	fy32 = y * 2;
	uf = CF|HF|WF;
	SetFlags(NF, 0);
	nfa = 0;
	return x + y;
}

void Z80R::ADCHL(uint y)
{
	uint cy = GetCF();
	uint x = RegHL & 0xffff;
	y &= 0xffff;
	RegHL = x + y + cy;

	fx32 = (uint32)(x) * 2 + 1;
	fy32 = (uint32)(y) * 2 + cy;
	uf = SF|ZF|HF|PF|CF|WF;
	nfa = 0;
	RegF &= ~NF;
	SetXF(RegH);
}

void Z80R::SBCHL(uint y)
{
	uint cy = GetCF();
	uint x = RegHL & 0xffff;
	y &= 0xffff;
	RegHL = x - y - cy;

	fx32 = (uint32)(x & 0xffff) * 2;
	fy32 = (uint32)(y & 0xffff) * 2 + cy;
	uf = SF|ZF|HF|PF|CF|WF;
	nfa = 1;
	RegF |= NF;
	SetXF(RegH);
}

// ---------------------------------------------------------------------------
//	[e[gEVtg ---------------------------------------------------

uint8 Z80R::RLC(uint8 d)
{
	uint8 f = (d & 0x80) ? CF : 0;
	d = (d << 1) + f;			/* CF == 1 */
	
	SetZSP(d);
	SetFlags(CF|NF|HF, f);
	return d;
}

uint8 Z80R::RRC(uint8 d)
{
	uint8 f = d & 1;
	d = (d >> 1) + (f ? 0x80 : 0);

	SetZSP(d);
	SetFlags(CF|NF|HF, f);
	return d;
}

uint8 Z80R::RL(uint8 d)
{
	uint8 f = (d & 0x80) ? CF : 0;
	d = (d << 1) + GetCF();

	SetZSP(d);
	SetFlags(CF|NF|HF, f);
	return d;
}

uint8 Z80R::RR(uint8 d)
{
	uint8 f = d & 1;
	d = (d >> 1) + (GetCF() ? 0x80 : 0);

	SetZSP(d);
	SetFlags(CF|NF|HF, f);
	return d;
}

uint8 Z80R::SLA(uint8 d)
{
	SetFlags(NF|HF|CF, (d & 0x80) ? CF : 0);
	d <<= 1;
	SetZSP(d);
	return d;
}

uint8 Z80R::SRA(uint8 d)
{
	SetFlags(NF|HF|CF, d & 1);
	d = (d >> 1) + (d & 0x80);
	SetZSP(d);
	return d;
}

uint8 Z80R::SLL(uint8 d)
{
	SetFlags(NF|HF|CF, (d & 0x80) ? CF : 0);
	d = (d << 1) + 1;
	SetZSP(d);
	return d;
}

uint8 Z80R::SRL(uint8 d)
{
	SetFlags(NF|HF|CF, d & 1);
	d = d >> 1;
	SetZSP(d);
	return d;
}

// ---------------------------------------------------------------------------
//  ݐtO -----------------------------------------------------

void Z80R::SetIFF(bool flag)
{
	reg.iff1 = flag;
}

// ---------------------------------------------------------------------------
//   tOe[u
// ---------------------------------------------------------------------------

static const uint8 ZSPTable[256] = 
{
	ZF|PF,      0,      0,     PF,      0,     PF,     PF,      0,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	    0,     PF,     PF,      0,     PF,      0,      0,     PF,
	   PF,      0,      0,     PF,      0,     PF,     PF,      0,
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	PF|SF,     SF,     SF,  PF|SF,     SF,  PF|SF,  PF|SF,     SF, 
	   SF,  PF|SF,  PF|SF,     SF,  PF|SF,     SF,     SF,  PF|SF, 
};


// ---------------------------------------------------------------------------
//   ߃NbNe[u
// ---------------------------------------------------------------------------
//   򖽗߂̏ꍇ͎sꍇ̃NbN

static const uint8 ClockTable[0x100] = 
{	
	4,	10,	7,	6,	4,	4,	7,	4,	4,	11,	7,	6,	4,	4,	7,	4,	/*0*/
	8,	10,	7,	6,	4,	4,	7,	4,	7,	11,	7,	6,	4,	4,	7,	4,	
	7,	10,	16,	6,	4,	4,	7,	4,	7,	11,	16,	6,	4,	4,	7,	4,
	7,	10,	13,	6,	11,	11,	7,	4,	7,	11,	13,	6,	4,	4,	7,	4,
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	/*4*/
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	
	7,	7,	7,	7,	7,	7,	4,	7,	4,	4,	4,	4,	4,	4,	7,	4,
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	/*8*/
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	
	4,	4,	4,	4,	4,	4,	7,	4,	4,	4,	4,	4,	4,	4,	7,	4,	
	5,	10,	10,	10,	10,	11,	7,	4,	5,	4,	10,	0,	10,	10,	7,	4,	/*C*/
	5,	10,	10,	11,	10,	11,	7,	4,	5,	4,	10,	11,	10,	4,	7,	4,
	5,	10,	10,	19,	10,	11,	7,	4,	5,	4,	10,	4,	10,	8,	7,	4,
	5,	10,	10,	4,	10,	11,	7,	4,	5,	6,	10,	4,	10,	4,	7,	4,
};

// ---------------------------------------------------------------------------
//	RXgN^EfXgN^
// ---------------------------------------------------------------------------

Z80R::Z80R()
{
#ifdef DEBUG_MONITOR
	fpmon = 0;
	monitor = false;
#endif
	Init();
#ifdef STAT_COUNT
	memset(etable,  0, sizeof(etable));
	memset(eitable, 0, sizeof(eitable));
	memset(ectable, 0, sizeof(ectable));
	memset(eetable, 0, sizeof(eetable));
#endif
}

Z80R::~Z80R()
{
#ifdef DEBUG_MONITOR
	if (fpmon)
		fclose(fpmon);
#endif
#ifdef STAT_COUNT
	{
		FILE* fp = fopen("Z80_EXEC.LOG", "w");
		if (fp)
		{
			int i;
			fprintf(fp, "Z80R ߎs\n");
			for (i=0; i<0x100; i++)
			{
				fprintf(fp, ((i & 7) == 7) ? "%0.2x:%8ld\n" : "%0.2x:%8ld   ", i, etable[i]);
			}
			fprintf(fp, "\n\nZ80R ED ߗ\n");
			for (i=0; i<0x100; i++)
			{
				fprintf(fp, ((i & 7) == 7) ? "%0.2x:%8ld\n" : "%0.2x:%8ld   ", i, eetable[i]);
			}
			fprintf(fp, "\n\nZ80R CB ߗ\n");
			for (i=0; i<0x100; i++)
			{
				fprintf(fp, ((i & 7) == 7) ? "%0.2x:%8ld\n" : "%0.2x:%8ld   ", i, ectable[i]);
			}
			fprintf(fp, "\n\nZ80R Indexed ߗ\n");
			for (i=0; i<0x100; i++)
			{
				fprintf(fp, ((i & 7) == 7) ? "%0.2x:%8ld\n" : "%0.2x:%8ld   ", i, eitable[i]);
			}

			fclose(fp);
		}
	}
#endif
}

// ---------------------------------------------------------------------------
//	荞݂̐\
//	intno	荞ݔԍ (0-31Ԃ܂)
// ---------------------------------------------------------------------------

void Z80R::RequestInterrupt(uint intno)
{
	intr_req |= 1 << intno;
}

void Z80R::CancelInterrupt(uint intno)
{
	intr_req &= ~(1 << intno);
}


// ---------------------------------------------------------------------------
//	荞
// ---------------------------------------------------------------------------

void Z80R::TestIntr()
{
	uint32 reqbit = intr_req & intr_mask;
	if (reqbit && reg.iff1)
	{
		int intno;
		uint32 m = 1;
		for (intno=0; intno<31; intno++, m *= 2)
		{
			if (reqbit & m)
				break;
		}
		
//		intr_req &= ~m;

		SetIFF(false);
		AcceptInterrupt(intno);
#ifdef DEBUG_MONITOR
		if (fpmon)
		{
			fprintf(fpmon, "Int%d accepted(%d)\n", intno / 2, reg.intmode);
		}
#endif

		if (reg.haltstate & 1)
		{
			reg.haltstate = 0;
			reg.pc++;
		}

		switch (reg.intmode)
		{
		case 0:
			switch (intno)
			{
				uint w;
			case 0x00:					// NOP
				break;
			case 0x02:
				Write8(RegBC, RegA);	// LD (BC),A
				break;
			case 0x04:					// INC B
				RegB = (Inc8(RegB)); 
				break;
			case 0x06:					// LD B,n
				RegB = Fetch8(); reg.pc++; 
				break;
			case 0x08:					// EX AF,AF'
				w = GetAF(); SetAF(reg.r_af); reg.r_af = w;
				break;
			case 0x0a:					// LD A,(BC)
				RegA = Read8(RegBC); 
				break;
			case 0x0c:					// INC C
				RegC = (Inc8(RegC)); 
				break;
			case 0x0e:					// LD C,n
				RegC = Fetch8(); reg.pc++; 
				break;
			}
			break;
		case 1:
			Push(reg.pc);
			Jump(0x38);
			break;
		case 2:
			Push(reg.pc);
			Jump(Read16((reg.ireg*256 + intno)));
			break;
		}
	}
}

// ---------------------------------------------------------------------------
//  荞
// ---------------------------------------------------------------------------

void Z80R::NMI()
{
	reg.iff2 = reg.iff1;
	SetIFF(false);
	Push(reg.pc);
	Jump(0x66);
}

// ---------------------------------------------------------------------------
//  Zbg
// ---------------------------------------------------------------------------

void Z80R::Reset(int rstat)
{
	SetIFF(false);		/* ݋֎~ */
	reg.iff2 = false;
	reg.ireg = 0;			/* I, R = 0 */
	reg.rreg = 0;
	SetFlags(0xff, 0);	/* tOZbg */
	reg.intmode = 0;			/* IM0 */
	reg.pc = 0;				/* pc, sp = 0 */
	RegSP = 0;
	intr_req = 0;
	reg.haltstate = 0;

	wait = rstat != 0;
}

// ---------------------------------------------------------------------------
//  \̏
// ---------------------------------------------------------------------------

void Z80R::Init()
{
	wait = true;
	/* e[u */
	ref_h[USEHL] = &RegH;	ref_l[USEHL] = &RegL;
	ref_h[USEIX] = &reg.r.b.xh;	ref_l[USEIX] = &reg.r.b.xl;
	ref_h[USEIY] = &reg.r.b.yh;	ref_l[USEIY] = &reg.r.b.yl;
	ref_hl[USEHL] = &RegHL;
	ref_hl[USEIX] = &reg.r.w.ix;
	ref_hl[USEIY] = &reg.r.w.iy;
	ref_byte[0] = &RegB;	ref_byte[1] = &RegC;
	ref_byte[2] = &RegD;	ref_byte[3] = &RegE;
	ref_byte[4] = &RegH;	ref_byte[5] = &RegL;
	ref_byte[6] = 0; 		ref_byte[7] = &RegA;
	
	index_mode = USEHL;

	clockcount = 0;
	intr_req=0;
//	next_index = USEHL;

	Reset(0);
}

// ---------------------------------------------------------------------------
//  ߐs
// ---------------------------------------------------------------------------

void Z80R::StepDual(uint clocks, Z80R* sub)
{
	TestIntr();
	sub->TestIntr();
	for (clockcount = clocks; clockcount > 0; )
	{
		SingleStep();	SingleStep();	SingleStep();	SingleStep();
		sub->SingleStep();	sub->SingleStep();
		sub->SingleStep();	sub->SingleStep();
	}
}

void Z80R::StepDual2(uint clocks, Z80R* sub)
{
	TestIntr();
	sub->TestIntr();
	for (clockcount = clocks; clockcount > 0; )
	{
		SingleStep();	SingleStep();	SingleStep();	SingleStep();
		SingleStep();	SingleStep();	SingleStep();	SingleStep();
		sub->SingleStep();	sub->SingleStep();
		sub->SingleStep();	sub->SingleStep();
	}
}

void Z80R::StepDualP(uint clocks, Z80R* sub, int x, int y)
{
	TestIntr();
	sub->TestIntr();
	for (clockcount = clocks; clockcount > 0; )
	{
		SingleStep();	SingleStep();	SingleStep();	SingleStep();
		SingleStep();	SingleStep();	SingleStep();	SingleStep();
		sub->SingleStep();	sub->SingleStep();
		sub->SingleStep();	sub->SingleStep();
	}
}

void Z80R::Step(uint clocks)
{
	TestIntr();
	for (clockcount = clocks; clockcount > 0; )
	{
		SingleStep();
#ifndef DEBUG_F
		SingleStep();
		SingleStep();
		SingleStep();
#endif
	}
}

// ---------------------------------------------------------------------------
//  Pߎs
// ---------------------------------------------------------------------------

void Z80R::SingleStep()
{
#ifdef DEBUG_F
	xf = RegA;
#endif
	reg.rreg++;
//	index_mode = idx;
//	next_index = USEHL;

#if !defined(DEBUG_MONITOR) && !defined(DEBUG_F)
	uint pc = reg.pc & 0xffff;
	uint m = (gvramselected && pc >= 0xc000) ? 
		uint(Read8(pc)) : uint(readaccess[pc >> 8][pc & 0xff]);
#else
	uint m = Read8(reg.pc);
#endif

#ifdef DEBUG_MONITOR
	if (monitor == 1 && index_mode == USEHL)
	{
		Monitor();
	}
#endif // DEBUG_MONITOR
	
	reg.pc++;
	
	CLK(ClockTable[m]);
	EXEC_REF(m);
	
	switch (m)
	{
		uint8 b;
		uint w;
		
	// [e[gVtgn
		
	case 0x07:	// RLCA
		b = (0 != (RegA & 0x80));
		RegA = RegA * 2 + b;
		SetFlags(NF|HF|CF, b);		/* Cn = 1 */
		break;

	case 0x0f:	// RRCA
		b = RegA & 1;
		RegA = (RegA >> 1) + (b ? 0x80 : 0);
		SetFlags(NF|HF|CF, b);
		break;

	case 0x17:	// RLA
		b = RegA;
		RegA = (b << 1) + GetCF();
		SetFlags(NF|HF|CF, (b & 0x80) ? CF : 0);
		break;

	case 0x1f:	// RRA
		b = RegA;
		RegA = (GetCF() ? 0x80 : 0) + (b >> 1);
		SetFlags(NF|HF|CF, b & 1);
		break;

// ZpZn

	case 0x27:	// DAA
		b = 0;
		if ((RegA & 0x0f) > 9 || GetHF())
		{
			RegA += GetNF() ? -6 : 6;
			b = HF;
		}
		if (RegA > 0x9f || GetCF())
		{
			RegA += GetNF() ? -0x60 : 0x60;
			b |= CF;
		}
		SetZSP(RegA);
		SetFlags(HF|CF, b);
		break;

	case 0x2f:	// CPL
		RegA = ~RegA;
		SetFlags(NF|HF, NF|HF);
		break;

	case 0x37:	// SCF
		SetFlags(CF|NF|HF, CF);
		break;

	case 0x3f:	// CCF
		b = GetCF(); SetFlags(CF|NF, b ^ CF);
		break;

// o͌n

	case 0xdb:	// IN A,(n)
#ifndef SHORT_IOADDRESS
		w = (((uint) RegA) << 8) + Fetch8();
#else
		w = Fetch8();
#endif
		RegA=Inp(w);
		INP_REF(w);
		reg.pc++;
		break;

	case 0xd3:	// OUT (n),A
#ifndef SHORT_IOADDRESS
		w = (((uint) RegA) << 8) + Fetch8();
#else
		w = Fetch8();
#endif
		Outp(w, RegA);
		OUTP_REF(w);
		reg.pc++;
		TestIntr();
		break;

// n

	case 0xc3:	// JP
		Jump(Fetch16());
		break;

	case 0xc2:	/*NZ*/ if (!GetZF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xca:	/* Z*/ if ( GetZF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xd2:	/*NC*/ if (!GetCF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xda:	/* C*/ if ( GetCF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xe2:	/*PO*/ if (!GetPF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xea:	/*PE*/ if ( GetPF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xf2:	/* P*/ if (!GetSF()) Jump(Fetch16()); else reg.pc+=2; break; 
	case 0xfa:	/* M*/ if ( GetSF()) Jump(Fetch16()); else reg.pc+=2; break; 

	case 0xcd:	// CALL
		Call();
		break;

	case 0xc4:	/*NZ*/ if (!GetZF()) Call(); else reg.pc+=2; break; 
	case 0xcc:	/* Z*/ if ( GetZF()) Call(); else reg.pc+=2; break; 
	case 0xd4:	/*NC*/ if (!GetCF()) Call(); else reg.pc+=2; break; 
	case 0xdc:	/* C*/ if ( GetCF()) Call(); else reg.pc+=2; break; 
	case 0xe4:	/*PO*/ if (!GetPF()) Call(); else reg.pc+=2; break; 
	case 0xec:	/*PE*/ if ( GetPF()) Call(); else reg.pc+=2; break; 
	case 0xf4:	/* P*/ if (!GetSF()) Call(); else reg.pc+=2; break; 
	case 0xfc:	/* M*/ if ( GetSF()) Call(); else reg.pc+=2; break; 

	case 0xc9:	// RET
		Ret();
		break;

	case 0xc0:	/*NZ*/ if (!GetZF()) Ret(); break; 
	case 0xc8:	/* Z*/ if ( GetZF()) Ret(); break; 
	case 0xd0:	/*NC*/ if (!GetCF()) Ret(); break; 
	case 0xd8:	/* C*/ if ( GetCF()) Ret(); break; 
	case 0xe0:	/*PO*/ if (!GetPF()) Ret(); break; 
	case 0xe8:	/*PE*/ if ( GetPF()) Ret(); break; 
	case 0xf0:	/* P*/ if (!GetSF()) Ret(); break; 
	case 0xf8:	/* M*/ if ( GetSF()) Ret(); break; 

	case 0x18:	// JR
		JumpR();
		break;

	case 0x20:	/*NZ*/ if (!GetZF()) JumpR(); else reg.pc++; break;
	case 0x28:	/* Z*/ if ( GetZF()) JumpR(); else reg.pc++; break;
	case 0x30:	/*NC*/ if (!GetCF()) JumpR(); else reg.pc++; break;
	case 0x38:	/* C*/ if ( GetCF()) JumpR(); else reg.pc++; break;

	case 0xe9:	// JP (HL)
		Jump(RegXHL);
		break;

	case 0x10:	// DJNZ
		if (0 != -- RegB) 
			JumpR();
		else 
			reg.pc++;
		break;

	case 0xc7:	/* RST 00H */	Push(reg.pc); Jump(0x00); break;
	case 0xcf:	/* RST 08H */	Push(reg.pc); Jump(0x08); break;
	case 0xd7:	/* RST 10H */	Push(reg.pc); Jump(0x10); break;
	case 0xdf:	/* RST 18H */	Push(reg.pc); Jump(0x18); break;
	case 0xe7:	/* RST 20H */	Push(reg.pc); Jump(0x20); break;
	case 0xef:	/* RST 28H */	Push(reg.pc); Jump(0x28); break;
	case 0xf7:	/* RST 30H */	Push(reg.pc); Jump(0x30); break;
	case 0xff:	/* RST 38H */	Push(reg.pc); Jump(0x38); break;

// 16 rbgZn

	// ADD XHL,dd
	case 0x09:	/*BC*/	RegXHL = ADD16(RegXHL, RegBC); break;
	case 0x19:	/*DE*/	RegXHL = ADD16(RegXHL, RegDE); break;
	case 0x29:	/*xHL*/	w = RegXHL; RegXHL = ADD16(w, w); break;
	case 0x39:	/*SP*/	RegXHL = ADD16(RegXHL, RegSP); break;
	
	// INC dd
	case 0x03:	/*BC*/	RegBC++; break;
	case 0x13:	/*DE*/	RegDE++; break;
	case 0x23:	/*xHL*/	RegXHL++; break;
	case 0x33:	/*SP*/	RegSP++; break;

	// DEC dd
	case 0x0B:	/*BC*/	RegBC--; break;
	case 0x1B:	/*DE*/	RegDE--; break;
	case 0x2B:	/*xHL*/	RegXHL--; break;
	case 0x3B:	/*SP*/	RegSP--; break;

// 

	case 0x08:	// EX AF,AF'
		w = GetAF();
		SetAF(reg.r_af);
		reg.r_af = w;
		break;

	case 0xe3:	// EX (SP),xHL
		w = Read16(RegSP);
		Write16(RegSP, RegXHL);
		RegXHL = w;
		break;

	case 0xeb:	// EX DE,HL
			w = RegDE;
		RegDE = RegHL;
		RegHL = w;
		break;

	case 0xd9:	// EXX
		w = RegHL; RegHL = reg.r_hl; reg.r_hl = w;
		w = RegDE; RegDE = reg.r_de; reg.r_de = w;
		w = RegBC; RegBC = reg.r_bc; reg.r_bc = w;
		break;

// bot

	case 0xf3:	// DI
		SetIFF(false);
		break;

	case 0xfb:	// EI
		b = Read8(reg.pc);
		if (b != 0xf3 && b != 0xfb)
		{
			SingleStep();
			SetIFF(true);
			TestIntr();
		}
		break;

	case 0x00:	// NOP
		break;

	case 0x76:	// HALT
		reg.pc--;
		reg.haltstate = 1;
		CLK(1000);
		break;

// 8 bit Zn

	// ADD A,-
	case 0x80: /*B*/ ADDA(RegB);	break;
	case 0x81: /*C*/ ADDA(RegC);	break;
	case 0x82: /*D*/ ADDA(RegD);	break;
	case 0x83: /*E*/ ADDA(RegE);	break;
	case 0x84: /*H*/ ADDA(RegXH);	break;
	case 0x85: /*L*/ ADDA(RegXL);	break;
	case 0x86: /*M*/ ADDA(GetM());	break;
	case 0x87: /*A*/ ADDA(RegA);	break;
	case 0xc6: /*n*/ ADDA(Fetch8()); reg.pc++; break;

	// ADC A,-
	case 0x88: /*B*/ ADCA(RegB);	break;
	case 0x89: /*C*/ ADCA(RegC);	break;
	case 0x8a: /*D*/ ADCA(RegD);	break;
	case 0x8b: /*E*/ ADCA(RegE);	break;
	case 0x8c: /*H*/ ADCA(RegXH);	break;
	case 0x8d: /*L*/ ADCA(RegXL);	break;
	case 0x8e: /*M*/ ADCA(GetM());	break;
	case 0x8f: /*A*/ ADCA(RegA);	break;
	case 0xce: /*n*/ ADCA(Fetch8()); reg.pc++; break;

	// SUB -
	case 0x90: /*B*/ SUBA(RegB);	break;
	case 0x91: /*C*/ SUBA(RegC);	break;
	case 0x92: /*D*/ SUBA(RegD);	break;
	case 0x93: /*E*/ SUBA(RegE);	break;
	case 0x94: /*H*/ SUBA(RegXH);	break;
	case 0x95: /*L*/ SUBA(RegXL);	break;
	case 0x96: /*M*/ SUBA(GetM());	break;
	case 0x97: /*A*/ SUBA(RegA);	break;
	case 0xd6: /*n*/ SUBA(Fetch8()); reg.pc++; break;

	// SBC A,-
	case 0x98: /*B*/ SBCA(RegB);	break;
	case 0x99: /*C*/ SBCA(RegC);	break;
	case 0x9a: /*D*/ SBCA(RegD);	break;
	case 0x9b: /*E*/ SBCA(RegE);	break;
	case 0x9c: /*H*/ SBCA(RegXH);	break;
	case 0x9d: /*L*/ SBCA(RegXL);	break;
	case 0x9e: /*M*/ SBCA(GetM());	break;
	case 0x9f: /*A*/ SBCA(RegA);	break;
	case 0xde: /*n*/ SBCA(Fetch8()); reg.pc++; break;

	// AND -
	case 0xa0: /*B*/ ANDA(RegB);	break;
	case 0xa1: /*C*/ ANDA(RegC);	break;
	case 0xa2: /*D*/ ANDA(RegD);	break;
	case 0xa3: /*E*/ ANDA(RegE);	break;
	case 0xa4: /*H*/ ANDA(RegXH);	break;
	case 0xa5: /*L*/ ANDA(RegXL);	break;
	case 0xa6: /*M*/ ANDA(GetM());	break;
	case 0xa7: /*A*/ ANDA(RegA);	break;
	case 0xe6: /*n*/ ANDA(Fetch8()); reg.pc++; break;

	// XOR -
	case 0xa8: /*B*/ XORA(RegB);	break;
	case 0xa9: /*C*/ XORA(RegC);	break;
	case 0xaa: /*D*/ XORA(RegD);	break;
	case 0xab: /*E*/ XORA(RegE);	break;
	case 0xac: /*H*/ XORA(RegXH);	break;
	case 0xad: /*L*/ XORA(RegXL);	break;
	case 0xae: /*M*/ XORA(GetM());	break;
	case 0xaf: /*A*/ XORA(RegA);	break;
	case 0xee: /*n*/ XORA(Fetch8()); reg.pc++; break;

	// OR -
	case 0xb0: /*B*/ ORA(RegB);	break;
	case 0xb1: /*C*/ ORA(RegC);	break;
	case 0xb2: /*D*/ ORA(RegD);	break;
	case 0xb3: /*E*/ ORA(RegE);	break;
	case 0xb4: /*H*/ ORA(RegXH);	break;
	case 0xb5: /*L*/ ORA(RegXL);	break;
	case 0xb6: /*M*/ ORA(GetM());	break;
	case 0xb7: /*A*/ ORA(RegA);	break;
	case 0xf6: /*n*/ ORA(Fetch8()); reg.pc++; break;

	// CP -
	case 0xb8: /*B*/ CPA(RegB);	break;
	case 0xb9: /*C*/ CPA(RegC);	break;
	case 0xba: /*D*/ CPA(RegD);	break;
	case 0xbb: /*E*/ CPA(RegE);	break;
	case 0xbc: /*H*/ CPA(RegXH);	break;
	case 0xbd: /*L*/ CPA(RegXL);	break;
	case 0xbe: /*M*/ CPA(GetM());	break;
	case 0xbf: /*A*/ CPA(RegA);	break;
	case 0xfe: /*n*/ CPA(Fetch8()); reg.pc++; break;

	// INC r
	case 0x04: /*B*/ RegB = (Inc8(RegB)); break;
	case 0x0c: /*C*/ RegC = (Inc8(RegC)); break;
	case 0x14: /*D*/ RegD = (Inc8(RegD)); break;
	case 0x1c: /*E*/ RegE = (Inc8(RegE)); break;
	case 0x24: /*H*/ RegXH = (Inc8(RegXH)); break;
	case 0x2c: /*L*/ RegXL = (Inc8(RegXL)); break;
	case 0x3c: /*A*/ RegA = (Inc8(RegA)); break;
	
	case 0x34: /*M*/
		w = RegXHL;
		if (index_mode != USEHL)
		{		
			w += (int8)(Fetch8()); reg.pc++;
			CLK(23-11);
		}
		Write8(w, Inc8(Read8(w)));
		break;

	// DEC r
	case 0x05: /*B*/ RegB = (Dec8(RegB)); break;
	case 0x0d: /*C*/ RegC = (Dec8(RegC)); break;
	case 0x15: /*D*/ RegD = (Dec8(RegD)); break;
	case 0x1d: /*E*/ RegE = (Dec8(RegE)); break;
	case 0x25: /*H*/ RegXH = (Dec8(RegXH)); break;
	case 0x2d: /*L*/ RegXL = (Dec8(RegXL)); break;
	case 0x3d: /*A*/ RegA = (Dec8(RegA)); break;
	
	case 0x35: /*M*/
		w = RegXHL;
		if (index_mode != USEHL)
		{		
			w += (int8)(Fetch8()); reg.pc++;
			CLK(23-11);
		}
		Write8(w, Dec8(Read8(w)));
		break;

// X^bNn

	// PUSH
	case 0xc5: /*BC*/ Push(RegBC); break;
	case 0xd5: /*DE*/ Push(RegDE); break;
	case 0xe5: /*xHL*/Push(RegXHL); break;
	case 0xf5: /*AF*/ Push(GetAF()); break;

	// POP
	case 0xc1: /*BC*/ RegBC = Pop(); break;
	case 0xd1: /*DE*/ RegDE = Pop(); break;
	case 0xe1: /*xHL*/RegXHL = Pop(); break;
	case 0xf1: /*AF*/ SetAF(Pop()); break;

// 16 rbg[h

	// LD dd,nn
	case 0x01: /*BC*/ RegBC  = Fetch16(); reg.pc+=2; break;
	case 0x11: /*DE*/ RegDE  = Fetch16(); reg.pc+=2; break;
	case 0x21: /*xHL*/RegXHL = Fetch16(); reg.pc+=2; break;
	case 0x31: /*SP*/ RegSP  = Fetch16(); reg.pc+=2; break;

	case 0x22: // LD (nn),xHL
		Write16(Fetch16(), RegXHL); reg.pc+=2;
		break;

	case 0x2a: // LD xHL,(nn)
		RegXHL = Read16(Fetch16()); reg.pc+=2;
		break;

	case 0xf9: // LD SP,HL
		RegSP = RegXHL;
		break;

// 8 bit [hn

	// LD B,-
	case 0x40: /*B*/              break; 
	case 0x41: /*C*/ RegB = RegC; break;
	case 0x42: /*D*/ RegB = RegD; break;
	case 0x43: /*E*/ RegB = RegE; break;
	case 0x44: /*H*/ RegB = RegXH; break;
	case 0x45: /*L*/ RegB = RegXL; break;
	case 0x46: /*M*/ RegB = GetM(); break;
	case 0x47: /*A*/ RegB = RegA; break;
	case 0x06: /*n*/ RegB = Fetch8(); reg.pc++; break;

	// LD C,-
	case 0x48: /*B*/ RegC = RegB; break;
	case 0x49: /*C*/              break;
	case 0x4a: /*D*/ RegC = RegD; break;
	case 0x4b: /*E*/ RegC = RegE; break;
	case 0x4c: /*H*/ RegC = RegXH; break;
	case 0x4d: /*L*/ RegC = RegXL; break;
	case 0x4e: /*M*/ RegC = GetM(); break;
	case 0x4f: /*A*/ RegC = RegA; break;
	case 0x0e: /*n*/ RegC = Fetch8(); reg.pc++; break;

	// LD D,-
	case 0x50: /*B*/ RegD = RegB; break;
	case 0x51: /*C*/ RegD = RegC; break;
	case 0x52: /*D*/              break;
	case 0x53: /*E*/ RegD = RegE; break;
	case 0x54: /*H*/ RegD = RegXH; break;
	case 0x55: /*L*/ RegD = RegXL; break;
	case 0x56: /*M*/ RegD = GetM(); break;
	case 0x57: /*A*/ RegD = RegA; break;
	case 0x16: /*n*/ RegD = Fetch8(); reg.pc++; break;

	// LD E,-
	case 0x58: /*B*/ RegE = RegB; break;
	case 0x59: /*C*/ RegE = RegC; break;
	case 0x5a: /*D*/ RegE = RegD; break;
	case 0x5b: /*E*/                   break;
	case 0x5c: /*H*/ RegE = RegXH; break;
	case 0x5d: /*L*/ RegE = RegXL; break;
	case 0x5e: /*M*/ RegE = GetM(); break;
	case 0x5f: /*A*/ RegE = RegA; break;
	case 0x1e: /*n*/ RegE = Fetch8(); reg.pc++; break;

	// LD H,-
	case 0x60: /*B*/ RegXH = RegB; break;
	case 0x61: /*C*/ RegXH = RegC; break;
	case 0x62: /*D*/ RegXH = RegD; break;
	case 0x63: /*E*/ RegXH = RegE; break;
	case 0x64: /*H*/                   break;
	case 0x65: /*L*/ RegXH = RegXL; break;
	case 0x66: /*M*/ RegH = GetM(); break;
	case 0x67: /*A*/ RegXH = RegA; break;
	case 0x26: /*n*/ RegXH = Fetch8(); reg.pc++; break;

	// LD L,-
	case 0x68: /*B*/ RegXL = RegB; break;
	case 0x69: /*C*/ RegXL = RegC; break;
	case 0x6a: /*D*/ RegXL = RegD; break;
	case 0x6b: /*E*/ RegXL = RegE; break;
	case 0x6c: /*H*/ RegXL = RegXH; break;
	case 0x6d: /*L*/                   break;
	case 0x6e: /*M*/ RegL = GetM(); break;
	case 0x6f: /*A*/ RegXL = RegA; break;
	case 0x2e: /*n*/ RegXL = Fetch8(); reg.pc++; break;

	// LD M,-
	case 0x70: /*B*/ SetM(RegB); break;
	case 0x71: /*C*/ SetM(RegC); break;
	case 0x72: /*D*/ SetM(RegD); break;
	case 0x73: /*E*/ SetM(RegE); break;
	case 0x74: /*H*/ SetM(RegH); break;
	case 0x75: /*L*/ SetM(RegL); break;
	case 0x77: /*A*/ SetM(RegA); break;
	case 0x36: /*n*/
		w = RegXHL;
		if (index_mode != USEHL)
		{
			w += int8(Fetch8()); reg.pc++; CLK(19-10);
		}
		Write8(w, Fetch8());
		reg.pc++;
		break;

	// LD A,-
	case 0x78: /*B*/ RegA = RegB; break;
	case 0x79: /*C*/ RegA = RegC; break;
	case 0x7a: /*D*/ RegA = RegD; break;
	case 0x7b: /*E*/ RegA = RegE; break;
	case 0x7c: /*H*/ RegA = RegXH; break;
	case 0x7d: /*L*/ RegA = RegXL; break;
	case 0x7e: /*M*/ RegA = GetM(); break;
	case 0x7f: /*A*/                   break;
	case 0x3e: /*n*/ RegA = Fetch8(); reg.pc++; break;

	// LD (--), A
	case 0x02: /*BC*/ Write8(RegBC, RegA); break;
	case 0x12: /*DE*/ Write8(RegDE, RegA); break;
	case 0x32: /*nn*/ Write8(Fetch16(), RegA); reg.pc += 2; break;

	// LD A, (--)
	case 0x0a: /*BC*/ RegA = Read8(RegBC); break;
	case 0x1a: /*DE*/ RegA = Read8(RegDE); break;
	case 0x3a: /*nn*/ RegA = Read8(Fetch16()); reg.pc += 2; break;

// DD / FD
	case 0xdd:
		b = Fetch8();
		if (b != 0xfd || b != 0xdd)		// ʂ͐
		{
			EXEC_REF_I(Read8(reg.pc));
			index_mode = USEIX;
			SingleStep();
			index_mode = USEHL;
		}
		break;

	case 0xfd:
		b = Fetch8();
		if (b != 0xfd || b != 0xdd)
		{
			EXEC_REF_I(Read8(reg.pc));
			index_mode = USEIY;
			SingleStep();
			index_mode = USEHL;
		}
		break;

// CB
	case 0xcb:
		if (index_mode == USEHL)
			reg.rreg++;
		CodeCB();
		break;

// ED
	case 0xed:
		b = Fetch8(); reg.pc++;
		reg.rreg++;
		EXEC_REF_ED(b);
		switch(b)
		{

		// o ED n
			
			// IN r,(c)
			case 0x40: INP_REF(RegBC); SetZSP(RegB=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x48: INP_REF(RegBC); SetZSP(RegC=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x50: INP_REF(RegBC); SetZSP(RegD=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x58: INP_REF(RegBC); SetZSP(RegE=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x60: INP_REF(RegBC); SetZSP(RegH=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x68: INP_REF(RegBC); SetZSP(RegL=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x70: INP_REF(RegBC); SetZSP(      InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			case 0x78: INP_REF(RegBC); SetZSP(RegA=InpC()); SetFlags(NF|HF,0); CLK(12-8); break;
			
			// OUT (C),r
			case 0x41: OUTP_REF(RegBC); OutpC(RegB); TestIntr(); CLK(12-8); break;
			case 0x49: OUTP_REF(RegBC); OutpC(RegC); TestIntr(); CLK(12-8); break;
			case 0x51: OUTP_REF(RegBC); OutpC(RegD); TestIntr(); CLK(12-8); break;
			case 0x59: OUTP_REF(RegBC); OutpC(RegE); TestIntr(); CLK(12-8); break;
			case 0x61: OUTP_REF(RegBC); OutpC(RegH); TestIntr(); CLK(12-8); break;
			case 0x69: OUTP_REF(RegBC); OutpC(RegL); TestIntr(); CLK(12-8); break;
			case 0x71: OUTP_REF(RegBC); OutpC(    0); TestIntr(); CLK(12-8); break;
			case 0x79: OUTP_REF(RegBC); OutpC(RegA); TestIntr(); CLK(12-8); break;

			case 0xa2:	// INI
				Write8(RegHL++, InpC()); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				break;

			case 0xaa: // IND
				Write8(RegHL--, InpC()); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				break;

			case 0xa3: // OUTI
				OutpC(Read8(RegHL++)); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF); 
				CLK(16-8);
				TestIntr(); 
				break;

			case 0xab: // OUTD
				OutpC(Read8(RegHL--)); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				TestIntr(); 
				break;
					
			case 0xb2: // INIR
				Write8(RegHL++, InpC()); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				if (RegB)
					reg.pc-=2;
				break;

			case 0xba: // INDR
				Write8(RegHL--, InpC()); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				if (RegB)
					reg.pc-=2;
				break;

			case 0xb3: // OTIR
				OutpC(Read8(RegHL++)); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF); 
				CLK(16-8);
				if (RegB)
					reg.pc-=2;
				TestIntr(); 
				break;

			case 0xbb: // OTDR
				OutpC(Read8(RegHL--)); 
				SetFlags(ZF|NF, (--RegB != 0) ? NF : NF|ZF);
				CLK(16-8);
				if (RegB)
					reg.pc-=2;
				TestIntr(); 
				break;

		// ubN]n

			case 0xa0: // LDI
				Write8(RegDE++, Read8(RegHL++));
				RegBC = (RegBC-1) & 0xffff;
				SetFlags(PF|NF|HF, RegBC ? PF : 0);
				CLK(16-8);
				break;

			case 0xa8: // LDD
				Write8(RegDE--, Read8(RegHL--));
				RegBC = (RegBC-1) & 0xffff;
				SetFlags(PF|NF|HF, RegBC ? PF : 0);
				CLK(16-8);
				break;

			case 0xb0: // LDIR
				Write8(RegDE++, Read8(RegHL++));
				RegBC = (RegBC-1) & 0xffff;
				SetFlags(PF|NF|HF, RegBC ? PF : 0);
				if (RegBC)
					reg.pc -= 2, CLK(20-8);
				else
					CLK(16-8);
				break;

			case 0xb8: // LDDR
				Write8(RegDE--, Read8(RegHL--));
				RegBC = (RegBC-1) & 0xffff;
				SetFlags(PF|NF|HF, RegBC ? PF : 0);
				CLK(16-8);
				if (RegBC)
					reg.pc -= 2, CLK(20-8);
				else
					CLK(16-8);
				break;

			// ubNT[`n

			case 0xa1: // CPI
				CPI();
				break;

			case 0xa9: // CPD
				CPD();
				break;

			case 0xb1: // CPIR
				CPI();
				if (!GetZF() && RegBC)
					reg.pc -= 2;
				break;

			case 0xb9: // CPDR
				CPD();
				if (!GetZF() && RegBC)
					reg.pc -= 2;
				break;

		// misc

			case 0x44:	// NEG
				b = RegA; RegA = (0); SUBA(b);
				break;

			case 0x46: /* IM 0 */ reg.intmode = 0; break;
			case 0x56: /* IM 1 */ reg.intmode = 1; break;
			case 0x5e: /* IM 2 */ reg.intmode = 2; break;

			case 0x57: // LD A,I
				RegA = (reg.ireg);
				SetZS(reg.ireg);
				SetFlags(NF|HF|PF, (reg.iff1 ? PF : 0));
				CLK(9-8);
				break;

			case 0x5F: // LD A,R
				RegA = (reg.rreg & 0x7f) + (reg.rreg7 & 0x80);
				SetZS(RegA);
				SetFlags(NF|HF|PF, (reg.iff1 ? PF : 0));
				CLK(9-8);
				break;

			case 0x47: // LD I,A
				reg.ireg = RegA; CLK(9-8); 
				break;

			case 0x4f: // LD R,A
				reg.rreg7 = reg.rreg = RegA; CLK(9-8); 
				break;

			case 0x45: // RETN
				Ret();
				CLK(14-8);
				break;

			case 0x4d: // RETI
				SetIFF(reg.iff2);
				Ret();
				CLK(14-8);
				break;

		// ړ

			case 0x6f: // RLD
				{
					uint8 d, e;
					
					d = Read8(RegHL);
					e = RegA & 0x0f;
					RegA = (RegA & 0xf0) + (d >> 4);
					d = ((d << 4) & 0xf0) + e;
					Write8(RegHL, d);
					
					SetZSP(RegA);
					SetFlags(NF|HF, 0);
					CLK(18-8);
				}
				break;

			case 0x67: // RRD
				{
					uint8 d, e;
					
					d = Read8(RegHL);
					e = RegA & 0x0f;
					RegA = (RegA & 0xf0) + (d & 0x0f);
					d = (d >> 4) + (e << 4);
					Write8(RegHL, d);
					
					SetZSP(RegA);
					SetFlags(NF|HF, 0);
					CLK(18-8);
				}
				break;

		// EDn 16 rbg[h
			
			// LD (nn),dd 
			case 0x43: /*BC*/ Write16(Fetch16(), RegBC); reg.pc+=2; break;
			case 0x53: /*DE*/ Write16(Fetch16(), RegDE); reg.pc+=2; break;
			case 0x63: /*HL*/ Write16(Fetch16(), RegHL); reg.pc+=2; break;
			case 0x73: /*SP*/ Write16(Fetch16(), RegSP); reg.pc+=2; break;

			// LD dd,(nn) 
			case 0x4b: /*BC*/ RegBC = Read16(Fetch16()); reg.pc+=2; break;
			case 0x5b: /*DE*/ RegDE = Read16(Fetch16()); reg.pc+=2; break;
			case 0x6b: /*HL*/ RegHL = Read16(Fetch16()); reg.pc+=2; break;
			case 0x7b: /*SP*/ RegSP = Read16(Fetch16()); reg.pc+=2; break;

		// EDn 16 rbgZ

			// ADC HL,dd
			case 0x4a: /*BC*/ ADCHL(RegBC); CLK(15-8); break;
			case 0x5a: /*DE*/ ADCHL(RegDE); CLK(15-8); break;
			case 0x6a: /*HL*/ ADCHL(RegHL); CLK(15-8); break;
			case 0x7a: /*SP*/ ADCHL(RegSP); CLK(15-8); break;

			// SBC HL,dd
			case 0x42: /*BC*/ SBCHL(RegBC); CLK(15-8); break;
			case 0x52: /*DE*/ SBCHL(RegDE); CLK(15-8); break;
			case 0x62: /*HL*/ SBCHL(RegHL); CLK(15-8); break;
			case 0x72: /*SP*/ SBCHL(RegSP); CLK(15-8); break;

			default: __assume(0);
		}
		break;		// 0xed
		
	default:
		__assume(0);
	}
#ifdef DEBUG_F
	GetAF();
#endif
}

// ---------------------------------------------------------------------------
//	CB n
// ---------------------------------------------------------------------------

void Z80R::CodeCB()
{
	typedef uint8 (Z80R::*RotFuncPtr)(uint8);

	static const RotFuncPtr func[8] =
	{
		&Z80R::RLC, &Z80R::RRC, &Z80R::RL,  &Z80R::RR, 
		&Z80R::SLA, &Z80R::SRA, &Z80R::SLL, &Z80R::SRL
	};
	
	int8 ref = (index_mode == USEHL) ? 0 : (int8)(Read8(reg.pc++));
	uint8 fn = Read8(reg.pc++);
	uint  rg = fn & 7;
	uint  bit = (fn >> 3) & 7;
	
	EXEC_REF_CB(fn);

	if (rg != 6)
	{
		uint8* p = ref_byte[rg];		/* Ώۂւ̃|C^ */
		switch ((fn >> 6) & 3)
		{
		case 0:
			*p = (this->*func[bit])(*p);
			break;
		case 1:
			BIT(*p, bit);
			break;
		case 2:
			*p = RES(*p, bit);
			break;
		case 3:
			*p = SET(*p, bit);
			break;
		}
		CLK(8);
	}
	else
	{
		uint b = *ref_hl[index_mode] + ref;
		uint8 d = Read8(b);
		switch ((fn >> 6) & 3)
		{
		case 0:
			Write8(b, (this->*func[bit])(d));
			CLK(15);
			break;
		case 1:
			BIT(d, bit);
			CLK(12);
			break;
		case 2:
			Write8(b, RES(d, bit));
			CLK(15);
			break;
		case 3:
			Write8(b, SET(d, bit));
			CLK(15);
			break;
		}
	}
}

void Z80R::CPI()
{
	uint8 n, f;
	n = Read8(RegHL++);
	RegBC = (RegBC-1) & 0xffff;
	f = (((RegA & 0x0f) < (n & 0x0f)) ? HF : 0)
	  | (RegBC ? PF : 0)
	  | NF;

	SetFlags(HF|PF|NF, f);
	SetZS(RegA-n);
	CLK(16-8);
}

void Z80R::CPD()
{
	uint8 n, f;
	n = Read8(RegHL--);
	RegBC = (RegBC-1) & 0xffff;
	f = (((RegA & 0x0f) < (n & 0x0f)) ? HF : 0)
	  | (RegBC ? PF : 0)
	  | NF;
	SetFlags(HF|PF|NF, f);
	SetZS(RegA-n);
	CLK(16-8);
}

// ---------------------------------------------------------------------------
//  tO֐ ---------------------------------------------------------------

uint8 Z80R::GetCF()
{
	if (uf & CF)
	{
		if (uf & WF)
		{
			if (nfa)
				SetFlags(CF, (fx32 < fy32) ? CF : 0);
			else
				SetFlags(CF, ((fx32 + fy32) & 0x20000ul) ? CF : 0); 
		}
		else
		{	
			if (nfa)
				SetFlags(CF, (fx < fy) ? CF : 0);
			else
				SetFlags(CF, ((fx + fy) & 0x200) ? CF : 0); 
		}
	}
	return RegF & CF;
}

uint8 Z80R::GetZF()
{
	if (uf & ZF)
	{
		if (uf & WF)
		{
			if (nfa)
				SetFlags(ZF, ((fx32 - fy32) & 0x1fffeul) ? 0 : ZF);
			else
				SetFlags(ZF, ((fx32 + fy32) & 0x1fffeul) ? 0 : ZF);
		}
		else
		{
			if (nfa)
				SetFlags(ZF, ((fx - fy) & 0x1fe) ? 0 : ZF);
			else
				SetFlags(ZF, ((fx + fy) & 0x1fe) ? 0 : ZF);
		}
	}
	return RegF & ZF;
}

uint8 Z80R::GetSF()
{
	if (uf & SF)
	{
		if (uf & WF)
		{
			if (nfa)
				SetFlags(SF, ((fx32 - fy32) & 0x10000ul) ? SF : 0);
			else
				SetFlags(SF, ((fx32 + fy32) & 0x10000ul) ? SF : 0);
		}
		else
		{
			if (nfa)
				SetFlags(SF, ((fx - fy) & 0x100) ? SF : 0);
			else
				SetFlags(SF, ((fx + fy) & 0x100) ? SF : 0);
		}
	}
	return RegF & SF;
}

uint8 Z80R::GetHF()
{
	if (uf & HF)
	{
		if (uf & WF)
		{
			if (nfa)
				SetFlags(HF, 
					((fx32 & 0x1ffful) < (fy32 & 0x1ffful)) ? HF : 0); 
			else
				SetFlags(HF, 
					(((fx32 & 0x1ffful) + (fy32 & 0x1ffful)) & 0x2000ul)
					? HF : 0); 
		}
		else
		{
			if (nfa)
				SetFlags(HF, ((fx & 0x1f) < (fy & 0x1f)) ? HF : 0);
			else
				SetFlags(HF, 
					(((fx & 0x1f) + (fy & 0x1f)) & 0x20) ? HF : 0);
		}
	}
	return RegF & HF;
}

uint8 Z80R::GetPF()
{
	if (uf & PF)
	{
		if (uf & WF)
		{
			if (nfa)
				SetFlags(PF, 
					((fx32^fy32)&(fx32^(fx32 - fy32))&0x10000ul)
					? PF : 0);
			else
				SetFlags(PF, 
					(~(fx32^fy32)&(fx32^(fx32 + fy32))&0x10000ul)
					? PF : 0);
		}
		else
		{
			if (nfa)
				SetFlags(PF, 
					((fx ^ fy) & (fx ^ (fx - fy)) & 0x100) 
					? PF : 0);
			else
				SetFlags(PF, 
					(~(fx ^ fy) & (fx ^ (fx + fy)) & 0x100)
					 ? PF : 0);
		}
	}
	return RegF & PF;
}

void Z80R::SetZS(uint8 a)
{
	SetFlags(ZF|SF, ZSPTable[a] & (ZF|SF));
	SetXF(a);
}

void Z80R::SetZSP(uint8 a)
{
	SetFlags(ZF|SF|PF, ZSPTable[a]);
	SetXF(a);
}

// ---------------------------------------------------------------------------
//	j^[
// ---------------------------------------------------------------------------

#ifdef DEBUG_MONITOR
static int fcount = 0;
#endif

void Z80R::StartMonitor()
{
#ifdef DEBUG_MONITOR
	if (fcount < 2)
	{
//		if (fcount == 0)
		{
			fpmon = fopen(fcount ? "z80_s.log" : "z80.log", "w");
			monitor = 1;
		}
		fcount++;
	}
#endif
}

void Z80R::StopMonitor()
{
#ifdef DEBUG_MONITOR
	monitor = 0;
	if (fpmon)
	{
		fclose(fpmon);
		fpmon = 0;
	}
	fcount --;
#endif
}

#ifdef DEBUG_MONITOR
void Z80R::Monitor()
{
	char f1[6], buf[20], p1[5], p2[5];
	static int id = 0;
	uint8 f;
	if (fpmon)
	{
		f = (uint8) GetAF();
		f1[0] = (f & 0x80) ? 'S' : '-'; f1[1] = (f & 0x40) ? 'Z' : '-';
		f1[2] = (f & 0x10) ? 'H' : '-';	f1[3] = (f & 0x04) ? 'P' : '-';
		f1[4] = (f & 0x02) ? 'N' : '-';	f1[5] = (f & 0x01) ? 'C' : '-';
		
//		f = (uint8) reg.r_af;
//		f2[0] = (f & 0x80) ? 'S' : '-';	f2[1] = (f & 0x40) ? 'Z' : '-';
//		f2[2] = (f & 0x10) ? 'H' : '-';	f2[3] = (f & 0x04) ? 'P' : '-';
//		f2[4] = (f & 0x02) ? 'N' : '-';	f2[5] = (f & 0x01) ? 'C' : '-';

		Disassemble(buf);
	#if 0
		fprintf(fpmon,
				"\t\t\t\t\t\t\t\t\t\t\t\t\t#%d\nPC:%.4x %-20.20s"
				"A:%.2x F:%.6s HL:%.4x DE:%.4x BC:%.4x IX:%.4x I:%.2x\n", id++,
				reg.pc, buf, RegA, f1,
				RegHL & 0xffff,RegDE & 0xffff,RegBC & 0xffff,
				reg.r.w.ix & 0xffff, reg.ireg);
		fprintf(fpmon, 
				"SP:%.4x=%.4x %.4x %.4x %.4x "
				"A:%.2x F:%.6s HL:%.4x DE:%.4x BC:%.4x IY:%.4x R:%.2x\n",
				 RegSP,
				 Read16((uint)(RegSP  )),
				 Read16((uint)(RegSP+2)),
				 Read16((uint)(RegSP+4)),
				 Read16((uint)(RegSP+6)),
				 reg.r_af/256, f2, reg.r_hl, reg.r_de, reg.r_bc, reg.r.w.iy, ((reg.rreg&0x7f)+(reg.rreg7&0x80)) );
	#else
		p1[0] = (intr_req & 0x0001 ? 'U' : '-');
		p1[1] = (intr_req & 0x0004 ? 'V' : '-');
		p1[2] = (intr_req & 0x0010 ? 'T' : '-');
		p1[3] = (intr_req & 0x0040 ? '4' : '-');
		p1[4] = (intr_req & 0x0100 ? 'S' : '-');
		p2[0] = (intr_mask & 0x0001 ? 'U' : '-');
		p2[1] = (intr_mask & 0x0004 ? 'V' : '-');
		p2[2] = (intr_mask & 0x0010 ? 'T' : '-');
		p2[3] = (intr_mask & 0x0040 ? '4' : '-');
		p2[4] = (intr_mask & 0x0100 ? 'S' : '-');

		fprintf(fpmon,
				"PC:%.4x %-20.20s  "
				"A:%.2x %.6s H:%.4x D:%.4x B:%.4x "
				"X:%.4x Y:%.4x R:%.2x SP:%.4x(%.4x %.4x %.4x) i:%c P:%.5s M:%.5s"
				"\n",
				reg.pc, buf, 
				RegA, f1, RegHL & 0xffff, RegDE & 0xffff, RegBC & 0xffff,
				reg.r.w.ix & 0xffff, reg.r.w.iy & 0xffff,  (reg.rreg & 0x7f) + (reg.rreg7 & 0x80), 
				RegSP & 0xffff,
				Read16(RegSP  ),
				Read16(RegSP+2),
				Read16(RegSP+4),
				reg.iff1 + '0', p1, p2
				);
	#endif
	}
}

typedef struct
{
	const char* text;
	int flag;
} MneTable;

static const MneTable mnetable[0x100] =
{
{"NOP",			0},	{"LD   BC,%s",	2}, {"LD   (BC),A",	0}, {"INC  BC",		0},
{"INC  B",		0}, {"DEC  B",		0}, {"LD   B,%s",   1}, {"RLCA",        0},
{"EX   AF,AF'", 0},	{"ADD  HL,BC",	0}, {"LD   A,(BC)",	0}, {"DEC  BC",		0},
{"INC  C",		0}, {"DEC  C",		0}, {"LD   C,%s",   1}, {"RRCA",        0},
{"DJNZ %s",		3},	{"LD   DE,%s",	2}, {"LD   (DE),A",	0}, {"INC  DE",		0},
{"INC  D",		0}, {"DEC  D",		0},	{"LD   D,%s",   1}, {"RLA",         0},
{"JR   %s",     3},	{"ADD  HL,DE",	0},	{"LD   A,(DE)",	0}, {"DEC  DE",		0},
{"INC  E",		0}, {"DEC  E",		0},	{"LD   E,%s",   1}, {"RRA",			0},
{"JR   NZ,%s",	3},	{"LD   HL,%s",	2},	{"LD   (%s),HL",2}, {"INC  HL",		0},
{"INC  H",		0}, {"DEC  H",		0},	{"LD   H,%s",   1}, {"DAA",         0},
{"JR   Z,%s",   3},	{"ADD  HL,HL",	0},	{"LD   HL,(%s)",2}, {"DEC  HL",		0},
{"INC  L",		0}, {"DEC  L",		0},	{"LD   L,%s",   1}, {"CPL",			0},
{"JR   NC,%s",	3},	{"LD   SP,%s",	2},	{"LD   (%s),A", 2}, {"INC  SP",		0},
{"INC  (HL)",	0}, {"DEC  (HL)",	0},	{"LD   (HL),%s",1}, {"SCF",         0},
{"JR   C,%s",   3},	{"ADD  HL,SP",	0}, {"LD   A,(%s)",	2}, {"DEC  SP",		0},
{"INC  A",		0}, {"DEC  A",		0}, {"LD   A,%s",   1}, {"CCF",			0},
{"LD   B,B",	0},	{"LD   B,C",	0},	{"LD   B,D",	0},	{"LD   B,E",	0},
{"LD   B,H",	0},	{"LD   B,L",	0},	{"LD   B,(HL)",	0},	{"LD   B,A",	0},
{"LD   C,B",	0},	{"LD   C,C",	0},	{"LD   C,D",	0},	{"LD   C,E",	0},
{"LD   C,H",	0},	{"LD   C,L",	0},	{"LD   C,(HL)",	0},	{"LD   C,A",	0},
{"LD   D,B",	0},	{"LD   D,C",	0},	{"LD   D,D",	0},	{"LD   D,E",	0},
{"LD   D,H",	0},	{"LD   D,L",	0},	{"LD   D,(HL)",	0},	{"LD   D,A",	0},
{"LD   E,B",	0},	{"LD   E,C",	0},	{"LD   E,D",	0},	{"LD   E,E",	0},
{"LD   E,H",	0},	{"LD   E,L",	0},	{"LD   E,(HL)",	0},	{"LD   E,A",	0},
{"LD   H,B",	0},	{"LD   H,C",	0},	{"LD   H,D",	0},	{"LD   H,E",	0},
{"LD   H,H",	0},	{"LD   H,L",	0},	{"LD   H,(HL)",	0},	{"LD   H,A",	0},
{"LD   L,B",	0},	{"LD   L,C",	0},	{"LD   L,D",	0},	{"LD   L,E",	0},
{"LD   L,H",	0},	{"LD   L,L",	0},	{"LD   L,(HL)",	0},	{"LD   L,A",	0},
{"LD   (HL),B",	0},	{"LD   (HL),C",	0},	{"LD   (HL),D",	0},	{"LD   (HL),E",	0},
{"LD   (HL),H",	0},	{"LD   (HL),L",	0},	{"HALT",	    0},	{"LD   (HL),A",	0},
{"LD   A,B",	0},	{"LD   A,C",	0},	{"LD   A,D",	0},	{"LD   A,E",	0},
{"LD   A,H",	0},	{"LD   A,L",	0},	{"LD   A,(HL)",	0},	{"LD   A,A",	0},
{"ADD  A,B",    0},	{"ADD  A,C",    0},	{"ADD  A,D",    0},	{"ADD  A,E",    0},
{"ADD  A,H",    0},	{"ADD  A,L",    0},	{"ADD  A,(HL)", 0},	{"ADD  A,A",    0},
{"ADC  A,B",    0},	{"ADC  A,C",    0},	{"ADC  A,D",    0},	{"ADC  A,E",    0},
{"ADC  A,H",    0},	{"ADC  A,L",    0},	{"ADC  A,(HL)", 0},	{"ADC  A,A",    0},
{"SUB  B",      0},	{"SUB  C",      0},	{"SUB  D",      0},	{"SUB  E",      0},
{"SUB  H",      0},	{"SUB  L",      0},	{"SUB  (HL)",   0},	{"SUB  A",      0},
{"SBC  A,B",    0},	{"SBC  A,C",    0},	{"SBC  A,D",    0},	{"SBC  A,E",    0},
{"SBC  A,H",    0},	{"SBC  A,L",    0},	{"SBC  A,(HL)", 0},	{"SBC  A,A",    0},
{"AND  B",      0},	{"AND  C",      0},	{"AND  D",      0},	{"AND  E",      0},
{"AND  H",      0},	{"AND  L",      0},	{"AND  (HL)",   0},	{"AND  A",      0},
{"XOR  B",      0},	{"XOR  C",      0},	{"XOR  D",      0},	{"XOR  E",      0},
{"XOR  H",      0},	{"XOR  L",      0},	{"XOR  (HL)",   0},	{"XOR  A",      0},
{"OR   B",      0},	{"OR   C",      0},	{"OR   D",      0},	{"OR   E",      0},
{"OR   H",      0},	{"OR   L",      0},	{"OR   (HL)",   0},	{"OR   A",      0},
{"CP   B",      0},	{"CP   C",      0},	{"CP   D",      0},	{"CP   E",      0},
{"CP   H",      0},	{"CP   L",      0},	{"CP   (HL)",   0},	{"CP   A",      0},
{"RET  NZ",     0},	{"POP  BC",     0},	{"JP   NZ,%s",  2}, {"JP   %s",     2},
{"CALL NZ,%s",  2}, {"PUSH BC",     0},	{"ADD  A,%s",   1},	{"RST  0",      0},
{"RET  Z",      0},	{"RET",         0},	{"JP   Z,%s",   2}, {"DB   0CBH,%s",1},
{"CALL Z,%s",   2}, {"CALL %s",     2},	{"ADC  A,%s",   1},	{"RST  8",      0},
{"RET  NC",     0},	{"POP  DE",     0},	{"JP   NC,%s",  2}, {"OUT  (%s),A", 1},
{"CALL NC,%s",  2}, {"PUSH DE",     0},	{"SUB  %s",     1},	{"RST  10H",    0},
{"RET  C",      0},	{"EXX",         0},	{"JP   C,%s",   2}, {"IN   A,(%s)", 1},
{"CALL C,%s",   2}, {NULL,          5},	{"SBC  A,%s",   1},	{"RST  18H",    0},
{"RET  PO",     0},	{"POP  HL",     0},	{"JP   PO,%s",  2}, {"EX   (SP),HL",0},
{"CALL PO,%s",  2}, {"PUSH HL",     0},	{"AND  %s",     1},	{"RST  20H",    0},
{"RET  PE",     0},	{"JP   (HL)",   0},	{"JP   PE,%s",  2}, {"EX   DE,HL",  0},
{"CALL PE,%s",  2}, {"DB   0EDH,%s",1},	{"XOR  %s",     1},	{"RST  28H",    0},
{"RET  P",      0},	{"POP  AF",     0},	{"JP   P,%s",   2}, {"DI",          0},
{"CALL P,%s",   2}, {"PUSH AF",     0},	{"OR   %s",     1},	{"RST  20H",    0},
{"RET  M",      0},	{"LD   SP,HL",  0},	{"JP   M,%s",   2}, {"EI",          0},
{"CALL M,%s",   2}, {NULL,          5},	{"CP   %s",     1},	{"RST  28H",    0},
};	

static const MneTable mnetablei[0x100] =
{
{"NOP",			0},	{"LD   BC,%s",	2}, {"LD   (BC),A",	0}, {"INC  BC",		0},	// 0
{"INC  B",		0}, {"DEC  B",		0}, {"LD   B,%s",   1}, {"RLCA",        0},
{"EX   AF,AF'", 0},	{"ADD  I%c,BC",	8}, {"LD   A,(BC)",	0}, {"DEC  BC",		0},
{"INC  C",		0}, {"DEC  C",		0}, {"LD   C,%s",   1}, {"RRCA",        0},
{"DJNZ %s",		3},	{"LD   DE,%s",	2}, {"LD   (DE),A",	0}, {"INC  DE",		0},	// 1
{"INC  D",		0}, {"DEC  D",		0},	{"LD   D,%s",   1}, {"RLA",         0},
{"JR   %s",     3},	{"ADD  I%c,DE",	8},	{"LD   A,(DE)",	0}, {"DEC  DE",		0},
{"INC  E",		0}, {"DEC  E",		0},	{"LD   E,%s",   1}, {"RRA",			0},
{"JR   NZ,%s",	3},	{"LD   I%c,%s",	10},{"LD   (%s),I%c",14},{"INC  I%c",	8},	// 2	!
{"INC  %cH",	8}, {"DEC  %cH",	8},	{"LD   %cH,%s", 9}, {"DAA",         0},
{"JR   Z,%s",   3},	{"ADD  I%c,I%c",13},{"LD   I%c,(%s)",10},{"DEC  I%c",	0},
{"INC  %cL",	8}, {"DEC  %cL",	8},	{"LD   %cL,%s", 9}, {"CPL",			0},
{"JR   NC,%s",	3},	{"LD   SP,%s",	2},	{"LD   (%s),A", 2}, {"INC  SP",		0},	// 3
{"INC  (%s)",  11}, {"DEC  (%s)",  11}, {"LD   (%s),%s",12},{"SCF",         0},
{"JR   C,%s",   3},	{"ADD  I%c,SP",	8}, {"LD   A,(%s)",	2}, {"DEC  SP",		0},
{"INC  A",		0}, {"DEC  A",		0}, {"LD   A,%s",   1}, {"CCF",			0},
{"LD   B,B",	0},	{"LD   B,C",	0},	{"LD   B,D",	0},	{"LD   B,E",	0},	// 4
{"LD   B,%cH",	8},	{"LD   B,%cL",	8},	{"LD   B,(%s)",	11},{"LD   B,A",	0},
{"LD   C,B",	0},	{"LD   C,C",	0},	{"LD   C,D",	0},	{"LD   C,E",	0},
{"LD   C,%cH",	8},	{"LD   C,%cL",	8},	{"LD   C,(%s)",	11},{"LD   C,A",	0},
{"LD   D,B",	0},	{"LD   D,C",	0},	{"LD   D,D",	0},	{"LD   D,E",	0},
{"LD   D,%cH",	8},	{"LD   D,%cL",	8},	{"LD   D,(%s)",	11},{"LD   D,A",	0},
{"LD   E,B",	0},	{"LD   E,C",	0},	{"LD   E,D",	0},	{"LD   E,E",	0},
{"LD   E,%cH",	8},	{"LD   E,%cL",	8},	{"LD   E,(%s)",	11},{"LD   E,A",	0},
{"LD   %cH,B",	8},	{"LD   %cH,C",	8},	{"LD   %cH,D",	8},	{"LD   %cH,E",	8},
{"LD   %cH,%cH",13},{"LD   %cH,%cL",13},{"LD   H,(%s)",	11},{"LD   %cH,A",	8},
{"LD   %cL,B",	8},	{"LD   %cL,C",	8},	{"LD   %cL,D",	8},	{"LD   %cL,E",	8},
{"LD   %cL,%cH",13},{"LD   %cL,%cL",13},{"LD   L,(%s)",	11},{"LD   %cL,A",	8},
{"LD   (%s),B",	11},{"LD   (%s),C",	11},{"LD   (%s),D",	11},{"LD   (%s),E",	11},
{"LD   (%s),H", 11},{"LD   (%s),L",	11},{"HALT",	    0},	{"LD   (%s),A",	11},
{"LD   A,B",	0},	{"LD   A,C",	0},	{"LD   A,D",	0},	{"LD   A,E",	0},
{"LD   A,%cH",	8},	{"LD   A,%cL",	8},	{"LD   A,(%s)",	11},{"LD   A,A",	0},
{"ADD  A,B",    0},	{"ADD  A,C",    0},	{"ADD  A,D",    0},	{"ADD  A,E",    0},
{"ADD  A,%cH",  8},	{"ADD  A,%cL",  8},	{"ADD  A,(%s)", 11},{"ADD  A,A",    0},
{"ADC  A,B",    0},	{"ADC  A,C",    0},	{"ADC  A,D",    0},	{"ADC  A,E",    0},
{"ADC  A,%cH",  8},	{"ADC  A,%cL",  8},	{"ADC  A,(%s)", 11},{"ADC  A,A",    0},
{"SUB  B",      0},	{"SUB  C",      0},	{"SUB  D",      0},	{"SUB  E",      0},
{"SUB  %cH",    8},	{"SUB  %cL",    8},	{"SUB  (%s)",   11},{"SUB  A",      0},
{"SBC  A,B",    0},	{"SBC  A,C",    0},	{"SBC  A,D",    0},	{"SBC  A,E",    0},
{"SBC  A,%cH",  8},	{"SBC  A,%cL",  8},	{"SBC  A,(%s)", 11},{"SBC  A,A",    0},
{"AND  B",      0},	{"AND  C",      0},	{"AND  D",      0},	{"AND  E",      0},
{"AND  %cH",    8},	{"AND  %cL",    8},	{"AND  (%s)",   11},{"AND  A",      0},
{"XOR  B",      0},	{"XOR  C",      0},	{"XOR  D",      0},	{"XOR  E",      0},
{"XOR  %cH",    8},	{"XOR  %cL",    8},	{"XOR  (%s)",   11},{"XOR  A",      0},
{"OR   B",      0},	{"OR   C",      0},	{"OR   D",      0},	{"OR   E",      0},
{"OR   %cH",    8},	{"OR   %cL",    8},	{"OR   (%s)",   11},{"OR   A",      0},
{"CP   B",      0},	{"CP   C",      0},	{"CP   D",      0},	{"CP   E",      0},
{"CP   %cH",    8},	{"CP   %cL",    8},	{"CP   (%s)",   11},{"CP   A",      0},
{"RET  NZ",     0},	{"POP  BC",     0},	{"JP   NZ,%s",  2}, {"JP   %s",     2},
{"CALL NZ,%s",  2}, {"PUSH BC",     0},	{"ADD  A,%s",   1},	{"RST  0",      0},
{"RET  Z",      0},	{"RET",         0},	{"JP   Z,%s",   2}, {"DB   0CBH,%s",1},
{"CALL Z,%s",   2}, {"CALL %s",     2},	{"ADC  A,%s",   1},	{"RST  8",      0},
{"RET  NC",     0},	{"POP  DE",     0},	{"JP   NC,%s",  2}, {"OUT  (%s),A", 1},
{"CALL NC,%s",  2}, {"PUSH DE",     0},	{"SUB  %s",     1},	{"RST  10H",    0},
{"RET  C",      0},	{"EXX",         0},	{"JP   C,%s",   2}, {"IN   A,(%s)", 1},
{"CALL C,%s",   2}, {"DB   0DDH",   0},	{"SBC  A,%s",   1},	{"RST  18H",    0},
{"RET  PO",     0},	{"POP  I%c",    8},	{"JP   PO,%s",  2}, {"EX   (SP),I%c",8},
{"CALL PO,%s",  2}, {"PUSH I%c",    8},	{"AND  %s",     1},	{"RST  20H",    0},
{"RET  PE",     0},	{"JP   (I%c)",  8},	{"JP   PE,%s",  2}, {"EX   DE,HL",  0},
{"CALL PE,%s",  2}, {"DB   0EDH,%s",1},	{"XOR  %s",     1},	{"RST  28H",    0},
{"RET  P",      0},	{"POP  AF",     0},	{"JP   P,%s",   2}, {"DI",          0},
{"CALL P,%s",   2}, {"PUSH AF",     0},	{"OR   %s",     1},	{"RST  20H",    0},
{"RET  M",      0},	{"LD   SP,I%c", 0},	{"JP   M,%s",   2}, {"EI",          0},
{"CALL M,%s",   2}, {"DB   0FDH",   0},	{"CP   %s",     1},	{"RST  28H",    0},
};	


//	 flag = 0		no arg				8:  XY  noarg
//	 flag = 1		arg is one byte		9:  XY	one byte
//	 flag = 2		arg is two bytes	10: XY  two bytes
//	 flag = 3		relative jump		11: XY  +d
//   flag = 4		dump 4 bytes		12: XY  +d one byte
//										13: XY  XY
//										14: two bytes	XY

void Z80R::Disassemble(char* buf)
{
	uint8 inst = Read8(reg.pc);
	char temp[10];
	uint arg = 0;
	int8 index;
	char i;
	int flag = mnetable[inst].flag;
	switch (flag)
	{
	case 0:
		strcpy(buf, mnetable[inst].text);
		return;
	case 1:
		arg = Read8(reg.pc+1);
		break;
	case 2:
		arg = Read16(reg.pc+1);
		break;
	case 3:
		arg = reg.pc + 2 + (int8) Read8(reg.pc+1);
		break;
	case 4:
		sprintf(buf, "(%.2x %.2x %.2x %.2x)", 
			Read8(reg.pc  ), Read8(reg.pc+1),
			Read8(reg.pc+2), Read8(reg.pc+3) );
		return;
	case 5:
		i = inst == 0xdd ? 'X': 'Y';
		inst = Read8(reg.pc + 1);
		flag = mnetablei[inst].flag;
		switch (flag)
		{
		case 0:
			strcpy(buf, mnetablei[inst].text);
			return;
		case 1: case 9:
			arg = Read8(reg.pc+2);
			break;
		case 2: case 10: case 14:
			arg = Read16(reg.pc+2);
			break;
		case 3:
			arg = reg.pc + 3 + (int8) Read8(reg.pc+2);
			break;
		case 4:
			sprintf(buf, "(%.2x %.2x %.2x %.2x)", 
				Read8(reg.pc  ), Read8(reg.pc+1),
				Read8(reg.pc+2), Read8(reg.pc+3) );
			return;
		case 8:
			sprintf(buf, mnetablei[inst].text, i);
			return;
		case 11:
			index = Read8(reg.pc+2);
			sprintf(temp, "I%c%+d", i, (int)index);
			sprintf(buf, mnetablei[inst].text, temp);
			return;
		case 12:
			{
				char temp2[12];
				index = Read8(reg.pc+2);
				sprintf(temp, "I%c%+d", i, (int)index);
				arg = Read8(reg.pc+2);
				if (arg < 10)
				{
					sprintf(temp2, "%d", arg);
					sprintf(buf, mnetablei[inst].text, temp, temp2);
				}
				else
				{
					sprintf(temp2, "0%XH", arg);
					if (isdigit(temp2[1]))
						sprintf(buf, mnetablei[inst].text, temp, temp2+1);
					else
						sprintf(buf, mnetablei[inst].text, temp, temp2);
				}
			}
			return;
		}
		break;
	default:
		buf[0] = 0;
		return;
	}

	if (arg < 10)
	{
		sprintf(temp, "%d", arg);
		if (flag & 8)
			sprintf(buf, mnetablei[inst].text, i, temp);
		else
			sprintf(buf, mnetable[inst].text, temp);
	}
	else
	{
		sprintf(temp, "0%XH", arg);
		if (flag == 14)
		{
			if (isdigit(temp[1]))
				sprintf(buf, mnetablei[inst].text, temp+1, i);
			else
				sprintf(buf, mnetablei[inst].text, temp, i);
		}
		else
		if (flag & 8)
		{
			if (isdigit(temp[1]))
				sprintf(buf, mnetablei[inst].text, i, temp+1);
			else
				sprintf(buf, mnetablei[inst].text, i, temp);
		}
		else
		{
			if (isdigit(temp[1]))
				sprintf(buf, mnetable[inst].text, temp+1);
			else
				sprintf(buf, mnetable[inst].text, temp);
		}
	}
}

#endif	// DEBUG_MONITOR

uint32 Z80R::GetClockCount()
{
	return 0;
}
