// ----------------------------------------------------------------------------
//	M88 - PC-88 emulator
//	Copyright (C) cisc 1998.
// ----------------------------------------------------------------------------
//	Z80 G~[VpbP[WrspNX
//	$Id: Z80Test.cpp,v 1.2 1999/01/11 15:32:19 cisc Exp $

#include <stdio.h>
#include "Z80Test.h"
#include "SubSys.h"

Z80Test::Z80Test()
{
	cpu1.Init(this); cpu2.Init(this);
	fp = fopen("z80.log", "w");
}

Z80Test::~Z80Test()
{
	if (fp)
		fclose(fp);
}

void Z80Test::Error(char* errtxt)
{
	if (fp)
	{
		fprintf(fp, "PC: %.4x   ", pc); 
		for (uint i=0; i<4; i++)
		{
			fprintf(fp, (i>codesize ? "   " : "%.2x "), code[i]);
		}

		Z80Reg* reg1 = cpu1.GetReg();
		Z80Reg* reg2 = cpu2.GetReg();

		fprintf(fp,	"%s  r:%d %d w:%d %d\n", errtxt, readcount, readcountt, writecount, writecountt);

		fprintf(fp, "     O PC:%.4x SP:%.4x AF:%.4x HL:%.4x DE:%.4x BC:%.4x IX:%.4x IY:%.4x IFF%d IR%.8x R:%.2x\n",
					reg.pc, reg.r.w.sp, reg.r.w.af, reg.r.w.hl,
					reg.r.w.de, reg.r.w.bc, reg.r.w.ix, reg.r.w.iy, reg.iff1, intr_req, reg.rreg);

		
		fprintf(fp, "     R PC:%.4x SP:%.4x AF:%.4x HL:%.4x DE:%.4x BC:%.4x IX:%.4x IY:%.4x IFF%d IR%.8x R:%.2x\n",
					reg1->pc, reg1->r.w.sp, reg1->r.w.af, reg1->r.w.hl,
					reg1->r.w.de, reg1->r.w.bc, reg1->r.w.ix, reg1->r.w.iy, reg1->iff1, cpu1.intr_req, reg1->rreg);
		
		fprintf(fp, "     T PC:%.4x SP:%.4x AF:%.4x HL:%.4x DE:%.4x BC:%.4x IX:%.4x IY:%.4x IFF%d IR%.8x R:%.2x\n\n",
					reg2->pc, reg2->r.w.sp, reg2->r.w.af, reg2->r.w.hl,
					reg2->r.w.de, reg2->r.w.bc, reg2->r.w.ix, reg2->r.w.iy, reg2->iff1, cpu2.intr_req, reg2->rreg);
	
	}
}

void Z80Test::Step(uint step)
{
	int s = step;
	if (!wait)
	{
		cpu1.gvramselected = true;
		cpu2.gvramselected = true;
		cpu1.reg.intmode = 2;
		cpu2.reg.intmode = 2;
		{
			for (; s > 0; s-=24)
			{
				Test();
				Test();
				Test();
				Test();
			}
		}
	}
}

void Z80Test::StepDual(uint step, Z80R* sub)
{
	int sss = step;
	if (!wait)
	{
		cpu1.gvramselected = true;
		cpu2.gvramselected = true;
		cpu1.reg.intmode = 2;
		cpu2.reg.intmode = 2;
		{
			for (; sss > 0; sss-=24)
			{
				Test();
				Test();
				Test();
				Test();
				sub->Step(10);
			}
		}
	}
}


void Z80Test::Test()
{
	CriticalSection::Lock lock(cs);

	Z80Reg* reg1 = cpu1.GetReg();
	Z80Reg* reg2 = cpu2.GetReg();

	// 
	readcount = writecount = readcountt = writecountt = codesize = 0;
	reg = *reg2 = *reg1;					// WX^v
	intr_req = cpu2.intr_req = cpu1.intr_req;
	cpu2.SetIMask(cpu1.intr_mask);

	pc = cpu1.reg.pc;

	// s
	cpu1.Step(1);
	cpu2.Step(1);

	// tÖvmF
	if ((reg1->r.b.flags & 0xd7) != (reg2->r.b.flags & 0xd7))
	{
		Error("tO̕sv");
		return;
	}

	// WX^̈vmF
	if (
		((reg1->r.b.a - reg2->r.b.a)
		| (reg1->r.w.bc - reg2->r.w.bc)
		| (reg1->r.w.de - reg2->r.w.de)
		| (reg1->r.w.hl - reg2->r.w.hl)
		| (reg1->r.w.ix - reg2->r.w.ix)
		| (reg1->r.w.iy - reg2->r.w.iy)
		| (reg1->r.w.sp - reg2->r.w.sp)
		| (reg1->r_af - reg2->r_af)
		| (!reg1->iff1 ^ !reg2->iff1)
		) & 0xffff)
	{
		Error("WX^̕sv");
		return;
	}

	if (cpu1.intr_req != cpu2.intr_req)
	{
		Error("荞݃tO̕sv");
		return;
	}

	if (readcount != readcountt)
	{
		Error("ǂݍ݉񐔂̕sv");
		return;
	}

	if (writecount != writecountt)
	{
		Error("݉񐔂̕sv");
		return;
	}

	// PC ̈vmF
	if (reg1->pc != reg2->pc)
	{
		Error("PC ̕sv");
		return;
	}
}

uint8 Z80Test::OpReadR(uint a)
{
	uint fcount = a - pc;
	if (fcount < 4)
	{
		codesize = (fcount > codesize) ? fcount : codesize;
		return code[fcount] = Read(a);
	}

	if (readcount < 8)
	{
		uint8 data = Read(a);
		readptr[readcount] = a;
		readdat[readcount] = data;
		readcount++;
		return data;
	}
	fprintf(fp, "%x %x\n", a, pc);
	Error("PߒɂWoCg𒴂f[^̓ǂݍ");
	return Read(a);
}

uint8 Z80Test::OpReadT(uint a)
{
	uint fcount = a - pc;
	if (fcount < 4)
		return code[fcount];
	
	readcountt++;
	for (int i=0; i<readcount; i++)
	{
		if (readptr[i] == a)
			return readdat[i];
	}

	char buf[128];
	sprintf(buf, "ǂݍ݃AhX̕sv: %.4x", a);
	Error(buf);
	return 0;
}

void Z80Test::OpWriteR(uint a, uint d)
{
	if (writecount < 8)
	{
		writeptr[writecount] = a;
		writedat[writecount] = d;
		writecount ++;
		Write(a, d);
		return;
	}
	Error("PߒɂWoCg𒴂f[^̏");
}

void Z80Test::OpWriteT(uint a, uint d)
{
	writecountt++;
	for (int i=0; i<writecount; i++)
	{
		if (writeptr[i] == a)
		{
			if (writedat[i] != d)
			{
				char buf[128];
				sprintf(buf, "݃f[^̕sv at %.4x:%.2x %.2x", a, writedat[i], d);
				Error(buf);
			}
			return;
		}
	}
	char buf[128];
	sprintf(buf, "݃AhX̕sv: %.4x", a);
	Error(buf);
}

uint8 Z80Test::OpInR(uint a)
{
	uint8 data = InPort(a);
	inptr = a;
	indat = data;
	return data;
}

uint8 Z80Test::OpInT(uint a)
{
	if (inptr == a)
		return indat;
	
	Error("̓|[g̕sv");
	return 0;
}

void Z80Test::OpOutR(uint a, uint d)
{
	outptr = a;
	outdat = d;
	OutPort(a, d);
}

void Z80Test::OpOutT(uint a, uint d)
{
	if (outptr == a)
	{
		if (outdat != d)
			Error("o̓f[^̕sv");
		return;
	}
	Error("o̓|[g̕sv");
}

void Z80Test::Reset(int rstat)
{
	cpu1.Reset(rstat);
	cpu2.Reset(rstat);
	wait = rstat ? true : false;
}

void Z80Test::NMI()
{
	cpu1.NMI();
	cpu2.NMI();
}

void Z80Test::RequestInterrupt(uint intno)
{
	CriticalSection::Lock lock(cs);
	cpu1.RequestInterrupt(intno);
	cpu2.RequestInterrupt(intno);
}

void Z80Test::CancelInterrupt(uint intno)
{
	CriticalSection::Lock lock(cs);
	cpu1.CancelInterrupt(intno);
	cpu2.CancelInterrupt(intno);
}
