// ----------------------------------------------------------------------------
//	M88 - PC-88 emulator
//	Copyright (C) cisc 1998.
// ----------------------------------------------------------------------------
//	Z80 G~[VpbP[WrspNX
//	$Id: Z80Test.h,v 1.2 1999/01/11 15:32:19 cisc Exp $


#ifndef Z80Test_h
#define Z80Test_h

#include <stdio.h>
#include "WinCriticalSection.h"
#include "Z80.h"

#include "Z80R.h"			// reference package
//#include "Z80C.h"			// target package
#include "Z80_x86.h"
//#include "Z80MK.h"

class Z80Test
{
	class CPURef : public Z80_x86
	{
	public:
		Z80Test* base;

		Z80Reg* GetReg() { return &reg; }
		void Init(Z80Test* _base) { base = _base; }
		void AcceptInterrupt(uint a) { base->AcceptInterrupt(a); }

	private:
		uint8 Read(uint a) { return base->OpReadR(a); }
		void Write(uint a, uint d) { base->OpWriteR(a, d); }
		uint8 InPort(uint p) { return base->OpInR(p); }
		void OutPort(uint p, uint d) { base->OpOutR(p, d); }
	
		friend class Z80Test;
	};

	class CPUTarget : public Z80R
	{
	public:
		Z80Test* base;

		Z80Reg* GetReg() { return &reg; }
		void Init(Z80Test* _base) { base = _base; }
		void AcceptInterrupt(uint) {}

	private:
		uint8 Read(uint a) { return base->OpReadT(a); }
		void Write(uint a, uint d) { base->OpWriteT(a, d); }
		uint8 InPort(uint p) { return base->OpInT(p); }
		void OutPort(uint p, uint d) { base->OpOutT(p, d); }

		friend class Z80Test;
	};

private:
	uint codesize;
	CPURef    cpu1;
	CPUTarget cpu2;
	CriticalSection cs;

	uint pc;
	uint8 code[4];
	uint  readptr[8], writeptr[8], inptr, outptr;
	uint8 readdat[8], writedat[8], indat, outdat;
	int readcount, writecount;
	int readcountt, writecountt;

	void Test();
	void Error(char*);

	FILE* fp;
	uint32 intr_req;

public:
	uint8 OpReadR(uint a);
	void  OpWriteR(uint a, uint d);
	uint8 OpInR(uint a);
	void  OpOutR(uint a, uint d);

	uint8 OpReadT(uint a);
	void  OpWriteT(uint a, uint d);
	uint8 OpInT(uint a);
	void  OpOutT(uint a, uint d);

	virtual void AcceptInterrupt(uint a) = 0;

protected:
	virtual uint8 Read(uint a) = 0;
	virtual void Write(uint a, uint d) = 0;
	virtual uint8 InPort(uint p) = 0;
	virtual void OutPort(uint p, uint d) = 0;
	void SetIMask(uint32 im) { cpu1.SetIMask(im); cpu2.SetIMask(im); }
	Z80Reg reg;


public:
	Z80Test();
	~Z80Test();

	void StartMonitor() { return; }
	void StopMonitor() { return; }
	uint32 GetCount() { return 0; }

	void Reset(int rstat);
	void NMI();
	void Step(uint);
	void StepDual(uint, Z80R*);
	void StepDual2(uint c, Z80R* r) { StepDual(c, r); }
	void StepDual3(uint c, Z80R* r) { StepDual(c, r); }
	void RequestInterrupt(uint intno);
	void CancelInterrupt(uint intno);

	volatile bool wait;
	bool gvramselected;
	uint8* readaccess[256];
	uint8* writeaccess[256];		
};


#endif // Z80Test_h
